package JavaQuat20;

import java.math.*;


public class BigQuaternion
{
  public double r;
  public double i;
  public double j;
  public double k;
  
  public BigQuaternion(double ra, double ia, double ja, double ka)
  {
    r = ra; i = ia; j = ja; k = ka;
  }
  
  public BigQuaternion(double ra, double ia, double ja, double ka, int scaleDesired)
  {
      r = ra;
      i = ia; 
      j = ja; 
      k = ka;
  }
  
  public BigQuaternion(BigDecimal ra, BigDecimal ia, BigDecimal ja, BigDecimal ka)
  {
    r = ra.doubleValue(); i = ia.doubleValue(); j = ja.doubleValue(); k = ka.doubleValue();
  }
  
  public BigQuaternion(BigDecimal ra, BigDecimal ia, BigDecimal ja, BigDecimal ka, int scaleDesired)
  {
      r = ra.doubleValue();
      i = ia.doubleValue(); 
      j = ja.doubleValue(); 
      k = ka.doubleValue();
  }
  
  public BigQuaternion ()
  {
    r = i = j = k = 0.0;
  }
  
  public BigQuaternion (BigQuaternion arg)
  {
    r = arg.r;
    i = arg.i;
    j = arg.j;
    k = arg.k;
  }
  
  public void setR (double arg)
  {
    r = arg;
  }
  
  public void setI (double arg)
  {
    i = arg;
  }
  
  public void setJ (double arg)
  {
    j = arg;
  }
  
  public void setK (double arg)
  {
    k = arg;
  }
  
  public void setR (BigDecimal arg)
  {
    r = arg.doubleValue();
  }
  
  public void setI (BigDecimal arg)
  {
    i = arg.doubleValue();
  }
  
  public void setJ (BigDecimal arg)
  {
    j = arg.doubleValue();
  }
  
  public void setK (BigDecimal arg)
  {
    k = arg.doubleValue();
  }
  
  // Misc
  
  public String rToString ()
  {
    return String.valueOf(r);
  }
  
  public String iToString ()
  {
    return String.valueOf(i);
  }
  
  public String jToString ()
  {
    return String.valueOf(j);
  }
  
  public String kToString ()
  {
    return String.valueOf(k);
  }
  
  public int scale ()
  {
    return 16;
  }
  
  
  public BigQuaternion assign_from (BigQuaternion arg)
  {
    r = arg.r;
    i = arg.i;
    j = arg.j;
    k = arg.k;
    return this;
  }
  
  public boolean equal_p (BigQuaternion arg)
  {
    return (r == (arg.r) &&
            i == (arg.i) &&
            j == (arg.j) &&
            k == (arg.k) );
  }

  // Math

  public BigQuaternion plus (BigQuaternion arg)
  {
    return new BigQuaternion(r + arg.r, i + arg.i, j + arg.j, k + arg.k);
  }
  
  public BigQuaternion plus_equal (BigQuaternion arg)
  {
    r = r + arg.r;
    i = i + arg.i;
    j = j + arg.j;
    k = k + arg.k;
    return this;
  }
  
  public BigQuaternion plus_equal (FastQuaternion arg)
  {
    return this.plus_equal(new BigQuaternion(arg.r, arg.i, arg.j, arg.k));
  }
  
   public BigQuaternion minus (BigQuaternion arg)
  {
    return new BigQuaternion(r - arg.r, i - arg.i, j - arg.j, k - arg.k);
  }
  
  public BigQuaternion minus_equal (BigQuaternion arg)
  {
    r = r - arg.r;
    i = i - arg.i;
    j = j - arg.j;
    k = k - arg.k;
    return this;
  }
  
 public BigQuaternion times (BigQuaternion point)
  {
     return new BigQuaternion(r * point.r - i * point.i - j * point.j - k * point.k,
                           r * point.i + i * point.r + j * point.k - k * point.j,
                           r * point.j - i * point.k + j * point.r + k * point.i,
                           r * point.k + i * point.j - j * point.i + k * point.r
                          );
  }
  
  public BigQuaternion times_scalar_n (double scalar)
  {
    r = r * scalar;
    i = i * scalar;
    j = j * scalar;
    k = k * scalar;
    return this;
  }
  
  public BigQuaternion times_scalar_n (int scalar)
  {
    r = r * scalar;
    i = i * scalar;
    j = j * scalar;
    k = k * scalar;
    return this;
  }
  
  public BigQuaternion times_scalar (double scalar)
  {
    return new BigQuaternion(r * scalar, i * scalar, j * scalar, k * scalar);
  }
  
  public BigQuaternion times_scalar (int scalar)
  {
    return new BigQuaternion(r * scalar, i * scalar, j * scalar, k * scalar);
  }
  
  public BigQuaternion times_equal (BigQuaternion point)
  {
    double rr, ii, jj, kk;
    rr = r * point.r - i * point.i - j * point.j - k * point.k;
    ii = r * point.i + i * point.r + j * point.k - k * point.j;
    jj = r * point.j - i * point.k + j * point.r + k * point.i;
    kk = r * point.k + i * point.j - j * point.i + k * point.r;
    r = rr;
    i = ii;
    j = jj;
    k = kk;
    return this;
  }
  
  public void setScale (int scale)
  {
  }
  
  public BigQuaternion setScaleLessOrEqual (int scale)
  {
    return this;   
  }
  
  public FastQuaternion fastQuaternionValue ()
  {
    return new FastQuaternion(r, i, j, k);
  }
  
  public double dot (BigQuaternion point)
  {
    return (r * point.r + i * point.i + j * point.j + k * point.k);
  }
  
  public BigQuaternion projection_onto (BigQuaternion normal_q)
  {
    return normal_q.times_scalar(this.dot(normal_q));
  }
  
  public BigQuaternion orthogonal_component_to (BigQuaternion normal_q)
   // We are a little extra careful here in case of parallel hidden by round off error.
   // Returns a new quaternion that is normalized.
  {
    BigQuaternion approx_answer = this.minus(this.projection_onto(normal_q));
    double answer_size = approx_answer.magnitude();
    if (answer_size < 1.0e-20)
      return new BigQuaternion();
    approx_answer.normalize_n();
  
    BigQuaternion better_normal = approx_answer.minus(approx_answer.projection_onto(normal_q));
    if (better_normal.magnitude() < 0.9)
      return new BigQuaternion(); // This is so nearly parallel to normal_q, answer 0.
//    if (Math.abs(better_normal.r.doubleValue()) < 1.0e-15)
//      better_normal.r = 0;
//    if (Math.abs(better_normal.i) < 1.0e-15)
//      better_normal.i = 0;
//    if (Math.abs(better_normal.j) < 1.0e-15)
//      better_normal.j = 0;
//    if (Math.abs(better_normal.k) < 1.0e-15)
//      better_normal.k = 0;
//    if (better_normal.magnitude() < 1.0e-15)
//      return new BigQuaternion();

    return better_normal.normalize_n().times_scalar(answer_size);
  }
  
  public double magnitude ()
  {
    return Math.sqrt(r * r + i * i + j * j + k * k);
  }
  
  public double magnitude_sq ()
  {
    return (r * r + i * i + j * j + k * k);
  }
  
  public BigQuaternion normalize_n ()
  {
    double mag = magnitude();
    if (mag <= 0) {
      System.err.println("BigQuaternion normalize_n called with a zero size vector");
    } else {
      double magInverse = 1 / mag;
      r = r * magInverse;
      i = i * magInverse;
      j = j * magInverse;
      k = k * magInverse;
    }
    return this;
  }
  
  public boolean normalizePerpendicularTo (BigQuaternion baseAxis)
    // Normalizes baseAxis destructively, sets this to normalized 
    // perpendicular component of this to baseAxis.
    // Returns true iff able to complete these operations.
  {
    if (baseAxis.magnitude() == 0) {
//      System.err.println("BigQuaternion:normalizePerpendicularTo - baseAxis must be non-zero.");
      return false;
    }
    
    if (magnitude() == 0) {
//      System.err.println("BigQuaternion:normalizePerpendicularTo - This must be non-zero.");
      return false;
    }
      
    BigQuaternion self = new BigQuaternion(this);
    BigQuaternion baseCopy = new BigQuaternion(baseAxis);
    baseCopy.normalize_n();
    self.normalize_n();
    self = self.orthogonal_component_to(baseCopy);
    if (self.magnitude() < 1.0e-15) {
//      System.err.println("BigQuaternion:normalizePerpendicularTo -  vectors are parallel.  That's invalid.");
      return false;
    }
    self.normalize_n();
    this.assign_from(self);
    baseAxis.assign_from(baseCopy);
    return true;
  }
    
  static public BigQuaternion rotate (double degrees, BigQuaternion base, BigQuaternion target)
    // Assigns self a value that is rotated 'degrees' from 'base' toward 'target'.
  {
    double radians = degrees * Math.PI / 180.0;
    double sin = Math.sin(radians);
    double cos = Math.cos(radians);
    if (cos == -0.0) cos = 0.0;
    if (sin == -0.0) sin = 0.0;
    return (base.times_scalar(cos).plus_equal(target.times_scalar(sin)));
  }
  
  public void print (String title)
  {
    System.out.println(title + " r=" + r + " i=" + i + " j=" + j + " k=" + k);
  }
  
  public String toString ()
  {
    return "<BigQuaternion r=" + r + " i=" + i + " j=" + j + " k=" + k + ">";
  }

}
