package JavaQuat20;

import java.awt.*;
import java.awt.image.ImageObserver;
import java.awt.event.*;

class ColorMapBox extends Canvas implements MouseListener
  // This counts on the ColorEditor being instantiated first.
{
  Color [] colors;
  ColorMapEditor parent;
  int rows;
  int columns;
  int height;
  int width;
  int rectWidth;
  int rectHeight;
  int numberOfColors = 1;
  int localColorMapIndex = 0;
  String [] hex = {"0", "1", "2", "3", "4", "5", "6", "7", "8","9", "A", "B", "C","D", "E", "F"};
  
  ColorMapBox (ColorMapEditor target, int rows1, int columns1, int height1)
    // This counts on the ColorEditor being instantiated first.
  {
    parent = target;
    rows = rows1;
    columns = columns1;
    height = height1;
    width = getBounds().width;
    rectWidth = width / (columns + 1);
    rectHeight = height / (rows + 1);
    numberOfColors = rows * columns;
    colors = new Color[numberOfColors];
    for(int i = 0; i < numberOfColors; i++) {
      colors[i] = Color.black;
    }
    //System.out.println("Exiting ColorMapBox");
    parent.colorEditor.setColor(colors[0]);
    addMouseListener(this);
  }
  
  public void rampColors(int index1, Color color1, int index2, Color color2, int editIndex, 
                         boolean rgb_p, boolean reverseP)
     // reverseP is meaningless for RGB ramping
  {
    int r = color1.getRed();
    int g = color1.getGreen();
    int b = color1.getBlue();
    if (index1 > index2) {
      index2 += numberOfColors;
    }
    double deltaIndex = (double)(index2 - index1);
    if (deltaIndex == 0)
      return;
    int i, inc;
    Color newColor;
    if (rgb_p) {
      double incR = (color2.getRed() - r)/deltaIndex;
      double incG = (color2.getGreen() - g)/deltaIndex;
      double incB = (color2.getBlue() - b)/deltaIndex;
      for (i = index1 + 1, inc = 1; i < index2; i++, inc++) {
        int red =   Math.max(0, Math.min(255, r + (int)(inc * incR)));
        int green = Math.max(0, Math.min(255, g + (int)(inc * incG)));
        int blue =  Math.max(0, Math.min(255, b + (int)(inc * incB)));
        int colorIndex = i % numberOfColors;
        newColor = new Color(red, green, blue);
        colors[colorIndex] = newColor;
        parent.parent.setIndexAndColor(colorIndex, newColor, i == index2 - 1); //3rd arg refreshes canvas
      }
    } else if (reverseP) {
      float[] hsb1 = Color.RGBtoHSB(r, g, b, (new float[3]));
      float[] hsb2 = Color.RGBtoHSB(color2.getRed(), color2.getGreen(), color2.getBlue(), (new float[3]));
      double incH = (hsb2[0] - hsb1[0]);
      if (incH < 0)
        incH += 1;
      incH /= deltaIndex;
      double incS = (hsb2[1] - hsb1[1])/deltaIndex;
      double incB = (hsb2[2] - hsb1[2])/deltaIndex;
      for (i = index1 + 1, inc = 1; i < index2; i++, inc++) {
        float newh = (float)Math.max(0.0, Math.min(1.0, (2.0 + hsb1[0] + (inc * incH)) % 1.0));
        float news = (float)Math.max(0.0, Math.min(1.0, hsb1[1] + (inc * incS)));
        float newb = (float)Math.max(0.0, Math.min(1.0, hsb1[2] + (inc * incB)));
        int colorIndex = i % numberOfColors;
        newColor = Color.getHSBColor(newh, news, newb);
        colors[colorIndex] = newColor;
        parent.parent.setIndexAndColor(colorIndex, newColor, i == index2 - 1); //3rd arg refreshes canvas
      }
    } else {
      float[] hsb1 = Color.RGBtoHSB(r, g, b, (new float[3]));
      float[] hsb2 = Color.RGBtoHSB(color2.getRed(), color2.getGreen(), color2.getBlue(), (new float[3]));
      double incH = (hsb1[0] - hsb2[0]); // Changed
      if (incH < 0)
        incH += 1;
      incH /= deltaIndex;
      double incS = (hsb2[1] - hsb1[1])/deltaIndex;
      double incB = (hsb2[2] - hsb1[2])/deltaIndex;
      for (i = index1 + 1, inc = 1; i < index2; i++, inc++) {
        float newh = (float)Math.max(0.0, Math.min(1.0, (2.0 + hsb1[0] - (inc * incH)) % 1.0));//Changed
        float news = (float)Math.max(0.0, Math.min(1.0, hsb1[1] + (inc * incS)));
        float newb = (float)Math.max(0.0, Math.min(1.0, hsb1[2] + (inc * incB)));
        int colorIndex = i % numberOfColors;
        newColor = Color.getHSBColor(newh, news, newb);
        colors[colorIndex] = newColor;
        parent.parent.setIndexAndColor(colorIndex, newColor, i == index2 - 1); //3rd arg refreshes canvas
      }
    }
    repaint();

    // Update the ColorEditor if it's color changed.
    if ((index1 <= editIndex && editIndex <= index2) || editIndex <= index2 - numberOfColors)
      parent.colorEditor.setColor(colors[editIndex]);
  }
  
  public int [] sortColors (int index1, int index2, Cmp comp)
  {
    if (parent.debugP()) {
      System.out.println("ColorMapBox index1 = " + hex[index1 / 16] + hex[index1 % 16] + "index2=" +
                         hex[index2 / 16] + hex[index2 % 16]);
    }
    if (index2 < index1) {
      index2 += 256;
    }
    int i;
    SortableVector sortedColors = new SortableVector(index2 - index1 + 1);
    for (i = index1; i <= index2; i++) {
      sortedColors.addElement(colors[i % 256]);
    }
    if (parent.debugP()) {
      for (i = 0; i < sortedColors.size(); i++) {
        System.out.println("ColorMapBox sortedColors[" + i + "] is " + colors[i]);
      }
    }
    int [] subMap = sortedColors.mappedQsort(comp);
    if (parent.debugP()) {
      for (i = 0; i < sortedColors.size(); i++) {
        System.out.println("  ColorMapBox sorted sortedColors[" + i + "] is " + colors[i] +
               " and subMap=" + subMap[i]);
      }
    }
    int sortIndex = 0;
    int ovarAllIndex;
    for (i = index1; i <= index2; i++, sortIndex++) {
      ovarAllIndex = i % 256;
      if (parent.debugP()) {
        System.out.println("ColorMapBox sortIndex=" + sortIndex + "'s color=" + 
                sortedColors.elementAt(sortIndex) + " put at location " + ovarAllIndex);
      }
      colors[ovarAllIndex] = (Color)sortedColors.elementAt(sortIndex);
      parent.parent.setIndexAndColor(ovarAllIndex, colors[ovarAllIndex], i == index2); //3rd arg refreshes canvas
    }
    repaint();
    if (subMap.length < 256) {
      int [] fullMap = new int [256];
      for (i = 0; i < 256; i++) {
        fullMap[i] = i;
      }
      int subIndex = 0;
      for (i = index1; i <= index2; i++, subIndex++) {
        fullMap[i % 256] = (subMap[subIndex] + index1) % 256;
      }
      if (parent.debugP()) {
        System.out.println("  ColorMapBox fullMap:");
        for (i = 0; i < sortedColors.size(); i++) {
          System.out.print(" [" + i + "] = " + fullMap[i]);
          if (i % 16 == 0) System.out.println(".");
        }
      }
      return fullMap;
    }
    return subMap;
  }
    

  public void rotateHues (int index1, int index2, int editIndex, int degrees1)
     // reverseP is meaningless for RGB ramping
  {
    int r;
    int g;
    int b;
    //System.out.println("ColorMapBox.rotateHues Entry");
    double deltaIndex = (double)(index2 - index1);
    if (deltaIndex == 0 || degrees1 == 0) {
      return;
    }
    float hueInc = (((degrees1 % 360) + 360) % 360) / (float)360; //range is [0, 1]
    int i, colorIndex;
    float newHue;
    Color newColor;
    float[] hsb1 = new float[3];
    if (index2 < index1) {
      index2 += 256;
    }
    for (i = index1; i <= index2; i++) {
      colorIndex = i % 256;
      r = colors[colorIndex].getRed();
      g = colors[colorIndex].getGreen();
      b = colors[colorIndex].getBlue();
      hsb1 = Color.RGBtoHSB(r, g, b, (new float[3]));
      if (hsb1[1] > 0 &&  hsb1[2] > 0) { // When white or black don't bother.
        newHue = (float)Math.max(0.0, Math.min(1.0, (hsb1[0] + hueInc) % 1.0));
        newColor = Color.getHSBColor(newHue, hsb1[1], hsb1[2]);
        colors[colorIndex] = newColor;
        parent.parent.setIndexAndColor(colorIndex, colors[colorIndex], i == index2); //3rd arg refreshes canvas
      } else if (i == index2) {
        parent.parent.setIndexAndColor(colorIndex, colors[colorIndex], true); //3rd arg refreshes canvas
      }
    }
    repaint();

    // Update the ColorEditor if it's color changed.
    if ((index1 <= editIndex && editIndex <= index2) || editIndex <= index2 - numberOfColors) {
      parent.colorEditor.setColor(colors[editIndex]);
    }
  }
  
  public void multiplySaturation (int index1, int index2, int editIndex, double multiplier)
     // reverseP is meaningless for RGB ramping
  {
    int r;
    int g;
    int b;
    //System.out.println("ColorMapBox.rotateHues Entry");
    double deltaIndex = (double)(index2 - index1);
    if (multiplier == 1.0) {
      return;
    }
    int i, colorIndex;
    float newSat;
    Color newColor;
    float[] hsb1 = new float[3];
    if (index2 < index1) {
      index2 += 256;
    }
    for (i = index1; i <= index2; i++) {
      colorIndex = i % 256;
      r = colors[colorIndex].getRed();
      g = colors[colorIndex].getGreen();
      b = colors[colorIndex].getBlue();
      hsb1 = Color.RGBtoHSB(r, g, b, (new float[3]));
      newSat = (float)Math.abs(Math.min(1.0, (hsb1[1] * multiplier)));
      newColor = Color.getHSBColor(hsb1[0], newSat, hsb1[2]);
      colors[colorIndex] = newColor;
      parent.parent.setIndexAndColor(colorIndex, colors[colorIndex], i == index2); //3rd arg refreshes canvas
    }
    repaint();

    // Update the ColorEditor if it's color changed.
    if ((index1 <= editIndex && editIndex <= index2) || editIndex <= index2 - numberOfColors) {
      parent.colorEditor.setColor(colors[editIndex]);
    }
  }
  
  public void multiplyBrightness (int index1, int index2, int editIndex, double multiplier)
     // reverseP is meaningless for RGB ramping
  {
    int r;
    int g;
    int b;
    //System.out.println("ColorMapBox.rotateHues Entry");
    double deltaIndex = (double)(index2 - index1);
    if (multiplier == 1.0) {
      return;
    }
    int i, colorIndex;
    float newBri;
    Color newColor;
    float[] hsb1 = new float[3];
    if (index2 < index1) {
      index2 += 256;
    }
    for (i = index1; i <= index2; i++) {
      colorIndex = i % 256;
      r = colors[colorIndex].getRed();
      g = colors[colorIndex].getGreen();
      b = colors[colorIndex].getBlue();
      hsb1 = Color.RGBtoHSB(r, g, b, hsb1);
      newBri = (float)Math.abs(Math.min(1.0, (hsb1[2] * multiplier)));
      newColor = Color.getHSBColor(hsb1[0], hsb1[1], newBri);
      colors[colorIndex] = newColor;
      parent.parent.setIndexAndColor(colorIndex, colors[colorIndex], i == index2); //3rd arg refreshes canvas
    }
    repaint();

    // Update the ColorEditor if it's color changed.
    if ((index1 <= editIndex && editIndex <= index2) || editIndex <= index2 - numberOfColors) {
      parent.colorEditor.setColor(colors[editIndex]);
    }
  }
  

  int getIndexMouse (int x, int y) 
    // x and y are pixel coordinates from top left of this.
  {
    if (width == 0) {
      width = getBounds().width;
      rectWidth = width / (columns + 1);
      System.out.println("ColorMapBox.getIndexMouse reset rectWidth to " + rectWidth);
    }
    int row1 = (int)Math.floor(y / (double) rectHeight) - 1;
        row1 = Math.min(row1, rows - 1);
        row1 = Math.max(row1, 0);
    int column1 = (int)Math.floor(x / (double) rectWidth) - 1;
        column1 = Math.min(column1, columns - 1);
        column1 = Math.max(column1, 0);
    return Math.min(row1 * columns + column1, numberOfColors);
  }
  
  Dimension getRowColMouse (int index) 
    // x and y are pixel coordinates from top left of this.
  {
    if (width == 0) {
      width = getBounds().width;
      rectWidth = width / (columns + 1);
//      System.out.println("ColorMapBox.getRowColMouse reset rectWidth to " + rectWidth);
    }
    int row1 = index / columns;
    int column1 = 1 + index - row1 * columns;
//    System.out.println("ColorMapBox.getRowColMouse columns=" + columns + " row1 =" + row1 +
//                       " column1=" + (column1 - 1) + " x=" + (column1 * rectWidth) + " y=" +
//                       ((row1 + 1) * rectHeight));
    return new Dimension(column1 * rectWidth, (row1 + 1) * rectHeight);
  }
  
  void paintHex (Graphics g)
  {
    int extensionV = rectHeight - 3;
    int insetH = 2;
    int insetV = rectHeight - 2;
    int extensionH = 2;
    for(int i = 0; i < 16; i++) {
      extensionV += rectHeight;
      extensionH += rectWidth;
      g.drawString(hex[i], extensionH, insetV);
      g.drawString(hex[i], insetH, extensionV);
    }
  }
      
  
  public void paint (Graphics g)
  {
    Dimension corner;
//    if (width == 0) {
      width = getBounds().width;
      int realHeight = getBounds().height;
      rectWidth = width / (columns + 1);
      rectHeight = realHeight / (rows + 1);
//      System.out.println("ColorMapBox.paint reset rectWidth to " + rectWidth);
//    }
    paintHex(g);
    for(int i = 0; i < numberOfColors; i++) {
      g.setColor(colors[i]);
      corner = getRowColMouse(i);
      g.fillRect(corner.width, corner.height, rectWidth - 2, rectHeight - 2);
    }
  }
  
  // Implementation of MouseListener
  
  public void mouseClicked(MouseEvent e) {}

  public void mousePressed(MouseEvent event1)
  {
    int i = getIndexMouse(event1.getX(), event1.getY());
//      System.out.println("ColorMapBox.handleEvent i = " + i);
    parent.setIndexAndColor(i + parent.offsetIntoColorMap, colors[i], true);
  }

  public void mouseReleased(MouseEvent e) {}

  public void mouseEntered(MouseEvent e) {}

  public void mouseExited(MouseEvent e) {}
  
//  public boolean handleEvent (Event event)
//   //This is right out of the API documentation for JDK 1.0.2
  
  public void setColor (Color c)
  {
//    System.out.println("CMBox setColor c = " + c + " localIndex=" + localColorMapIndex);
    colors[localColorMapIndex] = c;
    Dimension corner = getRowColMouse(localColorMapIndex);
    Graphics g = getGraphics();
    if (g == null) {
      repaint();
    } else {
      g.setColor(colors[localColorMapIndex]);
      g.fillRect(corner.width, corner.height, rectWidth - 2, rectHeight - 2);
    }
  }
  
  void setColorAt (int index, Color c)
  {
    colors[index] = c;
    Dimension corner = getRowColMouse(index);
    Graphics g = getGraphics();
    if (g == null) {
      repaint();
    } else {
      g.setColor(colors[index]);
      g.fillRect(corner.width, corner.height, rectWidth - 2, rectHeight - 2);
    }
  }

  public Insets insets ()
  {
    return new Insets(5, 5, 5, 5);
  }

  
  public Dimension getMinimumSize ()
  {
    return new Dimension(290, height);
  }
  
  public Dimension getPreferredSize ()
  {
    return new Dimension(290, height);
  }
}
  
