package JavaQuat20;

import java.math.*;

public class Complex
{
  public double r;
  public double i;
  public double j;
  public double k;
  
  public Complex(double ra, double ia)
  {
    r = ra; i = ia;
  }
  
  public Complex ()
  {
    r = i = 0;
  }
  
  public Complex (Complex arg)
  {
    r = arg.r;
    i = arg.i;
  }
  
  public void setR (double arg)
  {
    r = arg;
  }
  
  public void setI (double arg)
  {
    i = arg;
  }
  
  // Misc
  
  public String rToString ()
  {
    return String.valueOf(r);
  }
  
  public String iToString ()
  {
    return String.valueOf(i);
  }
  
  public Complex assign_from (Complex arg)
  {
    r = arg.r;
    i = arg.i;
    return this;
  }
  
  public boolean equal_p (Complex arg)
  {
    return (r == arg.r &&
            i == arg.i);
  }

  // Math

  public Complex plus (Complex arg)
  {
    return new Complex(r + arg.r, i + arg.i);
  }
  
  public Complex plus_equal (Complex arg)
  {
    r = r + arg.r;
    i = i + arg.i;
    return (Complex)this;
  }
  
   public Complex minus (Complex arg)
  {
    return (Complex)new Complex(r - arg.r, i - arg.i);
  }
  
  public Complex minus_equal (Complex arg)
  {
    r = r - arg.r;
    i = i - arg.i;
    return (Complex)this;
  }
  
  public Complex times (Complex point)
  {
     return (Complex)
      new Complex(r * point.r - i * point.i,
                  r * point.i + i * point.r);
  }
  
  public Complex times_scalar_n (double scalar)
  {
    r = r * scalar;
    i = i * scalar;
    return (Complex)this;
  }
  
  public Complex times_scalar_n (int scalar)
  {
    r = r * scalar;
    i = i * scalar;
    return (Complex)this;
  }
  
  public Complex times_scalar (double scalar)
  {
    return (Complex)new Complex(r * scalar, i * scalar);
  }
  
  public Complex times_scalar (int scalar)
  {
    return (Complex)new Complex(r * scalar, i * scalar);
  }
  
  public Complex times_equal (Complex point)
  {
    double rr, ii;
    rr = r * point.r - i * point.i;
    ii = r * point.i + i * point.r;
    r = rr;
    i = ii;
    return (Complex)this;
  }
  
  public Complex toPower (int power)
    // this works for non-negative integral powers.
  {
  	switch (power) {
  	case 0: return new Complex(1, 0);
  	case 1: return new Complex(this);
  	case 2: return times(this);
  	case 3: return times(this).times(this);
  	default: 
  	  if (power%2==1) {
  	    Complex temp = new Complex(toPower((power - 1)/2));
  	    return temp.times_equal(temp).times_equal(this);
  	  }
  	  Complex temp = new Complex(toPower(power/2));
  	  return temp.times_equal(temp);
    }
  }
  
  public double dot (Complex point)
  {
    return (r * point.r + i * point.i);
  }
  
  public Complex projection_onto (Complex normal_q)
  {
    return normal_q.times_scalar(this.dot(normal_q));
  }
  
  public Complex orthogonal_component_to (Complex normal_q)
   // We are a little extra careful here in case of parallel hidden by round off error.
  {
    Complex approx_answer = this.minus(this.projection_onto(normal_q));
    double answer_size = approx_answer.magnitude();
    if (answer_size < 1.0e-20)
      return new Complex();
    approx_answer.normalize_n();
  
    Complex better_normal = approx_answer.minus(approx_answer.projection_onto(normal_q));
    if (better_normal.magnitude() < 0.9)
      return new Complex(); // This is so nearly parallel to normal_q, answer 0.
    if (Math.abs(better_normal.r) < 1.0e-15)
      better_normal.r = 0;
    if (Math.abs(better_normal.i) < 1.0e-15)
      better_normal.i = 0;
    if (better_normal.magnitude() < 1.0e-15)
      return new Complex();

    return better_normal.normalize_n().times_scalar(answer_size);
  }
  
  public double magnitude ()
  {
    return Math.sqrt(r*r + i*i);
  }
  
  public double magnitude_sq ()
  {
    return (r*r + i*i);
  }
  
  public Complex normalize_n ()
  {
    double mag = magnitude();
    r = r / mag;
    i = i / mag;
    return this;
  }
  
  public boolean normalizePerpendicularTo (Complex baseAxis)
    // Normalizes baseAxis destructively, sets this to normalized 
    // perpendicular component of this to baseAxis.
    // Returns true iff able to complete these operations.
  {
    if (baseAxis.magnitude() == 0) {
//      System.err.println("Complex:normalizePerpendicularTo - baseAxis must be non-zero.");
      return false;
    }
    
    if (magnitude() == 0) {
//      System.err.println("Complex:normalizePerpendicularTo - This must be non-zero.");
      return false;
    }
      
    Complex self = new Complex(this);
    Complex baseCopy = new Complex(baseAxis);
    baseCopy.normalize_n();
    self.normalize_n();
    self = self.orthogonal_component_to(baseCopy);
    if (self.magnitude() < 1.0e-15) {
//      System.err.println("Complex:normalizePerpendicularTo -  vectors are parallel.  That's invalid.");
      return false;
    }
    self.normalize_n();
    this.assign_from(self);
    baseAxis.assign_from(baseCopy);
    return true;
  }
    
  static public Complex rotate (double degrees, Complex base, Complex target)
  {
    double radians = degrees * Math.PI / 180.0;
    double sin;
    double cos;
    if (degrees == 90) {
      cos = 0.0;
      sin = 1.0;
    } else if (degrees == 0) {
      cos = 1.0;
      sin = 0.0;
    } else {
      cos = Math.cos(radians);
      sin = Math.sin(radians);
    }
    double newR = base.r * cos + target.r * sin;
    if (newR == -0.0) newR = 0; // Such nonsense!
    double newI = base.i * cos + target.i * sin;
    if (newI == -0.0) newI = 0;
    return new Complex(newR, newI);
  }
  
  public void print (String title)
  {
    System.out.println(title + " r=" + r + " i=" + i);
  }

  public String toString ()
  {
    return (" r=" + r + " i=" + i);
  }
  
  static double atanIndex (double pt1r, double pt1i, double pt2r, double pt2i, boolean even_p)
    // 0.984375 == 126/128 so that we range in values which when multiplied by 256
    // take a range [0, 126] and [128, 255]
  {
    double angle;
    double deltaR = Math.abs(pt2r - pt1r);
    if (deltaR == 0) {
      angle = Math.PI / 2;
    } else {
      angle = Math.atan(Math.abs(pt2i - pt1i) / deltaR);//0 to PI/2
    }
    if (angle < 0) {
      angle += Math.PI;
    }
    // angle is in the range 0 thru PI.
    if (even_p) {  
      return 0.984375 * (angle / Math.PI);  //0 thru 0.5;
    }
    return 0.984375 * (angle / Math.PI) + 0.5;  //0.5 through 1;
  }
  
  static double atanIndex (Complex pt1, Complex pt2, boolean even_p)
    // 0.984375 == 126/128 so that we range in values which when multiplied by 256
    // take a range [0, 126] and [128, 255]
  {
    double angle;
    double deltaR = Math.abs(pt2.r - pt1.r);
    if (deltaR == 0) {
      angle = Math.PI / 2;
    } else {
      angle = Math.atan(Math.abs(pt2.i - pt1.i) / deltaR);//0 to PI/2
    }
    if (angle < 0) {
      angle += Math.PI;
    }
    // angle is in the range 0 thru PI.
    if (even_p) {  
      return 0.984375 * (angle / Math.PI);  //0 thru 0.5;
    }
    return 0.984375 * (angle / Math.PI) + 0.5;  //0.5 through 1;
  }
  


}
