package JavaQuat20;

import java.math.*;

public class FastQuaternion
{
  public double r;
  public double i;
  public double j;
  public double k;
  
  public FastQuaternion(double ra, double ia, double ja, double ka)
  {
    r = ra; i = ia; j = ja; k = ka;
  }
  
  public FastQuaternion ()
  {
    r = i = j = k = 0;
  }
  
  public FastQuaternion (FastQuaternion arg)
  {
    r = arg.r;
    i = arg.i;
    j = arg.j;
    k = arg.k;
  }
  
  public void setR (double arg)
  {
    r = arg;
  }
  
  public void setI (double arg)
  {
    i = arg;
  }
  
  public void setJ (double arg)
  {
    j = arg;
  }
  
  public void setK (double arg)
  {
    k = arg;
  }
  
  public void setR (BigDecimal arg)
  {
    r = arg.doubleValue();
  }
  
  public void setI (BigDecimal arg)
  {
    i = arg.doubleValue();
  }
  
  public void setJ (BigDecimal arg)
  {
    j = arg.doubleValue();
  }
  
  public void setK (BigDecimal arg)
  {
    k = arg.doubleValue();
  }
  
  // Misc
  
  public String rToString ()
  {
    return String.valueOf(r);
  }
  
  public String iToString ()
  {
    return String.valueOf(i);
  }
  
  public String jToString ()
  {
    return String.valueOf(j);
  }
  
  public String kToString ()
  {
    return String.valueOf(k);
  }
  
  // Misc
  
  public FastQuaternion assign_from (FastQuaternion arg)
  {
    r = arg.r;
    i = arg.i;
    j = arg.j;
    k = arg.k;
    return this;
  }
  
  public boolean equal_p (FastQuaternion arg)
  {
    return (r == arg.r &&
            i == arg.i &&
            j == arg.j &&
            k == arg.k);
  }

  // Math

  public FastQuaternion plus (FastQuaternion arg)
  {
    return new FastQuaternion(r + arg.r, i + arg.i, j + arg.j, k + arg.k);
  }
  
  public FastQuaternion plus_equal (FastQuaternion arg)
  {
    r = r + arg.r;
    i = i + arg.i;
    j = j + arg.j;
    k = k + arg.k;
    return (FastQuaternion)this;
  }
  
   public FastQuaternion minus (FastQuaternion arg)
  {
    return (FastQuaternion)new FastQuaternion(r - arg.r, i - arg.i, j - arg.j, k - arg.k);
  }
  
  public FastQuaternion minus_equal (FastQuaternion arg)
  {
    r = r - arg.r;
    i = i - arg.i;
    j = j - arg.j;
    k = k - arg.k;
    return (FastQuaternion)this;
  }
  
  public FastQuaternion times (FastQuaternion point)
  {
     return (FastQuaternion)
      new FastQuaternion(r * point.r - i * point.i - j * point.j - k * point.k,
                           r * point.i + i * point.r + j * point.k - k * point.j,
                           r * point.j - i * point.k + j * point.r + k * point.i,
                           r * point.k + i * point.j - j * point.i + k * point.r);
  }
  
  public FastQuaternion times_scalar_n (double scalar)
  {
    r = r * scalar;
    i = i * scalar;
    j = j * scalar;
    k = k * scalar;
    return (FastQuaternion)this;
  }
  
  public FastQuaternion times_scalar_n (int scalar)
  {
    r = r * scalar;
    i = i * scalar;
    j = j * scalar;
    k = k * scalar;
    return (FastQuaternion)this;
  }
  
  public FastQuaternion times_scalar (double scalar)
  {
    return (FastQuaternion)new FastQuaternion(r * scalar, i * scalar, j * scalar, k * scalar);
  }
  
  public FastQuaternion times_scalar (int scalar)
  {
    return (FastQuaternion)new FastQuaternion(r * scalar, i * scalar, j * scalar, k * scalar);
  }
  
  public FastQuaternion times_equal (FastQuaternion point)
  {
    double rr, ii, jj, kk;
    rr = r * point.r - i * point.i - j * point.j - k * point.k;
    ii = r * point.i + i * point.r + j * point.k - k * point.j;
    jj = r * point.j - i * point.k + j * point.r + k * point.i;
    kk = r * point.k + i * point.j - j * point.i + k * point.r;
    r = rr;
    i = ii;
    j = jj;
    k = kk;
    return (FastQuaternion)this;
  }
  
  public BigQuaternion bigQuaternionValue ()
  {
//    return new BigQuaternion(new BigDecimal(r), new BigDecimal(i), new BigDecimal(j), new BigDecimal(k));
    return new BigQuaternion(r, i, j, k);
  }
  
  public double dot (FastQuaternion point)
  {
    return (r * point.r + i * point.i + j * point.j + k * point.k);
  }
  
  public FastQuaternion projection_onto (FastQuaternion normal_q)
  {
    return normal_q.times_scalar(this.dot(normal_q));
  }
  
  public FastQuaternion orthogonal_component_to (FastQuaternion normal_q)
   // We are a little extra careful here in case of parallel hidden by round off error.
  {
    FastQuaternion approx_answer = this.minus(this.projection_onto(normal_q));
    double answer_size = approx_answer.magnitude();
    if (answer_size < 1.0e-20)
      return new FastQuaternion();
    approx_answer.normalize_n();
  
    FastQuaternion better_normal = approx_answer.minus(approx_answer.projection_onto(normal_q));
    if (better_normal.magnitude() < 0.9)
      return new FastQuaternion(); // This is so nearly parallel to normal_q, answer 0.
    if (Math.abs(better_normal.r) < 1.0e-15)
      better_normal.r = 0;
    if (Math.abs(better_normal.i) < 1.0e-15)
      better_normal.i = 0;
    if (Math.abs(better_normal.j) < 1.0e-15)
      better_normal.j = 0;
    if (Math.abs(better_normal.k) < 1.0e-15)
      better_normal.k = 0;
    if (better_normal.magnitude() < 1.0e-15)
      return new FastQuaternion();

    return better_normal.normalize_n().times_scalar(answer_size);
  }
  
  public double magnitude ()
  {
    return Math.sqrt(r*r + i*i + j*j + k*k);
  }
  
  public double magnitude_sq ()
  {
    return (r*r + i*i + j*j + k*k);
  }
  
  public FastQuaternion normalize_n ()
  {
    double mag = magnitude();
    r = r / mag;
    i = i / mag;
    j = j / mag;
    k = k / mag;
    return this;
  }
  
  public boolean normalizePerpendicularTo (FastQuaternion baseAxis)
    // Normalizes baseAxis destructively, sets this to normalized 
    // perpendicular component of this to baseAxis.
    // Returns true iff able to complete these operations.
  {
    if (baseAxis.magnitude() == 0) {
//      System.err.println("FastQuaternion:normalizePerpendicularTo - baseAxis must be non-zero.");
      return false;
    }
    
    if (magnitude() == 0) {
//      System.err.println("FastQuaternion:normalizePerpendicularTo - This must be non-zero.");
      return false;
    }
      
    FastQuaternion self = new FastQuaternion(this);
    FastQuaternion baseCopy = new FastQuaternion(baseAxis);
    baseCopy.normalize_n();
    self.normalize_n();
    self = self.orthogonal_component_to(baseCopy);
    if (self.magnitude() < 1.0e-15) {
//      System.err.println("FastQuaternion:normalizePerpendicularTo -  vectors are parallel.  That's invalid.");
      return false;
    }
    self.normalize_n();
    this.assign_from(self);
    baseAxis.assign_from(baseCopy);
    return true;
  }
    
  static public FastQuaternion rotate (double degrees, FastQuaternion base, FastQuaternion target)
  {
    double radians = degrees * Math.PI / 180.0;
    double sin;
    double cos;
    if (degrees == 90) {
      cos = 0.0;
      sin = 1.0;
    } else if (degrees == 0) {
      cos = 1.0;
      sin = 0.0;
    } else {
      cos = Math.cos(radians);
      sin = Math.sin(radians);
    }
    double newR = base.r * cos + target.r * sin;
    if (newR == -0.0) newR = 0; // Such nonsense!
    double newI = base.i * cos + target.i * sin;
    if (newI == -0.0) newI = 0;
    double newJ = base.j * cos + target.j * sin;
    if (newJ == -0.0) newJ = 0;
    double newK = base.k * cos + target.k * sin;
    if (newK == -0.0) newK = 0;
    return new FastQuaternion(newR, newI, newJ, newK);
  }
  
  public void print (String title)
  {
    System.out.println(title + " r=" + r + " i=" + i + " j=" + j + " k=" + k);
  }

  public String toString ()
  {
    return (" r=" + r + " i=" + i + " j=" + j + " k=" + k);
  }

}
