package JavaQuat20.FractalLoops;

import JavaQuat20.*;
import java.math.*;

public class Cpx2CardTrapOnly extends JQLooper
{
  public String rString () {return "real";}
  public String iString () {return "i";}
  public String jString () {return "-/rmid/imid";}
  public String kString () {return "-/hits/size";}  

  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
   //  Exits when near a Cardioid centered at  R = coeff.j and i = coeff.k.
   //  Magnitude is trans.k and initial iteraterns required to exit is trans.j
  {
    double reduction = trans.k;
    double piOver2 = Math.PI / 2.0;
    FastQuaternion add = coeff.times(point).plus(trans);
    double z_r = point.r;
    double z_i = point.i;
    double add_r = add.r;
    double add_i = add.i;
    double real;
    double zrsq = z_r * z_r;
    double zisq = z_i * z_i;
    double realCard;
    double imagCard;
    double angle;
    double bound;
    double magnitude;
    for (int i = 1; i < depth; i++) {
      real = zrsq - zisq + add_r;
      z_i = 2 * z_r * z_i + add_i;
      z_r = real;
      zrsq = z_r * z_r;
      zisq = z_i * z_i;
//      if (zrsq + zisq > 4)
//        return i;
      if (i > trans.j) {
        realCard = (z_r - coeff.j);
        imagCard = (z_i - coeff.k);
        if (realCard == 0)
          angle = piOver2;
        else if (realCard > 0)
          angle = Math.atan(imagCard / realCard);
        else 
          angle = Math.PI - Math.atan(imagCard / realCard);
        bound = reduction * (1 + Math.cos(angle));
        magnitude = Math.sqrt(realCard * realCard + imagCard * imagCard);
        if (magnitude < bound) {
          return (int)(magnitude * depth / bound);
        }
      }
    }
    return depth;
  }
}     

