package JavaQuat20.FractalLoops;

import JavaQuat20.*;
import java.math.*;

public class Cpx2InvTrapGIF extends JQLooper
{

//  private int debugCount_ = 0;

  public int loop (FastQuaternion Q1, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Have to define
   {return 0;}
   
  public boolean acceptsTrapGif () {return true;}
  public String formulaString () {return "Z(n+1) = Z(n)^2 + C*Z(0) + T";}
  public FastQuaternion q1Init () {return new FastQuaternion(-0.75,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,6);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0);}
  public String docString () {return "This is my original Image Trap loop.\n" +
                                     "Load a GIF file for the trap.\n" +
                                     "Ret is the color map index returned if the depth is reached." +
                                     "Image is reflected around circle of radius 2 out to Bnd." +
                                     "Cclk is a degrees rotation of the image.\n" +
                                     "Size is the size of the embedded image.";}
  public String q1String () {return "Z(0)//-/ret";}
  public String q2String () {return "C//-/Bnd";}
  public String q3String () {return "T//cclk/-";}
  
//  public String rString () {return "real";}
//  public String iString () {return "i";}
//  public String jString () {return "-/rmid/cclk";}
//  public String kString () {return "bg/imid/size";}  
   
  public int trapLoop (FastQuaternion Q1, FastQuaternion Q2, FastQuaternion Q3, int depth,
                       TrapFrame trap)
   // Returns value in range [1, depth] inclusive.
   //  Exits when near a Cardioid centered at  R = Q2.j and i = Q2.k.
   //  Magnitude is Q3.k and initial iteraterns required to exit is Q3.j
  {
  
    // Set up trap.
    trap.setCenter(0, 0);
    trap.setRotation(Q3.j);
    boolean oKay = trap.setRange(1);
    if (!oKay) {
      // Return background, trap is not set up right.
      return (int)Q1.k;
    }
    double bound = Q2.k;
    double donut = bound - 2;
    
    // Set up initial loop state.
    FastQuaternion Q1Prime = new FastQuaternion(Q1);
    Q1Prime.j = 0;
    Q1Prime.k = 0;
    FastQuaternion add = Q2.times(Q1Prime).plus(Q3);
    double z_r = Q1.r;
    double z_i = Q1.i;
    double add_r = add.r;
    double add_i = add.i;
    double real;
    double zrsq = z_r * z_r;
    double zisq = z_i * z_i;

    for (int i = 1; i < depth; i++) {
      real = zrsq - zisq + add_r;
      z_i = 2 * z_r * z_i + add_i;
      z_r = real;
      zrsq = z_r * z_r;
      zisq = z_i * z_i;
      
      // See if we are diverging and if so invert catch donut to unit circle where the trap gif thinks it is.
      if (z_r * z_r + z_i * z_i > 4) {
        if (trap.hitP((bound - z_r)/donut, (bound - z_i)/donut) ) {
          return trap.hitColorMapIndex();
        }
      }
    }
    return (int)Q1.k; // Never hit it, so return background.
  }
}     

