package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class Cpx3TrapGIF extends JQLooper
{
  public int loop (FastQuaternion Q1, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Have to define
   {return 0;}
   
  public boolean acceptsTrapGif () {return true;}
  public String formulaString () {return "Z(n+1) = Z(n)^3 + C*Z(0) + T";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0.18);}
  public String docString () {return "This is my second Image Trap loop.\n" +
                                     "Load a GIF file for the trap.\n" +
                                     "Ret is the color map index returned if the depth is reached." +
                                     "R and Imid set the position of the image.\n" +
                                     "Cclk is a degrees rotation of the image.\n" +
                                     "Size is the size of the embedded image.";}
  public String q1String () {return "Z(0)//-/ret";}
  public String q2String () {return "C//Rmid/Imid";}
  public String q3String () {return "T//cclk/size";}
  

  public int trapLoop (FastQuaternion Q1, FastQuaternion Q2, FastQuaternion Q3, 
                            int depth, TrapFrame trap)
   // Returns value in range [1, depth] inclusive.
  {
  
    // Set up trap.
    trap.setCenter(Q2.j, Q2.k);
    trap.setRotation(Q3.j);
    boolean oKay = trap.setRange((Q3.k == 0) ? 1 : Q3.k);
    if (!oKay) {
      // Return background, trap is not set up right.
      return (int)Q1.k;
    }
    double realOffCenter;
    double imagOffCenter;
    double diagonal = Math.sqrt(trap.height() * trap.height() +
                                trap.width() * trap.width() );
    double realHalfRange = diagonal / 1.41;
    double imagHalfRange = realHalfRange;
    
    //Set up loop
    FastQuaternion Q1Prime = new FastQuaternion(Q1);
    Q1Prime.j = 0;
    Q1Prime.k = 0;
    FastQuaternion add = Q2.times(Q1Prime).plus(Q3);
    double z_r = Q1.r;
    double z_i = Q1.i;
    double add_r = add.r;
    double add_i = add.i;
    double real;
    double zrsq = z_r * z_r;
    double zisq = z_i * z_i;
    double magnitude;
    for (int i = 1; i < depth; i++) {
      real = zrsq * z_r - 3 * z_r * zisq + add_r;
      z_i = 3 * zrsq * z_i - z_i * zisq + add_i;
      z_r = real;
      zrsq = z_r * z_r;
      zisq = z_i * z_i;
      
      // See if we hit a non-transparent part of the trap gif.
      realOffCenter = (z_r - Q2.j);
      imagOffCenter = (z_i - Q2.k);
      if (Math.abs(realOffCenter) <= realHalfRange &&
          Math.abs(imagOffCenter) <= imagHalfRange) {
        if (trap.hitP(z_r, z_i) ) {
          return trap.hitColorMapIndex();
        }
      }
      
      // Exit if diverging.
       if (zrsq + zisq > 4)
        return (int)Q1.k;
    }
    return (int)Q1.k;
  }
}

