package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class CpxN extends JQLooper
{
  public boolean outputsColorIndexP () {return false;}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,2,0);}
  public String docString () {return "Center power A=2 at (-0.75,0), higher powers at (0,0).";}
  public String formulaString () {return "Z(n+1) = Z(n)^A + C*Z(0) + T";}
  public String q1String () {return "Z(0)//-/-";}
  public String q2String () {return "C//-/-";}
  public String q3String () {return "T//A/-";}

  public int loop (FastQuaternion point, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Returns value in range [0, 255] inclusive.  255 is reserved for non-divergence.
   //     z(n+1) = z(n)^a + c*z(n)^b + p*z(n-1)
   // Init: complex y = (0,0); complex newz = (0,0); z = #pixel
   // Loop: newz = z^@power1  +  z^@power2 * @seed  +  @induct * y;  y = z;  z = newz
  {
    Complex Z = new Complex(point.r,point.i);
    Complex C = new Complex(Q2.r, Q2.i);
    Complex add = new Complex(Q3.r, Q3.i).plus_equal(C.times_equal(Z));
    int power1 = (int)Q3.j;
  
    for (int i = 1; i < depth; i++) {
      Z.assign_from(Z.toPower(power1).plus_equal(add));
      if (Z.magnitude() > 2) {
        return i;
      }
    }
    return depth;
  }
    
  public static void main(String[] args)
    // I expect the first arg to be a pathname or URL
    // The second might be a width and the third a height
    // None of these would be required.
  {
     System.out.println("break");
  }

}

