package JavaQuat20.FractalLoops;

import JavaQuat20.*;
import java.math.*;

public class CpxNTrapGIF extends JQLooper
{

//  private int debugCount_ = 0;

  public int loop (FastQuaternion Q1, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Have to define
   {return 0;}
   
  public boolean acceptsTrapGif () {return true;}
  public String formulaString () {return "Z(n+1) = Z(n)^n + C*Z(0) + T";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,4,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0.18);}
  public String docString () {return "This is an Image Trap loop.\n" +
                                     "Load a GIF file for the trap.  Transparency is honored.\n" +
                                     "Ret is the color map index returned if the depth is reached." +
                                     "R and Imid set the position of the image.\n" +
                                     "Cclk is a degrees rotation of the image.\n" +
                                     "Size is the size of the embedded image.";}
  public String q1String () {return "Z(0)//n/ret";}
  public String q2String () {return "C//Rmid/Imid";}
  public String q3String () {return "T//cclk/size";}
   
  public int trapLoop (FastQuaternion Q1, FastQuaternion Q2, FastQuaternion Q3, int depth,
                       TrapFrame trap)
   // Returns value in range [1, depth] inclusive.
   //  Exits when near a Cardioid centered at  R = Q2.j and i = Q2.k.
   //  Magnitude is Q3.k and initial iteraterns required to exit is Q3.j
  {
  
    // Set up trap.
    trap.setCenter(Q2.j, Q2.k);
    trap.setRotation(Q3.j);
    boolean oKay = trap.setRange((Q3.k == 0) ? 0.18 : Q3.k);
    if (!oKay) {
      // Return background, trap is not set up right.
      return (int)Q1.k;
    }
    double realOffCenter;
    double imagOffCenter;
    double halfRange = Q3.k * trap.radius();
    
    // Set up initial loop state.
    Complex Z = new Complex(Q1.r,Q1.i);
    Complex C = new Complex(Q2.r, Q2.i);
    Complex add = new Complex(Q3.r, Q3.i).plus_equal(C.times_equal(Z));
    int power1 = (int)Q1.j;
  
    for (int i = 1; i < depth; i++) {
      Z.assign_from(Z.toPower(power1).plus_equal(add));
      
      // See if we hit a non-transparent part of the trap gif.
      realOffCenter = (Z.r - Q2.j);
      imagOffCenter = (Z.i - Q2.k);
      if (Math.abs(realOffCenter) <= halfRange &&
          Math.abs(imagOffCenter) <= halfRange) {
        if (trap.hitP(Z.r, Z.i) ) {
          return trap.hitColorMapIndex();
        }
      }
    }
    return (int)Q1.k;
  }
}     

