package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class HcxPower2 extends JQLooper
  // HyperComplex space.
{
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0);}
  public String docString () {return "\nHypercomplex multiplication table:\n" +
  " __________________________\n" +
  "  * ||  r |  i |  j |  k |\n" +
  "==========================\n" +
  "| r ||  r |  i |  j |  k |\n" +
  "__________________________\n" +
  "| i ||  i | -r |  k | -j |\n" +
  "__________________________\n" +
  "| j ||  j |  k | -r | -i |\n" +
  "__________________________\n" +
  "| k ||  k | -j | -i |  r | \n" +
  "__________________________";}
  public String formulaString () {return "Z(n+1) = Z(n)^2 + C*Z(0) + T - using hypercomplex space";}
  public String q1String () {return "Z(0)";}
  public String q2String () {return "C";}
  public String q3String () {return "T";}

  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    FastQuaternion add = new FastQuaternion(point);
    add.assign_from(unique_times_equal(add, coeff).plus(trans));
    FastQuaternion Q0 = new FastQuaternion(point);
    for (int i = 1; i < depth; i++) {
      if (Q0.magnitude_sq() > 10)
        return i;
      unique_sq_self(Q0).plus_equal(add);
    }
    return depth;
  }
  
  private FastQuaternion unique_times_equal (FastQuaternion self, FastQuaternion point)
  {
    double rr, ii, jj, kk;
    rr = self.r * point.r - self.i * point.i - self.j * point.j + self.k * point.k;
    ii = self.r * point.i + self.i * point.r - self.j * point.k - self.k * point.j;
    jj = self.r * point.j - self.i * point.k + self.j * point.r - self.k * point.i;
    kk = self.r * point.k - self.i * point.j - self.j * point.i + self.k * point.r;
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    return self;
  }
  
    
   private FastQuaternion unique_sq_self (FastQuaternion self)
  {
    double rr, ii, jj, kk;
    rr = self.r * self.r - self.i * self.i - self.j * self.j + self.k * self.k;
    ii = 2 * (self.r * self.i - self.j * self.k);
    jj = 2 * (self.r * self.j - self.k * self.i);
    kk = 2 * (self.r * self.k - self.i * self.j);
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    return self;
  }
   
 
  
  /*  Times table
  __________________________
  HCx*||  r |  i |  j |  k |
  ==========================
  | r ||  r |  i |  j |  k |
  __________________________
  | i ||  i | -r |  k | -j |
  __________________________
  | j ||  j |  k | -r | -i |
  __________________________
  | k ||  k | -j | -i |  r | 
  __________________________
  */
  
}

