package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class Loop22 extends JQLooper
{
  public FastQuaternion q1Init () {return new FastQuaternion(-0.75,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0);}
  public String docString () {return " Mandelbrot and Julia fractal in the complex plane";}
  public String formulaString () {return "Z(n+1) = Z(n)^2 + C*Z(0) + T";}
  public String q1String () {return "Z(0)//-/-";}
  public String q2String () {return "C//-/-";}
  public String q3String () {return "T//-/-";}

  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    FastQuaternion add = coeff.times(point).plus(trans);
    double z_r = point.r;
    double z_i = point.i;
    double add_r = add.r;
    double add_i = add.i;
    double real;
    double zrsq = z_r * z_r;
    double zisq = z_i * z_i;
    double magnitude;
    for (int i = 1; i < depth; i++) {
      real = zrsq - zisq + add_r;
      z_i = 2 * z_r * z_i + add_i;
      z_r = real;
      zrsq = z_r * z_r;
      zisq = z_i * z_i;
      if (zrsq + zisq > 4)
        return i;
    }
    return depth;
  }
}

