package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class Loop23 extends JQLooper
{
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0);}
  public String docString () {return " Mandelbrot-like fractal using 3rd power in the complex plane";}
  public String formulaString () {return "Z(n+1) = Z(n)^3 + C*Z(0) + T";}
  public String q1String () {return "Z(0)//-/-";}
  public String q2String () {return "C//-/-";}
  public String q3String () {return "T//-/-";}

  public int loop (FastQuaternion Q1, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    FastQuaternion add = Q2.times(Q1).plus(Q3);
    double z_r = Q1.r;
    double z_i = Q1.i;
    double add_r = add.r;
    double add_i = add.i;
    double real;
    double zrsq = z_r * z_r;
    double zisq = z_i * z_i;
    double magnitude;
    for (int i = 1; i < depth; i++) {
      real = zrsq * z_r - 3 * z_r * zisq + add_r;
      z_i = 3 * zrsq * z_i - z_i * zisq + add_i;
      z_r = real;
      zrsq = z_r * z_r;
      zisq = z_i * z_i;
      if (zrsq + zisq > 4)
        return i;
    }
    return depth;
  }
}

