package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class Loop322k extends JQLooper
{
  double fullk;
  double halfk;
  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    fullk = 2 + point.k;
    halfk = 2 + point.k * 0.5;
    FastQuaternion add = new FastQuaternion(point);
    add.assign_from(unique_times_equal(add, coeff).plus(trans));
    FastQuaternion Q0 = new FastQuaternion(point);
    for (int i = 1; i < depth; i++) {
      if (Q0.magnitude_sq() > 15)
        return i;
      unique_sq_self(Q0).plus_equal(add);
    }
    return depth;
  }
  
  private FastQuaternion unique_times_equal (FastQuaternion self, FastQuaternion point)
  {
    double rr, ii, jj;
    rr =  self.r * point.r + self.i * point.j - self.j * point.i;
    ii =  self.r * point.i * halfk + 
          self.i * point.r * halfk - self.j * point.j;
    jj =  self.r * point.j * halfk - self.i * point.i +
          self.j * point.r * halfk;
    self.r = rr;
    self.i = ii;
    self.j = jj;
    return self;
  }
  
  private FastQuaternion unique_sq_self (FastQuaternion self)
  {
    double rr, ii, jj, kk;
    rr = self.r * self.r;
    ii = fullk * self.r * self.i - self.j * self.j;
    jj = fullk * self.r * self.j - self.i * self.i;
    self.r = rr;
    self.i = ii;
    self.j = jj;
    return self;
  }

  
  /*  Times table
  ___________________________
 322k*||   r  |   i  |   j  |
  ===========================
  | r ||   r  | Ki/2 | Kj/2 |
  ___________________________
  | i || Ki/2 |  -j  |   r  |
  ___________________________
  | j || Kj/2 |  -r  |  -i  |
  ___________________________

 */
  
}

