package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class LoopTetraPiHCi extends JQLooper
{
  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    FastQuaternion add = new FastQuaternion(point);
    add.assign_from(unique_times_equal(add, coeff).plus(trans));
    FastQuaternion Q0 = new FastQuaternion(point);
    for (int i = 1; i < depth; i++) {
      if (Q0.magnitude_sq() > 10)
        return i;
      unique_sq_self(Q0).plus_equal(add);
    }
    return depth;
  }
  
  private FastQuaternion unique_times_equal (FastQuaternion self, FastQuaternion point)
  {
    double rr, ii, jj, kk;
    rr = self.r * point.r + self.i * point.k - self.j * point.j - self.k * point.i;
    ii = self.r * point.i + self.i * point.r - self.j * point.k - self.k * point.j;
    jj = self.r * point.j + self.i * point.i + self.j * point.r - self.k * point.k;
    kk = self.r * point.k + self.i * point.j - self.j * point.i + self.k * point.r;
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    return self;
  }
  
   private FastQuaternion unique_sq_self (FastQuaternion self)
  {
    double rr, ii, jj, kk;
    rr = self.r * self.r - self.j * self.j;
    ii = 2 * self.r * self.i - 2 * self.j * self.k;
    jj = 2 * self.r * self.j + self.i * self.i - self.k * self.k;
    kk = 2 * self.r * self.k;
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    return self;
  }
  
 
  
  /*  Times table
       __________________________
TetraPiHCi*||  r |  i |  j |  k |
       ==========================
       | r ||  r |  i |  j |  k |
       __________________________
       | i ||  i |  j |  k |  r |
       __________________________
       | j ||  j | -k | -r | -i |
       __________________________
       | k ||  k | -r | -i | -j | 
       __________________________
  */
  
}

