package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class Phoenix extends JQLooper
{
  public boolean outputsColorIndexP () {return true;}
  public String formulaString () {return "Z(n+1) = Z(n)^A + C*Z(n)^B + P*Z(n-1)";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(0.5,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0.653,0,2,0);}
  public String docString () {return "This is, I believe, the Phoenix formula in the complex plane.";}
  public String q1String () {return "Z(0)//-/-";}
  public String q2String () {return "P//-/-";}
  public String q3String () {return "C//A/B";}
  

  public int loop (FastQuaternion point, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Returns value in range [0, 255] inclusive.  255 is reserved for non-divergence.
   //     z(n+1) = z(n)^a + c*z(n)^b + p*z(n-1)
   // Init: complex y = (0,0); complex newz = (0,0); z = #pixel
   // Loop: newz = z^@power1  +  z^@power2 * @seed  +  @induct * y;  y = z;  z = newz
  {
    Complex Z = new Complex(point.r,point.i);
    Complex induct = new Complex(Q2.r, Q2.i);
    Complex C = new Complex(Q3.r, Q3.i);
    int power1 = (int)Q3.j;
    int power2 = (int)Q3.k;
    Complex newZ;
    Complex Y = new Complex(0, 0);
  
    for (int i = 1; i < depth; i++) {
      newZ = Z.toPower(power1).plus_equal(Z.toPower(power2).times_equal(C)).plus_equal(Y.times_equal(induct));
      Y = Z;
      Z = newZ;
      if (Z.magnitude() > 2) {
        return (int)(256 * atanIndex(Y.r, Y.i, Z.r, Z.i, (i%2==0)));
      }
    }
    return 255;
  }
  
  static double atanIndex (double pt1r, double pt1i, double pt2r, double pt2i, boolean even_p)
    // 0.984375 == 126/128 so that we range in values which when multiplied by 256
    // take a range [0, 126] and [128, 255]
  {
    double angle;
    double deltaR = Math.abs(pt2r - pt1r);
    if (deltaR == 0) {
      angle = Math.PI / 2;
    } else {
      angle = Math.atan(Math.abs(pt2i - pt1i) / deltaR);//0 to PI/2
    }
    if (angle < 0) {
      angle += Math.PI;
    }
    // angle is in the range 0 thru PI.
    if (even_p) {  
      return 0.984375 * (angle / Math.PI);  //0 thru 0.5;
    }
    return 0.984375 * (angle / Math.PI) + 0.5;  //0.5 through 1;
  }
  
  public static void main(String[] args)
    // I expect the first arg to be a pathname or URL
    // The second might be a width and the third a height
    // None of these would be required.
  {
     System.out.println("for (0, 0) - (0, 10) atanIndex = " + atanIndex(0, 0, 0, 10,true)
                        + "index = " + (256 * atanIndex(0, 0, 0, 10,true)) );
     System.out.println("for (0, 0) - (8.6, 5) atanIndex = " + atanIndex(0, 0, 8.6, 5,true));
     System.out.println("for (0, 0) - (5, 5) atanIndex = " + atanIndex(0, 0, 5, 5,true));
     System.out.println("for (0, 0) - (5, 8.6) atanIndex = " + atanIndex(0, 0, 5, 8.6,true));
     System.out.println("for (0, 0) - (10, 0) atanIndex = " + atanIndex(0, 0, 10, 0,true)
                        + "index = " + (256 * atanIndex(0, 0, 10, 0,true)) );
     System.out.println("for (0, 0) - (-10, 0) atanIndex = " + atanIndex(0, 0, -10, 0,true));
     System.out.println("for (0, 0) - (5, -8.6) atanIndex = " + atanIndex(0, 0, 5, -8.6,true));
     System.out.println("for (0, 0) - (-5, 5) atanIndex = " + atanIndex(0, 0, -5, 5,true));
     System.out.println("for (0, 0) - (8.6, -5) atanIndex = " + atanIndex(0, 0, 8.6, -5,true));
     System.out.println("for (0, 0) - (0, -10) atanIndex = " + atanIndex(0, 0, 0, -10,true));
System.out.println("break");
     System.out.println("for (0, 0) - (0, 10) atanIndex = " + atanIndex(0, 0, 0, 10,false)
                         + "index = " + (256 * atanIndex(0, 0, 0, 10,false)) );
     System.out.println("for (0, 0) - (8.6, 5) atanIndex = " + atanIndex(0, 0, 8.6, 5,false));
     System.out.println("for (0, 0) - (5, 5) atanIndex = " + atanIndex(0, 0, 5, 5,false));
     System.out.println("for (0, 0) - (5, 8.6) atanIndex = " + atanIndex(0, 0, 5, 8.6,false));
     System.out.println("for (0, 0) - (10, 0) atanIndex = " + atanIndex(0, 0, 10, 0,false)
                        + "index = " + (256 * atanIndex(0, 0, 10, 0,false)) );
     System.out.println("for (0, 0) - (-10, 0) atanIndex = " + atanIndex(0, 0, -10, 0,false));
     System.out.println("for (0, 0) - (5, -8.6) atanIndex = " + atanIndex(0, 0, 5, -8.6,false));
     System.out.println("for (0, 0) - (-5, 5) atanIndex = " + atanIndex(0, 0, -5, 5,false));
     System.out.println("for (0, 0) - (8.6, -5) atanIndex = " + atanIndex(0, 0, 8.6, -5,false));
     System.out.println("for (0, 0) - (0, -10) atanIndex = " + atanIndex(0, 0, 0, -10,false));


  }

}

