package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class PhoenixDepth extends JQLooper
{
  public boolean outputsColorIndexP () {return false;}
  public String formulaString () {return "Z(n+1) = Z(n)^A + C*Z(n)^B + P*Z(n-1); K is bound.";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(0.5,0,2,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0.653,0,2,0);}
  public String docString () {return "This is, I believe, the Phoenix formula in the complex plane.\n" +
                                     "This one returns the number of iterations till divergence\n" +
                                     "at |Z(n)| > K.";}
  public String q1String () {return "Z(0)//-/-";}
  public String q2String () {return "P//K/-";}
  public String q3String () {return "C//A/B";}
  

  public int loop (FastQuaternion point, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Returns value in range [0, 255] inclusive.  255 is reserved for non-divergence.
   //     z(n+1) = z(n)^a + c*z(n)^b + p*z(n-1)
   // Init: complex y = (0,0); complex newz = (0,0); z = #pixel
   // Loop: newz = z^@power1  +  z^@power2 * @seed  +  @induct * y;  y = z;  z = newz
  {
    Complex Z = new Complex(point.r,point.i);
    Complex induct = new Complex(Q2.r, Q2.i);
    Complex C = new Complex(Q3.r, Q3.i);
    int power1 = (int)Q3.j;
    int power2 = (int)Q3.k;
    int bound = (int)Q2.j;
    Complex newZ;
    Complex Y = new Complex(0, 0);
  
    for (int i = 1; i < depth; i++) {
      newZ = Z.toPower(power1).plus_equal(Z.toPower(power2).times_equal(C)).plus_equal(Y.times_equal(induct));
      Y = Z;
      Z = newZ;
      if (Z.magnitude() > bound) {
        return i;
      }
    }
    return depth;
  }
  
}

