package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class PhoenixTrapGIF extends JQLooper
{
  public boolean acceptsTrapGif () {return true;}
  public boolean outputsColorIndexP () {return true;}
  public String formulaString () {return "Z(n+1) = Z(n)^A + C*Z(n)^B + P*Z(n-1)";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,2,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(-0.5,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0.653,0,0,0.18);}

  public String docString () {return "This is an Image Trap loop.\n" +
                                     "Load a GIF file for the trap.  Transparency is honored.\n" +
                                     "R and Imid set the position of the image.\n" +
                                     "Cclk is a degrees rotation of the image.\n" +
                                     "Size is the size of the embedded image.";}
  public String q1String () {return "Z(0)//A/B";}
  public String q2String () {return "P//Rmid/Imid";}
  public String q3String () {return "C//cclk/size";}


  public int loop (FastQuaternion point, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Have to define
   {return 0;}

  public int trapLoop (FastQuaternion Q1, FastQuaternion Q2, FastQuaternion Q3, int depth,
                       TrapFrame trap)
   // Returns value in range [0, 255] inclusive.
   //  0 is used for no GIF hit.
  {
  
    // Set up trap.
    trap.setCenter(Q2.j, Q2.k);
    trap.setRotation(Q3.j);
    boolean oKay = trap.setRange((Q3.k == 0) ? 0.18 : Q3.k);
    if (!oKay) {
      // Return background, trap is not set up right.
      return 0;
    }
    double realOffCenter;
    double imagOffCenter;
    double halfRange = Q3.k * trap.radius();
    
    // Set up initial loop state.
    Complex Z = new Complex(Q1.r,Q1.i);
    Complex induct = new Complex(Q2.r, Q2.i);
    Complex C = new Complex(Q3.r, Q3.i);
    int power1 = (int)Q1.j;
    int power2 = (int)Q1.k;
    Complex newZ;
    Complex Y = new Complex(0, 0);
  
    for (int i = 1; i < depth; i++) {
      // See if we hit a non-transparent part of the trap gif.
      realOffCenter = (Z.r - Q2.j);
      imagOffCenter = (Z.i - Q2.k);
      if (Math.abs(realOffCenter) <= halfRange &&
          Math.abs(imagOffCenter) <= halfRange) {
        if (trap.hitP(Z.r, Z.i) ) {
          return trap.hitColorMapIndex();
        }
      }
      
      newZ = Z.toPower(power1).plus_equal(Z.toPower(power2).times_equal(C)).plus_equal(Y.times_equal(induct));
      Y = Z;
      Z = newZ;
    }
    return 0;
  }
}

