package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class Phoenix_21 extends JQLooper
  // This is a specialized fast version of the Phoenix where A=2 and B=1;
{
  public boolean outputsColorIndexP () {return true;}
  public String formulaString () {return "Z(n+1) = Z(n)^2 + C*Z(n) + P*Z(n-1)";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(0.5,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0.653,0,2,0);}
  public String docString () {return "This is, I believe, the Phoenix formula in the complex plane.\n" +
                                     "Specialized to the case where A=2 and B=1";}
  public String q1String () {return "Z(0)//-/-";}
  public String q2String () {return "P//-/-";}
  public String q3String () {return "C//-/-";}
  

  public int loop (FastQuaternion point, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Returns value in range [0, 255] inclusive.  255 is reserved for non-divergence.
   //     z(n+1) = z(n)^a + c*z(n)^b + p*z(n-1)
   // Init: complex y = (0,0); complex newz = (0,0); z = #pixel
   // Loop: newz = z^@power1  +  z^@power2 * @seed  +  @induct * y;  y = z;  z = newz
  {
    double r = point.r;
    double i = point.i;
    double prev_r = 0;
    double prev_i = 0;
    double temp_r, temp_i;
    
    for (int count = 1; count < depth; count++) {
      temp_r = r * r - i * i + Q3.r * r - Q3.i * i + Q2.r * prev_r - Q2.i * prev_i;
      temp_i = 2 * r * i + Q3.r * i + Q3.i * r + Q2.r * prev_i + Q2.i * prev_r;
      prev_r = r;
      prev_i = i;
      r = temp_r;
      i = temp_i;
      if (r*r + i*i > 2) {
        return (int)(256 * atanIndex(prev_r, prev_i, r, i, (count%2==0)));
      }
    }
    return 255;
  }
  
  static double atanIndex (double pt1r, double pt1i, double pt2r, double pt2i, boolean even_p)
    // 0.984375 == 126/128 so that we range in values which when multiplied by 256
    // take a range [0, 126] and [128, 255]
  {
    double angle;
    double deltaR = Math.abs(pt2r - pt1r);
    if (deltaR == 0) {
      angle = Math.PI / 2;
    } else {
      angle = Math.atan(Math.abs(pt2i - pt1i) / deltaR);//0 to PI/2
    }
    if (angle < 0) {
      angle += Math.PI;
    }
    // angle is in the range 0 thru PI.
    if (even_p) {  
      return 0.984375 * (angle / Math.PI);  //0 thru 0.5;
    }
    return 0.984375 * (angle / Math.PI) + 0.5;  //0.5 through 1;
  }
  
  public static void main(String[] args)
    // I expect the first arg to be a pathname or URL
    // The second might be a width and the third a height
    // None of these would be required.
  {
     System.out.println("for (0, 0) - (0, 10) atanIndex = " + atanIndex(0, 0, 0, 10,true)
                        + "index = " + (256 * atanIndex(0, 0, 0, 10,true)) );
     System.out.println("for (0, 0) - (8.6, 5) atanIndex = " + atanIndex(0, 0, 8.6, 5,true));
     System.out.println("for (0, 0) - (5, 5) atanIndex = " + atanIndex(0, 0, 5, 5,true));
     System.out.println("for (0, 0) - (5, 8.6) atanIndex = " + atanIndex(0, 0, 5, 8.6,true));
     System.out.println("for (0, 0) - (10, 0) atanIndex = " + atanIndex(0, 0, 10, 0,true)
                        + "index = " + (256 * atanIndex(0, 0, 10, 0,true)) );
     System.out.println("for (0, 0) - (-10, 0) atanIndex = " + atanIndex(0, 0, -10, 0,true));
     System.out.println("for (0, 0) - (5, -8.6) atanIndex = " + atanIndex(0, 0, 5, -8.6,true));
     System.out.println("for (0, 0) - (-5, 5) atanIndex = " + atanIndex(0, 0, -5, 5,true));
     System.out.println("for (0, 0) - (8.6, -5) atanIndex = " + atanIndex(0, 0, 8.6, -5,true));
     System.out.println("for (0, 0) - (0, -10) atanIndex = " + atanIndex(0, 0, 0, -10,true));
System.out.println("break");
     System.out.println("for (0, 0) - (0, 10) atanIndex = " + atanIndex(0, 0, 0, 10,false)
                         + "index = " + (256 * atanIndex(0, 0, 0, 10,false)) );
     System.out.println("for (0, 0) - (8.6, 5) atanIndex = " + atanIndex(0, 0, 8.6, 5,false));
     System.out.println("for (0, 0) - (5, 5) atanIndex = " + atanIndex(0, 0, 5, 5,false));
     System.out.println("for (0, 0) - (5, 8.6) atanIndex = " + atanIndex(0, 0, 5, 8.6,false));
     System.out.println("for (0, 0) - (10, 0) atanIndex = " + atanIndex(0, 0, 10, 0,false)
                        + "index = " + (256 * atanIndex(0, 0, 10, 0,false)) );
     System.out.println("for (0, 0) - (-10, 0) atanIndex = " + atanIndex(0, 0, -10, 0,false));
     System.out.println("for (0, 0) - (5, -8.6) atanIndex = " + atanIndex(0, 0, 5, -8.6,false));
     System.out.println("for (0, 0) - (-5, 5) atanIndex = " + atanIndex(0, 0, -5, 5,false));
     System.out.println("for (0, 0) - (8.6, -5) atanIndex = " + atanIndex(0, 0, 8.6, -5,false));
     System.out.println("for (0, 0) - (0, -10) atanIndex = " + atanIndex(0, 0, 0, -10,false));


  }

}

