package JavaQuat20.FractalLoops;

import JavaQuat20.*;
import java.math.*;

public abstract class Tetra extends JQLooper
{
    BigDecimal TWO = null;
  Tetra ()
  {
    try {
      TWO = new BigDecimal("2.0");
    } catch (Throwable e) {
    }
  }
  
  /*  Tetra Times table
  __________________________
  428*||  r |  i |  j |  k |
  ==========================
  | r ||  r |  i |  j |  k |
  __________________________
  | i ||  i | -j |  k |  r |
  __________________________
  | j ||  j | -k | -r |  i |
  __________________________
  | k ||  k | -r | -i | -j | 
  __________________________
  */
  
  public FastQuaternion unique_times_equal (FastQuaternion self, FastQuaternion point)
  {
    double rr, ii, jj;
    rr = self.r * point.r + self.i * point.k - self.j * point.j - self.k * point.i;
    ii = self.r * point.i + self.i * point.r + self.j * point.k - self.k * point.j;
    jj = self.r * point.j - self.i * point.i + self.j * point.r - self.k * point.k;
    self.k = self.r * point.k + self.i * point.j - self.j * point.i + self.k * point.r;
    self.r = rr;
    self.i = ii;
    self.j = jj;
    return self;
  }
  
  public FastQuaternion unique_sq_self (FastQuaternion self)
  {
    double rr, ii;
    rr = self.r * self.r - self.j * self.j;
    ii = 2 * self.r * self.i;
    self.j = 2 * self.r * self.j - self.i * self.i - self.k * self.k;
    self.k = 2 * self.r * self.k;
    self.r = rr;
    self.i = ii;
    return self;
  }
  
/*
  public Quaternion unique_times_equal (Quaternion self, Quaternion point)
  {
    BigDecimal rr, ii, jj, kk;
    int scale = Math.max(self.scale(), point.scale());
    rr = self.r.multiply(point.r).add(self.i.multiply(point.k)).subtract(self.j.multiply(point.j)).subtract(self.k.multiply(point.i));
    ii = self.r.multiply(point.i).add(self.i.multiply(point.r)).add(self.j.multiply(point.k)).subtract(self.k.multiply(point.j));
    jj = self.r.multiply(point.j).subtract(self.i.multiply(point.i)).add(self.j.multiply(point.r)).subtract(self.k.multiply(point.k));
    kk = self.r.multiply(point.k).add(self.i.multiply(point.j)).subtract(self.j.multiply(point.i)).add(self.k.multiply(point.r));
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    self.setScale(scale);
    return self;
  }
  
  public Quaternion unique_sq_self (Quaternion self)
  {
    BigDecimal rr, ii, jj, kk;
    int scale = self.scale();
    rr = self.r.multiply(self.r).subtract(self.j.multiply(self.j));
    ii = TWO.multiply(self.r).multiply(self.i);
    jj = TWO.multiply(self.r).multiply(self.j).subtract(self.i.multiply(self.i)).subtract(self.k.multiply(self.k));
    kk = TWO.multiply(self.r).multiply(self.k);
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    self.setScale(scale);
    return self;
  }
  
  public BBinaryQuaternion unique_times_equal (BBinaryQuaternion self, BBinaryQuaternion point)
  {
    BigBinary rr, ii, jj, kk;
    rr = self.r.multiply(point.r).add(self.i.multiply(point.k)).subtract(self.j.multiply(point.j)).subtract(self.k.multiply(point.i));
    ii = self.r.multiply(point.i).add(self.i.multiply(point.r)).add(self.j.multiply(point.k)).subtract(self.k.multiply(point.j));
    jj = self.r.multiply(point.j).subtract(self.i.multiply(point.i)).add(self.j.multiply(point.r)).subtract(self.k.multiply(point.k));
    kk = self.r.multiply(point.k).add(self.i.multiply(point.j)).subtract(self.j.multiply(point.i)).add(self.k.multiply(point.r));
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    return self;
  }
  
  public BBinaryQuaternion unique_times_equalUPN (BBinaryQuaternion self, BBinaryQuaternion point)
  {
    //rr = self.r.multiply(point.r).add(self.i.multiply(point.k)).subtract(self.j.multiply(point.j)).subtract(self.k.multiply(point.i));
    BigBinary rrf = new BigBinary(self.r);
    BigBinary ii = new BigBinary(self.i);
    BigBinary jj = new BigBinary(self.j);
    BigBinary kk = new BigBinary(self.k);
    rrf.multiplyUPN(point.r);
    ii.multiplyUPN(point.k);
    jj.multiplyUPN(point.j);
    kk.multiplyUPN(point.i);
    rrf.addUPN(ii); rrf.subtractUPN(jj); rrf.subtractUPN(kk);

    //ii = self.r.multiply(point.i).add(self.i.multiply(point.r)).add(self.j.multiply(point.k)).subtract(self.k.multiply(point.j));
    BigBinary rr = new BigBinary(self.r);
    BigBinary iif = new BigBinary(self.i);
    jj = new BigBinary(self.j);
    kk = new BigBinary(self.k);
    rr.multiplyUPN(point.i);
    iif.multiplyUPN(point.r);
    jj.multiplyUPN(point.k);
    kk.multiplyUPN(point.j);
    iif.addUPN(rr); iif.addUPN(jj); iif.subtractUPN(kk);
    
    //jj = self.r.multiply(point.j).subtract(self.i.multiply(point.i)).add(self.j.multiply(point.r)).subtract(self.k.multiply(point.k));
    rr = new BigBinary(self.r);
    ii = new BigBinary(self.i);
    BigBinary jjf = new BigBinary(self.j);
    kk = new BigBinary(self.k);
    rr.multiplyUPN(point.j);
    ii.multiplyUPN(point.i);
    jjf.multiplyUPN(point.r);
    kk.multiplyUPN(point.k);
    jjf.subtractUPN(rr); jjf.subtractUPN(ii); jjf.subtractUPN(kk);
    
    //kk = self.r.multiply(point.k).add(self.i.multiply(point.j)).subtract(self.j.multiply(point.i)).add(self.k.multiply(point.r));
    self.r.multiplyUPN(point.k);
    self.i.multiplyUPN(point.j);
    self.j.multiplyUPN(point.i);
    self.k.multiplyUPN(point.r);
    self.k.addUPN(self.r); self.k.addUPN(self.i); self.k.subtractUPN(self.j);
    
    self.r = rrf;
    self.i = iif;
    self.j = jjf;
    return self;
  }
  
  public BBinaryQuaternion unique_sq_self (BBinaryQuaternion self)
  {
    BigBinary rr, ii, jj, kk;
    rr = self.r.multiply(self.r).subtract(self.j.multiply(self.j));
    ii = self.r.multiply(self.i).times2();
    jj = self.r.multiply(self.j).times2().subtract(self.i.multiply(self.i)).subtract(self.k.multiply(self.k));
    kk = self.r.multiply(self.k).times2();
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    return self;
  }
  
  public BBinaryQuaternion unique_sq_selfUPN (BBinaryQuaternion self)
    // This is a Looper4d derivative, not a Quaternion derivative.
  {
    BigBinary rr = new BigBinary(self.r);
    BigBinary jj = new BigBinary(self.j);
    rr.multiplyUPN(self.r);
    jj.multiplyUPN(self.j);
    rr.subtractUPN(jj);
    
    jj = new BigBinary(self.j);
    BigBinary ii = new BigBinary(self.i);
    BigBinary kk = new BigBinary(self.k);
    jj.multiplyUPN(self.r);
    jj.times2UPN();
    ii.multiplyUPN(self.i);
    kk.multiplyUPN(self.k);
    jj.subtractUPN(ii);
    jj.subtractUPN(kk);
    
    self.i.multiplyUPN(self.r);
    self.i.times2UPN();

    self.k.multiplyUPN(self.r);
    self.k.times2UPN();
    
    self.r = rr;
    self.j = jj;
    return self;
  }
  
  */
  
   /*  Times table
  __________________________
  428*||  r |  i |  j |  k |
  ==========================
  | r ||  r |  i |  j |  k |
  __________________________
  | i ||  i | -j |  k |  r |
  __________________________
  | j ||  j | -k | -r |  i |
  __________________________
  | k ||  k | -r | -i | -j | 
  __________________________
  */

}
