package JavaQuat20.FractalLoops;

import JavaQuat20.*;
import java.math.*;

public class Tetra2CardioidTrap extends Tetra
{
  static final double reduction = 0.01;
  static final double piOver2 = Math.PI / 2;
  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    FastQuaternion add = new FastQuaternion(point);
    add.assign_from(unique_times_equal(add, coeff).plus(trans));
    FastQuaternion Q0 = new FastQuaternion(point);
    FastQuaternion offset = new FastQuaternion(-0.15, -0.15, 0, 0);
    FastQuaternion QOffset = new FastQuaternion();
    for (int i = 1; i < depth; i++) {
      QOffset.assign_from(Q0).minus_equal(offset);
      double unrealMag = QOffset.i * QOffset.i + QOffset.j * QOffset.j + QOffset.k * QOffset.k;
      double denominator = Math.sqrt(unrealMag);
      double numerator = QOffset.r;
      double angle;
      if (numerator == 0)
        angle = piOver2;
      else if (numerator > 0)
        angle = Math.atan(unrealMag / numerator);
      else 
        angle = Math.PI - Math.atan(unrealMag / numerator);
      double bound = reduction * (1 + Math.cos(angle));
      double magnitude = unrealMag + numerator * numerator;
      if (magnitude < bound) {
        return (int)(magnitude * depth / bound);
      }
      if (Q0.magnitude() > 10)
        return depth;
      unique_sq_self(Q0).plus_equal(add);
    }
    return depth;
  }
     
  /*  Times table
  __________________________
  428*||  r |  i |  j |  k |
  ==========================
  | r ||  r |  i |  j |  k |
  __________________________
  | i ||  i | -j |  k |  r |
  __________________________
  | j ||  j | -k | -r |  i |
  __________________________
  | k ||  k | -r | -i | -j | 
  __________________________
  */
  
}

