package JavaQuat20.FractalLoops;

import JavaQuat20.*;
import java.math.*;

public class TetraPower2 extends Tetra
  // this is the same as LoopTetra2, with a different name.
{
  public String formulaString () {return "Z(n+1) = Z(n)^2 + C*Z(0) + T - using Tetra mult table.";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0);}
  public String docString () {return " __________________________\n" +
  "  * ||  r |  i |  j |  k |\n" +
  "==========================\n" +
  "| r ||  r |  i |  j |  k |\n" +
  "__________________________\n" +
  "| i ||  i | -j |  k |  r |\n" +
  "__________________________\n" +
  "| j ||  j | -k | -r |  i |\n" +
  "__________________________\n" +
  "| k ||  k | -r | -i | -j | \n" +
  "__________________________";}
  public String q1String () {return "Z(0)";}
  public String q2String () {return "C";}
  public String q3String () {return "T";}

  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    FastQuaternion add = new FastQuaternion(point);
    add.assign_from(unique_times_equal(add, coeff).plus(trans));
    FastQuaternion Q0 = new FastQuaternion(point);
    for (int i = 1; i < depth; i++) {
      if (Q0.magnitude_sq() > 10)
        return i;
      unique_sq_self(Q0).plus_equal(add);
    }
    return depth;
  }
     
  /*  Times table
  __________________________
  428*||  r |  i |  j |  k |
  ==========================
  | r ||  r |  i |  j |  k |
  __________________________
  | i ||  i | -j |  k |  r |
  __________________________
  | j ||  j | -k | -r |  i |
  __________________________
  | k ||  k | -r | -i | -j | 
  __________________________
  */
  
}

