package JavaQuat20.FractalLoops;

import JavaQuat20.*;
import java.math.*;

public class TetraPower3 extends Tetra
  // this is the same as LoopTetra2, with a different name.
{
  public String formulaString () {return "Z(n+1) = Z(n)^3 + C*Z(0) + T - using Tetra mult table.";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0);}
  public String docString () {return " __________________________\n" +
  "  * ||  r |  i |  j |  k |\n" +
  "==========================\n" +
  "| r ||  r |  i |  j |  k |\n" +
  "__________________________\n" +
  "| i ||  i | -j |  k |  r |\n" +
  "__________________________\n" +
  "| j ||  j | -k | -r |  i |\n" +
  "__________________________\n" +
  "| k ||  k | -r | -i | -j |\n" +
  "__________________________\n" +
  "Z^3 here is Z(Z^2).  For (Z^2)Z interchange i and k.";}
  public String q1String () {return "Z(0)";}
  public String q2String () {return "C";}
  public String q3String () {return "T";}

  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    double r = point.r; double i = point.i; double j = point.j; double k = point.k;
    double add_r = r * coeff.r + i * coeff.k - j * coeff.j - k * coeff.i + trans.r;
    double add_i = r * coeff.i + i * coeff.r + j * coeff.k - k * coeff.j + trans.i;
    double add_j = r * coeff.j - i * coeff.i + j * coeff.r - k * coeff.k + trans.j;
    double add_k = r * coeff.k + i * coeff.j - j * coeff.i + k * coeff.r + trans.k;
    double rr, ii;
    double rsq, jsq, isq_ksq, threersqLESSjsq;
    for (int count = 1; count < depth; count++) {
      rsq = r * r;
      jsq = j * j;
      isq_ksq = i * i + k * k;
      if (rsq + jsq + isq_ksq > 10)
        return count;
      rr = r * (rsq - 3 * jsq) + j * isq_ksq + add_r;
      threersqLESSjsq = 3 * rsq - jsq;
      ii = i * threersqLESSjsq + k * isq_ksq + add_i;
      j = j * threersqLESSjsq - 3 * r * isq_ksq + add_j;
      k = k * threersqLESSjsq - i * isq_ksq + add_k;
      r = rr;
      i = ii;
    }
    return depth;
  }

     
  /*  Times table
  __________________________
  428*||  r |  i |  j |  k |
  ==========================
  | r ||  r |  i |  j |  k |
  __________________________
  | i ||  i | -j |  k |  r |
  __________________________
  | j ||  j | -k | -r |  i |
  __________________________
  | k ||  k | -r | -i | -j | 
  __________________________
  */
  
}

