package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class PhoenixPaulC extends JQLooper
{
  public boolean outputsColorIndexP () {return true;}
   
  public String q1String () {return "Z(0)r/Z(0)i/-/-";}
  public String q2String () {return "-/-/-/-";}
  public String q3String () {return "Cr/Ci/-/-";}
  
  public String formulaString () {return "Phoenix-like formula.  See Help";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0.653,0,0,0);}
  public String docString () {return "This is, I believe, Paul Chandler's Phoenix.\n" +
                                     "X = Zr^2 - 0.5*Y + C;\n" +
                                     "Y = Z;\n" +
                                     "Z = X;\n" +
                                     "if (|Z| > 2) return atanIndex(Y,Z,loopCountEvenP);"
                              ;}



  public int loop (FastQuaternion point, FastQuaternion Q2, FastQuaternion Q3, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    double Zr = point.r;
    double Zi = point.i;
    double Cr = Q3.r;
    double Ci = Q3.i;
    double Yr = 0;
    double Yi = 0;
    double Xr = 0;
    double Xi = 0;
    for (int i = 1; i < depth; i++) {
      //Zr = Zr*Zr - Zi*Zi - 0.5*Zr + Cr;
      //Zi = 2*Zr*Zi - 0.5*Zi + Ci;
      Xr = Zr*Zr - Zi*Zi - 0.5*Yr + Cr;
      Xi = 2*Zr*Zi - 0.5*Yi + Ci;
      Yr = Zr;
      Yi = Zi;
      Zr = Xr;
      Zi = Xi;
      if (Zr*Zr + Zi*Zi > 2) {
        return (int)(256 * atanIndex(Yr, Yi, Zr, Zi, (i%2==0)));
      }
    }
    return 255;
  }
  
  static double atanIndex (double pt1r, double pt1i, double pt2r, double pt2i, boolean even_p)
    // 0.984375 == 126/128 so that we range in values which when multiplied by 256
    // take a range [0, 126] and [128, 255]
  {
    double angle;
    double deltaR = Math.abs(pt2r - pt1r);
    if (deltaR == 0) {
      angle = Math.PI / 2;
    } else {
      angle = Math.atan(Math.abs(pt2i - pt1i) / deltaR);//0 to PI/2
    }
    if (angle < 0) {
      angle += Math.PI;
    }
    // angle is in the range 0 thru PI.
    if (even_p) {  
      return 0.984375 * (angle / Math.PI);  //0 thru 0.5;
    }
    return 0.984375 * (angle / Math.PI) + 0.5;  //0.5 through 1;
  }
  
  public static void main(String[] args)
    // I expect the first arg to be a pathname or URL
    // The second might be a width and the third a height
    // None of these would be required.
  {
     System.out.println("for (0, 0) - (0, 10) atanIndex = " + atanIndex(0, 0, 0, 10,true)
                        + "index = " + (256 * atanIndex(0, 0, 0, 10,true)) );
     System.out.println("for (0, 0) - (8.6, 5) atanIndex = " + atanIndex(0, 0, 8.6, 5,true));
     System.out.println("for (0, 0) - (5, 5) atanIndex = " + atanIndex(0, 0, 5, 5,true));
     System.out.println("for (0, 0) - (5, 8.6) atanIndex = " + atanIndex(0, 0, 5, 8.6,true));
     System.out.println("for (0, 0) - (10, 0) atanIndex = " + atanIndex(0, 0, 10, 0,true)
                        + "index = " + (256 * atanIndex(0, 0, 10, 0,true)) );
     System.out.println("for (0, 0) - (-10, 0) atanIndex = " + atanIndex(0, 0, -10, 0,true));
     System.out.println("for (0, 0) - (5, -8.6) atanIndex = " + atanIndex(0, 0, 5, -8.6,true));
     System.out.println("for (0, 0) - (-5, 5) atanIndex = " + atanIndex(0, 0, -5, 5,true));
     System.out.println("for (0, 0) - (8.6, -5) atanIndex = " + atanIndex(0, 0, 8.6, -5,true));
     System.out.println("for (0, 0) - (0, -10) atanIndex = " + atanIndex(0, 0, 0, -10,true));
System.out.println("break");
     System.out.println("for (0, 0) - (0, 10) atanIndex = " + atanIndex(0, 0, 0, 10,false)
                         + "index = " + (256 * atanIndex(0, 0, 0, 10,false)) );
     System.out.println("for (0, 0) - (8.6, 5) atanIndex = " + atanIndex(0, 0, 8.6, 5,false));
     System.out.println("for (0, 0) - (5, 5) atanIndex = " + atanIndex(0, 0, 5, 5,false));
     System.out.println("for (0, 0) - (5, 8.6) atanIndex = " + atanIndex(0, 0, 5, 8.6,false));
     System.out.println("for (0, 0) - (10, 0) atanIndex = " + atanIndex(0, 0, 10, 0,false)
                        + "index = " + (256 * atanIndex(0, 0, 10, 0,false)) );
     System.out.println("for (0, 0) - (-10, 0) atanIndex = " + atanIndex(0, 0, -10, 0,false));
     System.out.println("for (0, 0) - (5, -8.6) atanIndex = " + atanIndex(0, 0, 5, -8.6,false));
     System.out.println("for (0, 0) - (-5, 5) atanIndex = " + atanIndex(0, 0, -5, 5,false));
     System.out.println("for (0, 0) - (8.6, -5) atanIndex = " + atanIndex(0, 0, 8.6, -5,false));
     System.out.println("for (0, 0) - (0, -10) atanIndex = " + atanIndex(0, 0, 0, -10,false));


  }

}

