package JavaQuat20;

import java.math.*;
import java.util.*;
import java.text.*;

public class FractalSpec
{
  static final int CENTER = 0;
  static final int COEFF = 1;
  static final int TRANS = 2;
  public BigQuaternion center;
  public BigQuaternion coeff;
  public BigQuaternion trans;
  public FastQuaternion vx; // Horizontal axis
  public FastQuaternion vy; // Vertical axis
  public FastQuaternion vz; // Primary perpendicular to image
  public double range = 2.5;
  public int rangeOver = CENTER;
  public int pix_width = 100;
  public int pix_height = 100;
  public int depth = 254;
  public String which = "Loop22";
  public int panels = 1;
  public int panel_x = 0;
  public int panel_y = 0;
  public boolean colorsIs256_p = true;
  public double color_cycles = 1;
  public int pts_per_pixel = 1;
  public boolean hifi_p = false;
  public double version = 2.0;
  
  public double hAxisScale = 1.0; // for printers that can print more densely horiz
  public ThumbNailProducer thumbNail = null;
  public String pathname = null;
  
  public FractalSpec ()
  {
    center = new BigQuaternion();
    coeff = new BigQuaternion();
    trans = new BigQuaternion();
    vx = new FastQuaternion();
    vy = new FastQuaternion();
    vz = new FastQuaternion();
    center.r = -0.75;
    vx.r = 1;
    vy.i = 1;
    coeff.setR(1.0);
    thumbNail = null;
  }
  
  public String toString (int log_id, String userComment) throws Throwable
  {    
    StringBuffer buf = new StringBuffer();
    buf.append("((log_id ");  buf.append(Integer.toString(log_id));

    // The Bs make the numbers into words so that these BigDecimals cal
    // be read in entirely and not truncated to float or double.
    buf.append(" )(ct_r B");    buf.append(Utility.removeExponent(center.rToString()));
    buf.append(" )(ct_i B");    buf.append(Utility.removeExponent(center.iToString()));
    buf.append(" )(ct_j B");    buf.append(Utility.removeExponent(center.jToString()));
    buf.append(" )(ct_k B");    buf.append(Utility.removeExponent(center.kToString()));
    
    buf.append(" )(co_r B");    buf.append(Utility.removeExponent(coeff.rToString()));
    buf.append(" )(co_i B");    buf.append(Utility.removeExponent(coeff.iToString()));
    buf.append(" )(co_j B");    buf.append(Utility.removeExponent(coeff.jToString()));
    buf.append(" )(co_k B");    buf.append(Utility.removeExponent(coeff.kToString()));

    buf.append(" )(tr_r B");    buf.append(Utility.removeExponent(trans.rToString()));
    buf.append(" )(tr_i B");    buf.append(Utility.removeExponent(trans.iToString()));
    buf.append(" )(tr_j B");    buf.append(Utility.removeExponent(trans.jToString()));
    buf.append(" )(tr_k B");    buf.append(Utility.removeExponent(trans.kToString()));

    buf.append(" )(vx_r ");    buf.append(String.valueOf(vx.rToString()));
    buf.append(" )(vx_i ");    buf.append(String.valueOf(vx.iToString()));
    buf.append(" )(vx_j ");    buf.append(String.valueOf(vx.jToString()));
    buf.append(" )(vx_k ");    buf.append(String.valueOf(vx.kToString()));
    
    buf.append(" )(vy_r ");    buf.append(String.valueOf(vy.rToString()));
    buf.append(" )(vy_i ");    buf.append(String.valueOf(vy.iToString()));
    buf.append(" )(vy_j ");    buf.append(String.valueOf(vy.jToString()));
    buf.append(" )(vy_k ");    buf.append(String.valueOf(vy.kToString()));

    buf.append(" )(range ");     buf.append(Double.toString(range));
    buf.append(" )(hifi ");      buf.append(hifi_p ? "T" : "F");
    buf.append(" )(version ");   buf.append(Double.toString(version));
    
    buf.append(" )(rangeOver "); buf.append(Integer.toString(rangeOver));
    buf.append(" )(panels ");    buf.append(Integer.toString(panels));   
    buf.append(" )(panel_x ");   buf.append(Integer.toString(panel_x));  
    buf.append(" )(panel_y ");   buf.append(Integer.toString(panel_y));
    
    buf.append(" )(width ");     buf.append(Integer.toString(pix_width));
    buf.append(" )(height ");    buf.append(Integer.toString(pix_height));
    buf.append(" )(depth ");     buf.append(Integer.toString(depth));
    buf.append(" )(id ");        buf.append(which);  
    buf.append(" )(ccycles ");   buf.append(Double.toString(color_cycles)); 
    buf.append(" )(pt_pix ");    buf.append(Integer.toString(pts_per_pixel)); 
    buf.append(" )(comment ");
    
    buf.append("Created: ");
    DateFormat df = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.SHORT);
    Date now = new Date();
    buf.append(df.format(now));
    buf.append("\n");
    buf.append(userComment); 
    buf.append(" end_of_comment ))");
    return buf.toString();
    
//    buf.append("Created:" + Integer.toString(now.getMonth() + 1) +
//                    "/" + now.getDate() + "/" + now.getYear() +
//                    "@" + now.getHours() + ":" + now.getMinutes() + ":" +       
//                    now.getSeconds());
  }
 
  
  public FractalSpec (FractalSpec arg)
  {
    center = new BigQuaternion();
    coeff = new BigQuaternion();
    trans = new BigQuaternion();
    vx = new FastQuaternion();
    vy = new FastQuaternion();
    vz = new FastQuaternion();
    center.assign_from(arg.center);
    vx.assign_from(arg.vx);
    vy.assign_from(arg.vy);
    vz.assign_from(arg.vz);
    coeff.assign_from(arg.coeff);
    trans.assign_from(arg.trans);
    range = arg.range;
    rangeOver = arg.rangeOver;
    pix_width = arg.pix_width;
    pix_height = arg.pix_height;
    depth = arg.depth;
    which = new String(arg.which);
    panels = arg.panels;
    panel_x = arg.panel_x;
    panel_y = arg.panel_y;
    colorsIs256_p = arg.colorsIs256_p;
    color_cycles = arg.color_cycles;
    pts_per_pixel = arg.pts_per_pixel;
    hifi_p = arg.hifi_p;
    version = arg.version;
    hAxisScale = arg.hAxisScale;
    thumbNail = arg.thumbNail;
  }
  
  public void setThumbNail (ThumbNailProducer arg)
  {
    thumbNail = arg;
  }
  
  public FractalSpec assign_from (FractalSpec arg)
  {
    return assignFrom(arg);
  }
  
  public FractalSpec assignFrom (FractalSpec arg)
  {
    center.assign_from(arg.center);
    coeff.assign_from(arg.coeff);
    trans.assign_from(arg.trans);
    vx.assign_from(arg.vx);
    vy.assign_from(arg.vy);
    vz.assign_from(arg.vz);
    range = arg.range;
    rangeOver = arg.rangeOver;
    pix_width = arg.pix_width;
    pix_height = arg.pix_height;
    depth = arg.depth;
    which = new String(arg.which);
    panels = arg.panels;
    panel_x = arg.panel_x;
    panel_y = arg.panel_y;
    colorsIs256_p = arg.colorsIs256_p;
    color_cycles = arg.color_cycles;
    pts_per_pixel = arg.pts_per_pixel;
    hifi_p = arg.hifi_p;
    version = arg.version;
    hAxisScale = arg.hAxisScale;
    thumbNail = arg.thumbNail;
    return this;
  }
  
  public boolean equal_p (FractalSpec arg)
    // This does not compare the primary perpendicular, vz, since it does not
    // affect the picture.  It is only useful for pivots.
  {
    return (center.equal_p(arg.center) &&
            vx.equal_p(arg.vx) &&
            vy.equal_p(arg.vy) &&
            coeff.equal_p(arg.coeff) &&
            trans.equal_p(arg.trans) &&
            range == arg.range &&
            rangeOver == arg.rangeOver &&
            pix_width == arg.pix_width &&
            pix_height == arg.pix_height &&
            depth == arg.depth &&
            ((which != null &&
              which.equals(arg.which) ) ||
             which == arg.which) &&
            panels == arg.panels &&
            panel_x == arg.panel_x &&
            panel_y == arg.panel_y &&
            colorsIs256_p == arg.colorsIs256_p &&
            color_cycles == arg.color_cycles &&
            pts_per_pixel == arg.pts_per_pixel &&
            hifi_p == arg.hifi_p &&
            version == arg.version &&
            hAxisScale == arg.hAxisScale);
  }
  
  
  BigQuaternion centralCoordinates ()
  {
    if (rangeOver == COEFF) {
      return coeff;
    }
    if (rangeOver == TRANS) {
      return trans;
    }
    return center;
  }

  
  public int correctScale ()
    // This sets the scale in the BigDecimal numbers used for center, coeff, and trans
    // to be no more than 1 bigger than necessary to handle the difference between pixels
    // in the picture that will be (has been) drawn from this parameter set.
  {  
    double currentPixSpacing = range / pix_width;
    int newScale =  (int)(Math.ceil(-1 * Math.log(currentPixSpacing)/Math.log(10))) + 1;
    if (newScale < 0) {
      newScale = 0;
    }
    center.setScaleLessOrEqual(newScale);
    coeff.setScaleLessOrEqual(newScale);
    trans.setScaleLessOrEqual(newScale);
    return newScale;
  }

}

