/*
 * @(#)GIFDecoder.java    0.10 9/13/97 Garr Lystad, with thanks to Adam Doppelt
 */

package JavaQuat20.GIFDecoderPkg;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.image.*;

/**
 * GIFDecoder is a class which reads a GIF file from an InputStream
 **/
 
public class GIFDecoder 
{
  private boolean initialized_p_ = false;
  private Header header_;
  private LogicalScreenDescriptor logicalScreenDescriptor_;
  private RGBxSizeSequence globalRGBTable_ = null;
  
  private Vector applicationExtensions_;
  private Vector commentExtensions_;
  private Vector graphicBlocks_;
  
  final int TRAILER = 0x3B;
  final int ZERO = 0;
  final int IMAGE = 0x2C;
  final int EXTENSION = 0x21;
  final int APPLICATION = 0xFF;
  final int COMMENT = 0xFE;
  final int CONTROL = 0xF9;
  final int TEXT = 0x01;
  
  public boolean initialized_p () {return initialized_p_;}
  public Header header () {return header_;}
  public LogicalScreenDescriptor logicalScreenDescriptor ()
  {
    return logicalScreenDescriptor_;
  }
  public RGBxSizeSequence globalRGBTable () {return globalRGBTable_;}
  public Vector applicationExtensions () {return applicationExtensions_;}
  public Vector commentExtensions () {return commentExtensions_;}
  public Vector graphicBlocks () {return graphicBlocks_;}

  public int getWidth () 
  { 
    return width();
    //return ((GraphicsBlock)graphicBlocks_.elementAt(0)).width();
  }
  
  public int getWidth (int pictureNumber) 
  { 
    return width(pictureNumber);
    //return ((GraphicsBlock)graphicBlocks_.elementAt(pictureNumber)).width();
  }
  
  public int getHeight() 
  { 
    return height();
    //return ((GraphicsBlock)graphicBlocks_.elementAt(0)).height();
  }
  
  public int getHeight(int pictureNumber) 
  { 
    return height(pictureNumber);
    //return ((GraphicsBlock)graphicBlocks_.elementAt(pictureNumber)).height();
  }
  
  public byte [] pixelsAsIndices (int pictureNumber) 
  { 
    if (pictureNumber >= graphicBlocks_.size())
      return null;
    return ((GraphicsBlock)graphicBlocks_.elementAt(pictureNumber)).pixels;
  }
  
  public byte [] pixelsAsIndices () 
  { 
    int pictureNumber = 0;
    if (pictureNumber >= graphicBlocks_.size())
      return null;
    return ((GraphicsBlock)graphicBlocks_.elementAt(pictureNumber)).pixels;
  }
 
  public int width (int pictureNumber) 
  { 
    if (pictureNumber >= graphicBlocks_.size())
      return -1;
    return    
      ((GraphicsBlock)graphicBlocks_.elementAt(pictureNumber)).imageDescriptor().width();
  }
  public int width () 
  { 
    int pictureNumber = 0;
    if (pictureNumber >= graphicBlocks_.size())
      return -1;
    return
      ((GraphicsBlock)graphicBlocks_.elementAt(pictureNumber)).imageDescriptor().width();
  }
 
  public int height (int pictureNumber) 
  { 
    if (pictureNumber >= graphicBlocks_.size())
      return -1;
    return    
      ((GraphicsBlock)graphicBlocks_.elementAt(pictureNumber)).imageDescriptor().height();
  }
  public int height () 
  { 
    int pictureNumber = 0;
    if (pictureNumber >= graphicBlocks_.size())
      return -1;
    return
      ((GraphicsBlock)graphicBlocks_.elementAt(pictureNumber)).imageDescriptor().height();
  }
  
  public boolean transparentPixelP (int pictureNumber)
  { 
    if (pictureNumber >= graphicBlocks_.size())
      return false;
    GraphicsBlock block = (GraphicsBlock)graphicBlocks_.elementAt(pictureNumber);
    if (block == null) {
//      System.out.println("In GIFDecoder.transparentPixelP(int) block is null unexpectedly.");
      return false;
    }
    if (block.control() == null) {
//      System.out.println("In GIFDecoder.transparentPixelP(int) block.control() is null " +
//                         "unexpectedly.");
      return false;
    }
    return block.control().transparentColorFlag();
  }
  
  public boolean transparentPixelP () 
  { 
    int pictureNumber = 0;
    if (pictureNumber >= graphicBlocks_.size())
      return false;
    GraphicsBlock block = (GraphicsBlock)graphicBlocks_.elementAt(pictureNumber);
    if (block == null) {
//      System.out.println("In GIFDecoder.transparentPixelP() block is null unexpectedly.");
      return false;
    }
    if (block.control() == null) {
//      System.out.println("In GIFDecoder.transparentPixelP() block.control() is null unexpectedly.");
      return false;
    }
    return block.control().transparentColorFlag();
  }

  public int transparentPixelIndex (int pictureNumber)
  { 
    if (pictureNumber >= graphicBlocks_.size())
      return -1;
    GraphicsBlock block = (GraphicsBlock)graphicBlocks_.elementAt(pictureNumber);
    return block.control().transparentColorIndex();
  }
  
  public int transparentPixelIndex () 
  { 
    int pictureNumber = 0;
    if (pictureNumber >= graphicBlocks_.size())
      return -1;
    GraphicsBlock block = (GraphicsBlock)graphicBlocks_.elementAt(pictureNumber);
    return block.control().transparentColorIndex();
  }
    
  public GIFDecoder (InputStream stream, boolean skipImageP) throws AWTException, IOException
  {
//    System.out.println("GIFDecoder entry.");
    if (stream == null)
      throw new IOException("File does not exist.");
    applicationExtensions_ = new Vector();
    commentExtensions_ = new Vector();
    graphicBlocks_ = new Vector();
    BufferedInputStream bstream = new BufferedInputStream(stream);
    header_ = new Header(bstream);
    logicalScreenDescriptor_ = new LogicalScreenDescriptor(bstream);
    if (logicalScreenDescriptor_.globalRGBTableFlag()) {
      globalRGBTable_ = 
        new RGBxSizeSequence(bstream, logicalScreenDescriptor_.sizeOfRGBTable());
    }
    
    // READ DATA BLOCKS
    int bb = bstream.read();
    while (bb != TRAILER) {
        int needlessVariable = bb;
      switch (needlessVariable) {  // Was bb.
        case 0: //ZERO:
         initialized_p_ = true; return;
        case 0x21: {// EXTENSION: {
//          System.out.println("Extension recognized.");
          if (header().version87_p())
            throw new AWTException("Illegal extension in GIF87a file.");
          GraphicsControl control = null;
          bb = bstream.read();
          if (bb == CONTROL) {
//            System.out.println("Control block recognized.");
            control = new GraphicsControl(bstream);
            bb = bstream.read();
          }
          switch (bb) {
            case 0xFF: // APPLICATION: 
//              System.out.println("Application recognized.");
              if (control != null)
                throw new AWTException("Control block preceeds Application block.");
              applicationExtensions_.addElement(new ApplicationExtension(bstream));
              break;
            case 0xFE: // COMMENT:
//              System.out.println("Comment recognized.");
              if (control != null)
                throw new AWTException("Control block preceeds Comment block.");
              commentExtensions_.addElement(new CommentExtension(bstream));
              break;
            case 1: // TEXT:
//              System.out.println("Text recognized.");
              graphicBlocks_.addElement(new TextExtension(bstream, control));
              break;
            case 0x2C: // IMAGE:
//              System.out.println("Image recognized 1.");
              graphicBlocks_.addElement(new GraphicsBlock(bstream, control, skipImageP));
              break;
            default:
              throw new AWTException("Unknown extension label code " + bb);
          }}
          break;
        case 0x2C: // IMAGE:
//          System.out.println("Image recognized 2.");
          graphicBlocks_.addElement(new GraphicsBlock(bstream, null, skipImageP));
          break;
        default:
          throw new AWTException("Unknown block code " + needlessVariable);
      }
      bb = bstream.read();
    }
//    System.out.println("Trailer recognized.");
    initialized_p_ = true;
  }
  
  public String getApplicationData(String id, String authCode)
  {
    if (applicationExtensions_ == null) {
//      System.out.println("getApplicationData Early exit no AEs.");
      return null;
    }
    for (int i = 0; i < applicationExtensions_.size(); i++) {
      ApplicationExtension ae = 
        (ApplicationExtension)applicationExtensions_.elementAt(i);
//      System.out.println("getApplicationData ident=" + ae.identifier());
      if (id.equals(ae.identifier())) {
        String aeAuthCode = Util.bytesToString(ae.authenticationCode);
//        System.out.println("getApplicationData authCode=" + aeAuthCode);
        if (aeAuthCode.equals(authCode)) {
          return ae.data();
        }
      }
    }
    return null;
  }  
}

class Util
{
  public static String bytesToString (byte [] bytes)
  {
    StringBuffer buf = new StringBuffer(bytes.length);
    for (int i = 0; i < bytes.length; i++) {
      buf.append((char)bytes[i]);
    }
    return buf.toString();
  }
  
  public static int bytesToInt (byte [] bytes)
  {
    int answer = 0;
    for (int i = 0; i < bytes.length; i++) {
      answer |= (bytes[i] & 0xFF) << (i * 8); //Remember without the "& 0xFF" it is signed.
    }
    return answer;
  }
  
  public static int extractBitsInt (byte val, int offset, int width)
  {
    return (val >> offset)& ((((int)Math.pow((double)2, (double)width)) - 1));
  }
  
  public static boolean extractBitBoolean (byte val, int bit)
  {
    if (1 == ((val >> bit) & 1))
      return true;
    return false;
  }
  
  public static String readDataString (BufferedInputStream bstream)
                                       throws IOException
  {
    int length;
    char ch;
    StringBuffer data_ = new StringBuffer();
    
    length = bstream.read();
    while (length > 0) {
      for (int i = 0; i < length; i++) {
        ch = (char)bstream.read();
        data_.append(ch);
      }
      length = bstream.read();
    }
    return data_.toString();
  }
  
  public static void skipDataBlocks (BufferedInputStream bstream) throws IOException
  {
    int length;
    length = bstream.read();
    while (length > 0) {
      bstream.skip(length);
      length = bstream.read();
    }
    // Note that at this point we have read the terminator block with length 0.
  }
}

class Header
{
  private boolean version87_p_ = false;
  byte [] buf;
  
  Header (BufferedInputStream bstream) throws IOException, AWTException
  {
    buf = new byte [3];
    bstream.read(buf, 0, 3);
    if (!Util.bytesToString(buf).equals("GIF"))
      throw new AWTException("Invalid File Type, not GIF");
    bstream.read(buf, 0, 3);
    if (Util.bytesToString(buf).equals("87a"))
      version87_p_ = true;
  }
  
  boolean version87_p () {return version87_p_;}
}

class LogicalScreenDescriptor
{
  int width_;
  int height_;
  boolean globalRGBTableFlag_;
  int colorResolution_;
  boolean sortFlag_;
  int sizeOfRGBTable_ = 0;
  int backgroundColorIndex_;
  int pixelAspectRatio_;

  int width () {return width_;}
  int height () {return height_;}
  boolean globalRGBTableFlag () {return globalRGBTableFlag_;}
  int colorResolution () {return colorResolution_;}
  boolean sortFlag () {return sortFlag_;}
  int sizeOfRGBTable () 
  {
    return (int)Math.pow((double)2, (double)(sizeOfRGBTable_ + 1));
  }
  int backgroundColorIndex () {return backgroundColorIndex_;}
  int pixelAspectRatio () {return pixelAspectRatio_;}

  LogicalScreenDescriptor (BufferedInputStream bstream)  
                          throws IOException, AWTException
  {
    byte byte1;
    byte [] word = new byte[2];
    bstream.read(word, 0, 2);
    width_ = Util.bytesToInt(word);
    bstream.read(word, 0, 2);
    height_ = Util.bytesToInt(word);
    byte1 = (byte)bstream.read();
    globalRGBTableFlag_ = Util.extractBitBoolean(byte1, 7);
    colorResolution_ = Util.extractBitsInt(byte1, 4, 3);
    sortFlag_ = Util.extractBitBoolean(byte1, 3);
    // We are to set this regardless of whether there is a global color table.
    sizeOfRGBTable_ = Util.extractBitsInt(byte1, 0, 3);
    backgroundColorIndex_ = bstream.read();
    pixelAspectRatio_ = bstream.read();
//    System.out.println("LogicalScreenDescriptor width=" + width_ + " height=" +
//      height_ + " flags=" + byte1 + " gColorTFlag=" + globalRGBTableFlag_ +
//      "colorResolution_=" + colorResolution_ + " sortFlag_=" + sortFlag_ );
  }
}

class ImageDescriptor
{
  private int leftPosition_;
  private int topPosition_;
  private int width_;
  private int height_;
  private boolean localRGBTableFlag_;
  private boolean interlaceFlag_;
  private boolean sortFlag_;
  private int sizeOfRGBTable_;
  
  public int leftPosition () {return leftPosition_;}
  public int topPosition () {return topPosition_;}
  public int width () {return width_;}
  public int height () {return height_;}
  public boolean localRGBTableFlag () {return localRGBTableFlag_;}
  public boolean interlaceFlag () {return interlaceFlag_;}
  public boolean sortFlag () {return sortFlag_;}
  public int sizeOfLocalRGBTable ()
  {
    return (int)Math.pow((double)2, (double)(sizeOfRGBTable_ + 1));  
  }
  public int codesize () { return sizeOfRGBTable_ + 1; }
  
  ImageDescriptor (BufferedInputStream bstream) throws IOException
  {
    byte byte1;
    byte [] word = new byte[2];
    bstream.read(word, 0, 2);
    leftPosition_ = Util.bytesToInt(word);
    bstream.read(word, 0, 2);
    topPosition_ = Util.bytesToInt(word);
    bstream.read(word, 0, 2);
    width_ = Util.bytesToInt(word);
    bstream.read(word, 0, 2);
    height_ = Util.bytesToInt(word);
    byte1 = (byte)bstream.read();
    localRGBTableFlag_ = Util.extractBitBoolean(byte1, 7);
    interlaceFlag_ = Util.extractBitBoolean(byte1, 6);
    sortFlag_ = Util.extractBitBoolean(byte1, 5);
    sizeOfRGBTable_ = Util.extractBitsInt(byte1, 0, 3);
  }
}

class GraphicsBlock
  // This is the "Table Based Image", consisting of an ImageDescriptor
  // an optional localRGBTable, and the image itself, known as pixels.
  // Since I personally don't want to the pixels and it's trouble to decode them
  // pixels is always null.
{
  private GraphicsControl control_ = null;
  private ImageDescriptor imageDescriptor_;
  private RGBxSizeSequence localRGBTable_ = null;
  
  public GraphicsControl control () {return control_;}
  public ImageDescriptor imageDescriptor () {return imageDescriptor_;}
  public RGBxSizeSequence localRGBTable () {return localRGBTable_;} //********** Needed
  public byte [] pixels = null;
  public int width () { return imageDescriptor_.width(); }
  public int height () { return imageDescriptor_.height(); }
  
  GraphicsBlock (BufferedInputStream bstream, GraphicsControl control, boolean skipImageP)
                 throws IOException
  {
    control_ = control;
    imageDescriptor_ = new ImageDescriptor(bstream);
    if (imageDescriptor_.localRGBTableFlag()) {
      localRGBTable_ =
        new RGBxSizeSequence(bstream, imageDescriptor_.sizeOfLocalRGBTable());
    }
    
    // Decode the picture.
    int width = imageDescriptor_.width();
    int height = imageDescriptor_.height();
    pixels = new byte[width * height];
    int LZWMinCodeSize = bstream.read();
    if (skipImageP) {
      Util.skipDataBlocks(bstream);
    } else {
      LZWInflator.inflate(bstream, LZWMinCodeSize, pixels);
    }
  }
}

class GraphicsControl
{
  private int disposalMethod_;
  private boolean userInputFlag_;
  private boolean transparentColorFlag_;
  private int displayTime_;
  private int transparentColorIndex_ = 0;
  
  public int disposalMethod () {return disposalMethod_;}
  public boolean userInputFlag () {return userInputFlag_;}
  public boolean transparentColorFlag () {return transparentColorFlag_;}
  public int displayTime () {return displayTime_;}
  public int transparentColorIndex () {return transparentColorIndex_;}
  
  GraphicsControl (BufferedInputStream bstream) throws IOException, AWTException
  {
    byte byte1;
    byte [] word = new byte[2];
    byte1 = (byte)bstream.read(); //This should read a block size of 4.
    if (byte1 != 4) {
      throw new AWTException("GraphicsControl block size should have been 4, was " +
                             byte1);
    }
    byte1 = (byte)bstream.read();
    disposalMethod_ = Util.extractBitsInt(byte1, 2, 3);
    userInputFlag_ = Util.extractBitBoolean(byte1, 1);
    transparentColorFlag_ = Util.extractBitBoolean(byte1, 0);
    bstream.read(word, 0, 2);
    displayTime_ = Util.bytesToInt(word);
    transparentColorIndex_ = bstream.read();
    if (!transparentColorFlag_)
      transparentColorIndex_ = 0;
    byte1 = (byte)bstream.read(); // This is the terminator and should be 0;
    if (byte1 != 0) {
      throw new AWTException("GraphicsControl end should have been 0, was " + byte1);
    }
  }
}

class CommentExtension
{
  private String data_;
  
  public String data ()
  {
    return data_;
  }
  
  CommentExtension (BufferedInputStream bstream) throws IOException
  {
    data_ = Util.readDataString(bstream);
  }
}

class TextExtension
  // This is for text that is to be displayed in the picture
{
  private GraphicsControl control_;
  private int gridLeftPosition_;
  private int gridTopPosition_;
  private int gridWidth_;
  private int gridHeight_;
  private int cellWidth_;
  private int cellHeight_;
  private int foregroundColorIndex_;
  private int backgroundColorIndex_;
  private String data_;

  public GraphicsControl control () {return control_;}
  public int gridLeftPosition () {return gridLeftPosition_;}
  public int gridTopPosition () {return gridTopPosition_;}
  public int gridWidth () {return gridWidth_;}
  public int gridHeight () {return gridHeight_;}
  public int cellWidth () {return cellWidth_;}
  public int cellHeight () {return cellHeight_;}
  public int foregroundColorIndex () {return foregroundColorIndex_;}
  public int backgroundColorIndex () {return backgroundColorIndex_;}
  public String data ()
  {
    return data_;
  }

  TextExtension (BufferedInputStream bstream, GraphicsControl control) 
                throws IOException
  {
    byte [] word = new byte[2];
    byte byte1;
    
    control_ = control;
    bstream.read(word, 0, 2); gridLeftPosition_ = Util.bytesToInt(word);
    bstream.read(word, 0, 2); gridTopPosition_ = Util.bytesToInt(word);
    bstream.read(word, 0, 2); gridWidth_ = Util.bytesToInt(word);
    bstream.read(word, 0, 2); gridHeight_ = Util.bytesToInt(word);
    cellWidth_ = bstream.read();
    cellHeight_ = bstream.read();
    foregroundColorIndex_ = bstream.read();
    backgroundColorIndex_ = bstream.read();
    data_ = Util.readDataString(bstream);
  }
}

class ApplicationExtension
{
  private String identifier_;
  private String data_;
  
  public String identifier () {return identifier_;}
  public String data () {return data_;}
  public byte [] authenticationCode;
  
  ApplicationExtension (BufferedInputStream bstream) throws IOException, AWTException
  {
    byte byte1;
    byte [] id = new byte[8];
    authenticationCode = new byte[3];
    byte1 = (byte)bstream.read(); // This should be the value 11 decimal.
    if (byte1 != 11) {
      throw new AWTException("ApplicationExtension block length should be 11, was " +
                             byte1);
    }
    bstream.read(id, 0, 8);
    identifier_ = Util.bytesToString(id);
    bstream.read(authenticationCode, 0, 3);
    
    // Read data
    data_ = Util.readDataString(bstream);
  }
}


class BitFile 
{
    InputStream input_;
    byte buffer_[];
    int index_, bitsLeft_;
    int bytesRead_;
    
//public static int byteBlock = 0; 

    public BitFile(InputStream input) {
        input_ = input;
        buffer_ = new byte[256];
        index_ = 0;
        bytesRead_ = 0;
        bitsLeft_ = 8;
    }

    public int ReadBits(int numbits) throws IOException {
        int bitsRead = 0;
        int answer = 0;
        do {
            if ((index_ == bytesRead_ - 1 && bitsLeft_ == 0) || index_ >= bytesRead_) {
                bytesRead_ = input_.read();
                if (bytesRead_ < 0)
                  bytesRead_ += 256;
  // THIS IS THE BUGGY LINE OF CODE
  //              input_.read(buffer_, 0, bytesRead_);
                for (int j = 0; j < bytesRead_; j++ ) {
                  buffer_[j] = (byte)(input_.read() & 0xFF);
                }
//System.out.println("*** Begin block " + byteBlock);
//byteBlock++;
                index_ = 0;
                bitsLeft_ = 8;
            }
            
            if (numbits <= bitsLeft_) {
                int bits = 
                  ((buffer_[index_] >> (8 - bitsLeft_)) & ((1 << numbits) - 1));
                answer |= bits << bitsRead;
                if (bitsLeft_ == numbits) {
                  ++index_;
                  bitsLeft_ = 8;
                } else {
                  bitsLeft_ -= numbits;
                }
                numbits = 0;
            } else {
                int bits = ((buffer_[index_] >> (8 - bitsLeft_)) & ((1 << bitsLeft_) - 1));
                answer |= bits << bitsRead;
                bitsRead += bitsLeft_;
                numbits -= bitsLeft_;
                ++index_;
                bitsLeft_ = 8;
            }
        } while (numbits != 0);
        return answer;
    }
}

class LZWStringTable {
    private final static int RES_CODES = 2;  // clearcode_ and endofinfo_
    private final static short NEXT_FIRST = (short)0xFFFF;
    private final static int MAXBITS = 12;
    private final static int MAXSTR = (1 << MAXBITS);

    byte strChr_[];
    short strNxt_[];
    short numStrings_;
    
    byte reverseChr_[];
    short previousFindIndex_;

    public LZWStringTable() {
        strChr_ = new byte[MAXSTR];
        strNxt_ = new short[MAXSTR];
        reverseChr_ = new byte[MAXSTR];
        previousFindIndex_ = NEXT_FIRST;
    }
    
    public short numStrings () { return numStrings_;}

    public void AddCharString(short index, byte b)
    {
      if (numStrings_ >= MAXSTR) {
        System.err.println("GIFDecoder.AddCharString numStrings_ " + numStrings_ +
                           " is impossibly large.");
        return;
      }
      if (index >= numStrings_) {
        System.err.println("GIFDecoder.AddCharString index " + index +
                           " is larger than numStrings_ " + numStrings_);
        return;
      }
      strChr_[numStrings_] = b;
      strNxt_[numStrings_] = index; //(index != HASH_FREE) ? index : NEXT_FIRST;
      numStrings_++;
    }
    
    public int FindCharString(short index, byte [] pixelIndices, int startFill) 
    {
        // DEBUG
//        byte debugLastByte = 0;
//        int debugRepeatCount = 0;
//        int debugRepeatLoopValue = 0;
        // END DEBUG
      int reverseIndex = 0;  //We extract the bytes from the table in reverse order.
      short tableIndex = index;
      boolean unknownEntryP = false;
      
      if (index > numStrings_) {
        System.err.println("GIFDecoder.FindCharString index " + index +
                           " is too large.");
      }
      if (index == numStrings_ && previousFindIndex_ != NEXT_FIRST) {
        // This happens only when a sequence repeats and then the next char is the
        // first one of the sequence, causing the new table code to be output before
        // its value can be totally known in this decoding.  To learn the last char we
        // must first find the first char.
        unknownEntryP = true;
        // We need this to chain in the following do-loop, but don't know the char
        // so we cannot call AddCharString yet.  We will below since unknownEntryP == true.
        strNxt_[numStrings_] = previousFindIndex_;            //<<<< FIX???????
        // DEBUG - LA-rose 32 panels, (0, 0) panel.
//        if (startFill > 450 * 485 && startFill < 450 * 504 + 1) {
//          System.out.println("(FindCharString index == numStrings_ 1=" + index + 
//                             ",  previousFindIndex_" + previousFindIndex_ + ")");
//        }
        // END DEBUG
        // was AddCharString(previousFindIndex_, strChr_[previousFindIndex_]);
      }
      
      do {
        reverseChr_[reverseIndex++] = strChr_[tableIndex];            
      } while ((tableIndex = strNxt_[tableIndex]) != NEXT_FIRST);
      
      if (previousFindIndex_ != NEXT_FIRST) {
        byte subsequentChar = reverseChr_[reverseIndex - 1];  //<<<< FIX???????
        AddCharString(previousFindIndex_, subsequentChar);    //<<<< FIX???????        
        if (unknownEntryP) {                                  //<<<< FIX???????
          // Fill in the unknown char, the last (i.e. reverse 0) is the same as the first.
          reverseChr_[0] = subsequentChar;                    //<<<< FIX???????
        }
        // DEBUG
//        if (startFill > 450 * 485 && startFill < 450 * 504 + 1) {
//          System.out.println("(FindCharString !unknownEntryP 2,  previousFindIndex_" +  
//                             previousFindIndex_ + " char=" + 
//                             reverseChr_[reverseIndex - 1] + ")");
//        }
        // END DEBUG
      }
      previousFindIndex_ = index;
      
      // DEBUG
//        if (startFill > 450 * 489 && startFill < 450 * 504 + 1) {
//          System.out.println("(*Tdx at " + (startFill - 1) + " is " +
//                             index + " giving " + reverseIndex + "characters.)");
//        }
      // END DEBUG
      
      for(; reverseIndex > 0;) {
        pixelIndices[startFill++] = reverseChr_[--reverseIndex];
        // DEBUG
//        if (startFill > 450 * 490 && startFill < 450 * 504 + 1) {
//          if (reverseChr_[reverseIndex] == debugLastByte) {
//            debugRepeatCount++;
//          } else {
//            if (debugLastByte != 0) {
//              System.out.println("(FindCharString starting Pdx " + debugRepeatLoopValue +
//                                 " is " +
//                                 debugLastByte + " * " + debugRepeatCount + ")");
//            }
//            debugRepeatCount = 1;
//            debugRepeatLoopValue = (startFill - 1);
//            debugLastByte = reverseChr_[reverseIndex];
//          }
//        }
        // END DEBUG
      }
      
      // DEBUG
//      if (debugRepeatCount > 0) {
//        System.out.println("(FindCharString starting Pdx " + debugRepeatLoopValue +
//                            " is " + debugLastByte + " * " + debugRepeatCount + ")");
//      }
      // END DEBUG

      return startFill;
    }

    public void ClearTable(int codesize) {
        numStrings_ = 0;
        previousFindIndex_ = NEXT_FIRST;
        
        int w = (1 << codesize) + RES_CODES;
        for (int q = 0; q < w; q++)
            AddCharString(NEXT_FIRST, (byte)q);
    }
    
}

class LZWInflator 
{
  private static InputStream input_;  /* NEW */
  private static int codesize_;   /* NEW */
  private static short index_;
  private static int clearcode_, endofinfo_, numbits_, limit_, errcode_;
  private static short prefix_ = (short)0xFFFF;
  public static BitFile bitFile_ = null;
  private static LZWStringTable strings_ = null;
  private static boolean freshlyResetP_ = true;
  private static int nextPixel_ = 0;
  
  // These duplicate the above except that they are not static, for reading in parts.
  private InputStream input_2;  /* NEW */
  private int codesize_2;   /* NEW */
  private short index_2;
  private int clearcode_2, endofinfo_2, numbits_2, limit_2, errcode_2;
  private short prefix_2 = (short)0xFFFF;
  private BitFile bitFile_2;
  private LZWStringTable strings_2;
  private boolean freshlyResetP_2 = true;  
  private int nextPixel_2 = 0;
  
    
  public static void inflate (InputStream input, int codesize,
                                 byte pixels[]) throws IOException 
  {
// BitFile.byteBlock = 0;
 
     input_ = input;
     codesize_ = codesize;

     bitFile_ = new BitFile(input);
     strings_ = new LZWStringTable();

     clearcode_ = 1 << codesize;
     endofinfo_ = clearcode_ + 1;
    
     numbits_ = codesize + 1;
     limit_ = (1 << numbits_) - 1;
     
     strings_.ClearTable(codesize);
     nextPixel_ = 0;
     boolean expectClearcodeP_ = true;
     
     //DEBUG
//     int count796 = 0;
     
     for (; nextPixel_ < pixels.length + 1;) { // Plus 1 to be sure to read the endofinfo
       index_ = (short)bitFile_.ReadBits(numbits_);
       // DEBUG
//       if (index_ == 796) {
//         count796++;
//       }
//       if (count796 >= 19) {
//         System.out.println("index =" + index_ );
//       }
       // END DEBUG
       if (index_ == clearcode_) {
         strings_.ClearTable(codesize_);
         numbits_ = codesize_ + 1;
         limit_ = (1 << numbits_) - 1;
         expectClearcodeP_ = false;
       } else if (index_ == endofinfo_) {
         if (nextPixel_ != pixels.length)
           System.err.println("GIFDecode.LZWDepress: nextPixel = " + nextPixel_ +
                              " whereas pixels.length = " + pixels.length);
         break;
       } else {
         if (expectClearcodeP_)
           System.err.println("GIFDecode.LZWDepress: Expected clearcode. index_=" + index_);
         nextPixel_ = strings_.FindCharString(index_, pixels, nextPixel_);
       }
       if (strings_.numStrings() > limit_ && nextPixel_ < pixels.length) {
         // The last part of the test is for when the final code bumps up the numStrings
         // here.  In the encoder the last code does not write to the stringTable and
         // the encoder just outputs the endOfInfo code in the same number of bits. 
         // Thus we don't want to change it here and read the wrong thing.
         if (++numbits_ > 12) {
           numbits_ = 12;
           expectClearcodeP_ = true;
         }
         limit_ = (1 << numbits_) - 1;
//         System.out.println(" debug numbits_=" + numbits_);
       }
     }
   }
    
   public LZWInflator (InputStream input, int codesize) throws IOException  /* NEW */
   {
     input_2 = input;
     codesize_2 = codesize;

     bitFile_2 = new BitFile(input);
     strings_2 = new LZWStringTable();

     clearcode_2 = 1 << codesize;
     endofinfo_2 = clearcode_2 + 1;
    
     numbits_2 = codesize + 1;
     limit_2 = (1 << numbits_2) - 1;
     
     strings_2.ClearTable(codesize);
     nextPixel_2 = 0;
     freshlyResetP_2 = true;
   }
    
   public int readPixels (byte[] pixels, int nextPixel, int boundExclusive) 
       throws IOException  /* NEW */
      // Returns boundExclusive if all goes well, 
      //  something less if we stop filling pixels sooner
   {
     for (; nextPixel < boundExclusive;) {
       index_2 = (short)bitFile_2.ReadBits(numbits_2);
       if (index_2 == clearcode_2) {
         if (!freshlyResetP_2) {
           System.err.println("GIFDecode.LZWDepress: Clearcode error detected. - ABORT");
           return nextPixel;
         }
       } else if (index_2 == endofinfo_2) {
         if (nextPixel != pixels.length)
           System.err.println("GIFDecode.LZWDepress: nextPixel = " + nextPixel +
                              " whereas boundExclusive = " + boundExclusive);
         break;
       } else {
         if (freshlyResetP_2) {
           System.err.println("GIFDecode.LZWDepress: index_2 = " + index_2 +
                              " when clearcode_ was expected.");
         }
         nextPixel = strings_2.FindCharString(index_2, pixels, nextPixel);
       }
       freshlyResetP_2 = false;
       if (strings_2.numStrings() > limit_2) {
         if (++numbits_2 > 12) {
           strings_2.ClearTable(codesize_2);
           numbits_2 = codesize_2 + 1;
           freshlyResetP_2 = true;
         }
         limit_2 = (1 << numbits_2) - 1;
       }
     }
     return nextPixel;
   }
}


    
    
    
    
    
