package JavaQuat20;

import java.awt.*;
import java.io.*;
import java.math.*;
import java.net.*;

public class JQArgReader 
{
  private StreamTokenizer tokens;
  private FractalSpec values;
  private StringBuffer buf;
  
  public FractalSpec readArgFile (Reader stream1, int id, Messenger messenger) 
   // if id = -1 we read the first image_numger record
  {    
    values = new FractalSpec();
    values.version = 1.4; // This is the only version that did not record version, so that's
                          // the default.
    values.pix_height = 0; // Old images did not record height.  They had height = width.
                           // This signals code below that height has not been read.
    buf = new StringBuffer();
    
    // Set up the StreamTokenizer.
    tokens = new StreamTokenizer(stream1);
//    tokens.lowerCaseMode(true); - commented out because loop22 != Loop22
    tokens.eolIsSignificant(true);
    
    // Thank you Emacs (& RMS) for the following info.
    //Control characters 0 - 31, then space
    int bang = (int)'!'; // = 33
    int quote = (int)'''; // = 47
    tokens.wordChars(bang, quote);
    // the parens are next
    int openP = (int)'('; // = 48
    int closeP = (int)')'; // = 49
    tokens.whitespaceChars(openP, closeP);
    int star = (int)'*'; // = 32
    int slash = (int)'/'; // = 47
    tokens.wordChars(star, slash);
    // Then come the digits
    int colon = (int)':'; // = 58
    int atSign = (int)'@'; // 64
    tokens.wordChars(colon, atSign);
    // Then come the Capital letters
    int leftBracket = (int)'['; // = 91
    int backquote = (int)'`'; // 96
    tokens.wordChars(leftBracket, backquote);
    // Then come the small letters
    int leftBrace = (int)'{'; // = 123
    int tilde = (int)'~'; // 126
    tokens.wordChars(leftBrace, tilde);
    

    if (!skipToRightId(id)) {
      messenger.message("JQArgReader did not find the desired log-id", true);
      return null;
    }
 
    // Read the values
    boolean finished_p = false;
    try {
      // Advance to the correct arg set
      tokens.nextToken();
      while (tokens.ttype != StreamTokenizer.TT_EOF &&
             !finished_p) {
        switch (tokens.ttype) {
          case StreamTokenizer.TT_WORD: 
            if (tokens.sval.equals("log_id")) {
              finished_p = true; // We're at next record.
            } else {
              fillField(tokens.sval, buf); //calls nextToken.
            }
            break;
          default:
            tokens.nextToken();
            //finished_p = true;
        }
      }
    } catch (Throwable e) {
      buf.append("\nRead error: " + e.getMessage());
    }
    messenger.message(buf.toString(), false);
    values.correctScale();
    return values;
  }
  
  private String translateWhich (String arg)
  {
    if (arg.equals("Loop428"))
      return "LoopTetra2";
    if (arg.equals("LoopHCx"))
      return "HcxPower2";
    return arg;
  }
  
  private boolean skipToRightId (int id)
  {
    try {
//      System.out.println("skipToRightId looking for id=" + id);
      tokens.nextToken();
      while (true) {
//        System.out.println("skipToRightId " + tokens.ttype
//                           + " sval=" + tokens.sval);
        if (tokens.ttype == StreamTokenizer.TT_EOF) {
//          System.out.println("skipToRightId Exit F");
          return false;
        } else if (tokens.ttype == StreamTokenizer.TT_WORD &&
                   tokens.sval.equals("log_id") ) {
          tokens.nextToken();
          if (tokens.ttype == StreamTokenizer.TT_NUMBER &&
              (id < 0 || tokens.nval == id) ) {
//            System.out.println("skipToRightId Exit T");
            return true;
          }
        } else {
          tokens.nextToken(); // ADVANCE TO NEXT TOKEN
        }
      }
    } catch (IOException e) {
      System.err.println("skipToRightId: " + e.getMessage());
      return false;
    }
  }
    
  private void fillField (String field, StringBuffer buf) 
                            throws IOException
    // NOTE THAT THIS CALLS nextToken TO GET TO THE NEXT
    // ITEM AFTER THE LAST THING USED.  THIS IS NEEDED FOR
    // readNumber() (or readBigDecimal()) TO PROCESS SCIENTIFIC NOTATION.
  {
    //For warnings and comments.
    tokens.nextToken();
//    System.out.println("fillField field=" + field);
    
    if (field.equals("ct_r") || field.equals("Q1_r")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.center.setR(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.center.setR(readBigDecimal());
      } else {
        buf.append("Bad value: center.r\n");
      }
    } else if (field.equals("ct_i") || field.equals("Q1_i")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.center.setI(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.center.setI(readBigDecimal());
      } else {
        buf.append("Bad value: center.i\n");
      }
    } else if (field.equals("ct_j") || field.equals("Q1_j")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.center.setJ(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.center.setJ(readBigDecimal());
      } else {
        buf.append("Bad value: center.j\n");
      }
    } else if (field.equals("ct_k") || field.equals("Q1_k")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.center.setK(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.center.setK(readBigDecimal());
      } else {
        buf.append("Bad value: center.k\n");
      }
    } else if (field.equals("co_r") || field.equals("Q2_r")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.coeff.setR(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.coeff.setR(readBigDecimal());
      } else {
        buf.append("Bad value: coeff.r\n");
      }
    } else if (field.equals("co_i") || field.equals("Q2_i")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.coeff.setI(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.coeff.setI(readBigDecimal());
      } else {
        buf.append("Bad value: coeff.i\n");
      }
    } else if (field.equals("co_j") || field.equals("Q2_j")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.coeff.setJ(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.coeff.setJ(readBigDecimal());
      } else {
        buf.append("Bad value: coeff.j\n");
      }
    } else if (field.equals("co_k") || field.equals("Q2_k")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.coeff.setK(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.coeff.setK(readBigDecimal());
      } else {
        buf.append("Bad value: coeff.k\n");
      }
    } else if (field.equals("tr_r") || field.equals("Q3_r")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.trans.setR(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.trans.setR(readBigDecimal());
      } else {
        buf.append("Bad value: trans.r\n");
      }
    } else if (field.equals("tr_i") || field.equals("Q3_i")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.trans.setI(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.trans.setI(readBigDecimal());
      } else {
        buf.append("Bad value: trans.i\n");
      }
    } else if (field.equals("tr_j") || field.equals("Q3_j")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.trans.setJ(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.trans.setJ(readBigDecimal());
      } else {
        buf.append("Bad value: trans.j\n");
      }
    } else if (field.equals("tr_k") || field.equals("Q3_k")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.trans.setK(readNumberAsBigDecimal());
      } else if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.trans.setK(readBigDecimal());
      } else {
        buf.append("Bad value: trans.k\n");
      }
    } else if (field.equals("vx_r")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vx.r = readNumber();
      } else {
        buf.append("Bad value: vx.r\n");
      }
    } else if (field.equals("vx_i")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vx.i = readNumber();
      } else {
        buf.append("Bad value: vx.i\n");
      }
    } else if (field.equals("vx_j")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vx.j = readNumber();
      } else {
        buf.append("Bad value: vx.j\n");
      }
    } else if (field.equals("vx_k")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vx.k = readNumber();
      } else {
        buf.append("Bad value: vx.k\n");
      }
    } else if (field.equals("vy_r")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vy.r = readNumber();
      } else {
        buf.append("Bad value: vy.r\n");
      }
    } else if (field.equals("vy_i")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vy.i = readNumber();
      } else {
        buf.append("Bad value: vy.i\n");
      }
    } else if (field.equals("vy_j")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vy.j = readNumber();
      } else {
        buf.append("Bad value: vy.j\n");
      }
    } else if (field.equals("vy_k")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vy.k = readNumber();
      } else {
        buf.append("Bad value: vy.k\n");
      }
    } else if (field.equals("vz_r")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vz.r = readNumber();
      } else {
        buf.append("Bad value: vz.r\n");
      }
    } else if (field.equals("vz_i")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vz.i = readNumber();
      } else {
        buf.append("Bad value: vz.i\n");
      }
    } else if (field.equals("vz_j")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vz.j = readNumber();
      } else {
        buf.append("Bad value: vz.j\n");
      }
    } else if (field.equals("vz_k")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.vz.k = readNumber();
      } else {
        buf.append("Bad value: vz.k\n");
      }
    } else if (field.equals("range")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.range = readNumber();
      } else {
        buf.append("Bad value: range\n");
      }
    } else if (field.equals("hifi")) {
      if (tokens.ttype == StreamTokenizer.TT_WORD) {
        if (tokens.sval.equals("T") || tokens.sval.equals("t")) {
          values.hifi_p = true;
        } else if (tokens.sval.equals("F") || tokens.sval.equals("f")) {
          values.hifi_p = false;
        } else {
          buf.append("Bad value: hifi\n");
          values.hifi_p = false;
        }
        tokens.nextToken();
      } else {
//        System.out.println("Hifi found non-TT_WORD value.");
      }
    } else if (field.equals("version")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.version = readNumber();
      } else {
        buf.append("Bad value: version\n");
      }
    } else if (field.equals("full_colr")) {
      if (tokens.ttype == StreamTokenizer.TT_WORD) {
        if (tokens.sval.equals("T") || tokens.sval.equals("t")) {
          values.colorsIs256_p = false;
        } else if (tokens.sval.equals("F") || tokens.sval.equals("f")) {
          values.colorsIs256_p = true;
        } else {
          buf.append("Bad value: full_colr\n");
          values.colorsIs256_p = true;
        }
        tokens.nextToken();
      } else {
//        System.out.println("Hifi found non-TT_WORD value.");
      }
    } else if (field.equals("rangeOver")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.rangeOver = (int)tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: panels\n");
      }
    } else if (field.equals("panels")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.panels = (int)tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: panels\n");
      }
    } else if (field.equals("panel_x")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.panel_x = (int)tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: panel_x\n");
      }
    } else if (field.equals("panel_y")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.panel_y = (int)tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: panel_y\n");
      }
    } else if (field.equals("width")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.pix_width = (int)tokens.nval;
        if (values.pix_height == 0) // We haven't read height.  On old pictures we won't.
          values.pix_height = values.pix_width;
        tokens.nextToken();
      } else {
        buf.append("Bad value: picture width\n");
      }
    } else if (field.equals("height")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.pix_height = (int)tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: picture height\n");
      }
    } else if (field.equals("multwd")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        // NEVER USED IN RELEASED VERSION. values.mult_width = (int)tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: picture height\n");
      }
    } else if (field.equals("multht")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        // NEVER USED IN RELEASED VERSION. values.mult_height = (int)tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: picture height\n");
      }
    } else if (field.equals("depth")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.depth = (int)tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: depth\n");
      }
    } else if (field.equals("id")) {
      if (tokens.ttype == StreamTokenizer.TT_WORD) {
        values.which = tokens.sval;
        values.which = translateWhich(values.which);
        tokens.nextToken();
      } else {
        buf.append("Bad value: power/id\n");
      }
    } else if (field.equals("ccycles")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.color_cycles = tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: color cycles\n");
      }
    } else if (field.equals("pt_pix")) {
      if (tokens.ttype == StreamTokenizer.TT_NUMBER) {
        values.pts_per_pixel = (int)tokens.nval;
        tokens.nextToken();
      } else {
        buf.append("Bad value: points per pixel\n");
      }
    } else if (field.equals("comment")) {
      getComment(buf);
      tokens.nextToken();
    } else {
      buf.append("Unknown field in args record: " + field);
    }
  }
  
  private double readNumber () throws IOException
  {
    double front = tokens.nval;
    tokens.nextToken();
    if (tokens.ttype == StreamTokenizer.TT_WORD) {
      String numbers = new String("-0123456789");
      if ((tokens.sval.indexOf("e") == 0 || tokens.sval.indexOf("E") == 0) &&
          numbers.indexOf(tokens.sval.charAt(1)) >= 0 ) {
        // Above is a test for "e" followed by a digit or minus.
        int exponent = Integer.valueOf(tokens.sval.substring(1)).intValue();
        tokens.nextToken(); //We have used a token so advance.
        return front * Math.pow(10.0, (double)exponent);
      }
    }
    return front;
  } 
  
  private BigDecimal readNumberAsBigDecimal  () throws IOException
  {
    double value = readNumber ();
    try {
      return new BigDecimal(value);
    } catch (NumberFormatException exceptn) {
      String message = exceptn.getMessage();
      if (message == null)
        message = exceptn.toString();
      throw new IOException(message);
    }
  }
   
  private BigDecimal readBigDecimal () throws IOException
  {
    String asString = new String(tokens.sval);
    // Remove the leading letter that makes this a word.
    asString = asString.substring(1);
    tokens.nextToken();
    try {
      //System.out.println("JQArgReader.readBigDecimal sval substr=" + asString);
      return new BigDecimal(asString);
    } catch (NumberFormatException exceptn) {
      String message = exceptn.getMessage();
      if (message == null)
        message = exceptn.toString();
      throw new IOException(message);
    }
  }
  
  private void getComment(StringBuffer buf) throws IOException
  {
    while (!(tokens.ttype == StreamTokenizer.TT_WORD &&
             tokens.sval.equals("end_of_comment") )) {
      switch (tokens.ttype) {
        case StreamTokenizer.TT_WORD:
          buf.append(tokens.sval);
          break;
        case StreamTokenizer.TT_NUMBER:
          buf.append(tokens.nval);     //This was nval but got 1.0 for 1
          break;
        case StreamTokenizer.TT_EOL:
          buf.append("\n");
          break;
        case StreamTokenizer.TT_EOF:
          return; // Should never get here
        default:
//          buf.append(tokens.ttype);
      }
      buf.append(" ");
      tokens.nextToken();
    }
  }
  
}                 
                 
                 
