package JavaQuat20;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.io.*;
import java.math.*;
import java.util.*;
import java.net.*;
import JavaQuat20.GIFDecoderPkg.*;
import JavaQuat20.GIFEncoderPkg.*;
import JavaQuat20.FractalLoops.JQLooper;

class JQCanvas extends Canvas implements Runnable, MouseListener, ClipboardOwner, TransferCMap
  // As a Component this is an ImageObserver
{
  static final int bitsPerByte = 8;
  static final int bitsPerInt = 32;
  static final int COMPUTED = 1;
  static final int LOADgIF = 2;
  static final int LOADpNG = 3;
   
  private JQCentral jQCentral_;
  private JQFrame jQFrame_;
  int who_am_i_;
  Thread runner_;
  
  // Image computation or loading
  private int typeOfImagePreparation_ = COMPUTED;

  //Drawing control
  private boolean parameter_change_p_ = false;
  private boolean ever_drawn_p = false;
  private boolean other_change_requires_redraw_p_ = false;
  
  // Fractal Parameters
  public FractalSpec spec_;
  
  // Fractal drawing helpers.
  static int min_pix_width_ = 9; // Doubles as min_pix_height_
  int h_offset_ = 0;
  int v_offset_ = 0;
  private boolean stop_requires_new_image = false;
  
  // Mouse marking helpers
  int drag_anchor_x = -1;
  int drag_anchor_y = -1;
  
  // For box drawing
  boolean drawBox_p = false;
  boolean drawCross_p = false;
  int box_low_x;
  int box_low_y;
  int box_hi_x;
  int box_hi_y;
  
  // Color control
  int colors = 256;
  public Color [] color_map_ = new Color[colors]; // SHOULD NOT BE PUBLIC
  public byte [] reds = new byte[colors];         // SHOULD NOT BE PUBLIC
  byte [] greens = new byte[colors];
  byte [] blues = new byte[colors];
  ColorModel colorModel = null;
  IndexColorModel colorModel256 = null;
  
  Q1ImageProducer fractalProducer_ = null;
  Image refresher_ = null;

  TransImage clipBoardFlavor_ = null; 

  /* **************** Drawing Methods ************** */
  
  public void set_other_change_requires_redraw_p (boolean val)
  {
    other_change_requires_redraw_p_ = val;
  }
  
  public void set_stop_requires_new_image (boolean val)
  {
    stop_requires_new_image = val;
  }

  /* **************** Misc methods ******************* */
  
  JQCanvas (JQCentral central, JQFrame frame, int who_am_i, FractalSpec spec)
  {
    jQCentral_ = central;
    jQFrame_ = frame;
    who_am_i_ = who_am_i;
    spec_ = spec;
    SetUpColorMap();
    updateColorModel();
    spec_.which = "Loop22";
    if (who_am_i_ == 1) {
      spec_.center.setR(-0.75);
      spec_.range = 2.5;
      spec_.pix_width = 100;
      spec_.pix_height = 100;
      spec_.depth = 255;
      spec_.panels = 1;
    } else if (who_am_i_ == 2) {
      spec_.center.setR(0.3587762);
      spec_.center.setI(0.26421307);
      spec_.coeff.setR(2.0);
      spec_.range = 0.003054;
      spec_.pix_width = 600;
      spec_.pix_height = 600;
      spec_.which = "Loop25";
      spec_.depth = 510;
      spec_.panels = 1;
    } else if (who_am_i_ == 3) { // Mandelbrot
      spec_.panels = 1;
      spec_.depth = colors;
      spec_.center.setR(-0.75);
      spec_.range = 3.0;
    } else if (who_am_i_ == 4) { // The Squid
      spec_.center.setR(0.18535);
      spec_.center.setI(0.202147);
      spec_.range = 0.00822;
      spec_.depth = 1024;
      spec_.which = "Loop23";
      spec_.coeff.setR(3.0);
      spec_.pix_width = 300;
      spec_.pix_height = 300;
      spec_.colorsIs256_p = true;
      spec_.color_cycles = 4;
      spec_.panels = 1;
    } else if (who_am_i_ == 5) { // The Dragon
      spec_.center.setR(0.16022);
      spec_.center.setI(0.37655);
      spec_.range = 0.00025;
      spec_.depth = 512;
      spec_.which = "Loop25";
      spec_.coeff.setR(2.0);
      spec_.pix_width = 300;
      spec_.pix_height = 300;
      spec_.colorsIs256_p = true;
      spec_.color_cycles = 1;
      spec_.panels = 1;
    } else if (who_am_i_ == 6) { // The Torch
      spec_.hifi_p = true;
      spec_.center.setR(0.246373);
      spec_.center.setI(0.275996);
      spec_.range = 0.0003;
      spec_.depth = 5250;
      spec_.which = "Loop24";
      spec_.coeff.setR(2.5);
      spec_.pix_width = 300;
      spec_.pix_height = 300;
      spec_.colorsIs256_p = true;
      spec_.color_cycles = 1;
      spec_.panels = 1;
    }
    
    addMouseListener(this);
  }
  
  public int who_are_you ()
  {
    return who_am_i_;
  }
  
  public FractalSpec spec ()
  {
    return spec_;
  }
  
  public JQFrame parent ()
  {
    return jQFrame_;
  }
  
  public void message (String msg, boolean expose)
  {
    jQCentral_.message(msg, expose);
  }
  
  public boolean debugP ()
  {
    return jQCentral_.debugP();
  }
  
  public boolean copyImageToSystemClipboard ()
    // In progress
  {
    boolean debug = false;
    //if (refresher_ == null || drawing_p ()) {
    //  return false;
    //}
    Toolkit tk = Toolkit.getDefaultToolkit();
    if (debug) System.out.println("JQCanval.copyImageToSystemClipboard pt1");
    Clipboard sysClip = tk.getSystemClipboard(); 
    if (debug) System.out.println("JQCanval.copyImageToSystemClipboard pt2");
    if (clipBoardFlavor_ == null) {
      clipBoardFlavor_ = new TransImage();
    }
    
    //
    if (debug) System.out.println("JQCanval.copyImageToSystemClipboard pt3");
    Transferable data = sysClip.getContents(this);
    if (debug) System.out.println("JQCanval.copyImageToSystemClipboard pt4");
    if (data == null) {
      System.out.println("JQCanval.copyImageToSystemClipboard Clipboard data was null.");
      return false;
    }
    DataFlavor[] flavors = data.getTransferDataFlavors();
    if (debug) System.out.println("JQCanval.copyImageToSystemClipboard pt5");
    for (int i = 0; i < flavors.length; i++) {
      System.out.println("Clipboard supports " + flavors[i].getMimeType() + " which is " +
                         flavors[i].getRepresentationClass() );
    }
    if (debug) System.out.println("JQCanval.copyImageToSystemClipboard pt6");
        
    //sysClip.setContents(clipBoardFlavor_, this);
    return true;
  }
  
  // To implement ClipboardOwner
  public void lostOwnership(Clipboard clipboard, Transferable contents)
  {
    // We don't care that we lost ownership of the clipboard.
  } 
  /* ****************** Color Methods ******************* */
  
  public ColorModel colorModel ()
  {
    return colorModel;
  }
  
  public IndexColorModel colorModel256 ()
  {
    return colorModel256;
  }
  
  public DirectColorModel colorModelFull ()
  {
    if (colorModel == colorModel256)
      return null;
    return (DirectColorModel)colorModel;
  }
  
  public byte[] pixels ()
  {
    if (fractalProducer_ == null) 
      return null;
    return fractalProducer_.pixels();
  }
  
  public int[] intPixels ()
  {
    if (fractalProducer_ == null) 
      return null;
    return fractalProducer_.intPixels();
  }
  
  public void remapPixelIndices (int [] pixelMap)
    //This is called from the colorMapEditor when colors have been sorted.
  {
    if (fractalProducer_ != null) {
      fractalProducer_.remapPixelIndices(pixelMap);
      refresher_ = createImage(fractalProducer_);
      repaint();
    }
  }
  
  public Color getColorAt (int index)
  {
    if (index < color_map_.length) {
      return color_map_[index];
    }
    return null; 
  }
  
  public void transferColorMapTo (ColorMapOwner cmo)
   // tell cmo the color at each index.
  {
    for (int i = 0; i < color_map_.length; i++) {
      cmo.setIndexAndColor(i, getColorAt(i), i == (color_map_.length - 1));
    }
  }
  
  void updateColorModel ()
  {
    colorModel256 = new IndexColorModel(bitsPerByte, colors, reds, greens, blues);
    if (spec_.colorsIs256_p) {
      colorModel = colorModel256;
    } else /* if (colorModel == null || !(colorModel.getClass().toString().equals("DirectColorModel"))) */ {
      int maskBase = 255;
      colorModel = new DirectColorModel(24, maskBase << 16, maskBase << 8, maskBase);
    }
  }
  
  boolean setColorAtIndex (int index, Color c, boolean repaint_p)
    // This should only be called externally to change the colormap and redisplay.
  {
    if (index < color_map_.length) {
      color_map_[index] = c;
      reds[index] = (byte)c.getRed();
      greens[index] = (byte)c.getGreen();
      blues[index] = (byte)c.getBlue();
      if (repaint_p) {
        updateColorModel();
      }
      if (fractalProducer_ != null && repaint_p) { // For initial setting of colormap refresher == null.
        // This is for colorEditor initiated update.
        refresher_ = createImage(fractalProducer_);
        repaint(); //This may be a problem ***************
      }
      return true;
    }
    return false;
  }
  
  
  void SetUpColorMap ()
    // Set up the initial color map.
  {
    float colors_f = colors;
    int purple_offset = (int)(5 * colors_f / 6.0) + 12;
    int rotation;
    if (spec_.version == 1.4) {
      for (int i = 0; i < colors; i++) {
        rotation = i + purple_offset;
        if (rotation >= colors)
          rotation = rotation - colors;
        color_map_[i] = Color.getHSBColor((rotation / colors_f), (float)1.0, (float)1.0);
        reds[i] = (byte)color_map_[i].getRed();
        greens[i] = (byte)color_map_[i].getGreen();
        blues[i] = (byte)color_map_[i].getBlue();
      }
    } else {
      colors_f = colors - 1; // Divide them up into 255 places since last is set black.
      int lastColoredIndex = (int)colors_f -1; // Which is 254
      for (int i = 0; i < colors - 1; i++) {   // From 0 thru 254
        int rgbIndex = lastColoredIndex - i;
        rotation = i + purple_offset;
        if (rotation >= colors)
          rotation = rotation - colors;
        Color rainbow1 = Color.getHSBColor((rotation / colors_f), (float)1.0, (float)1.0);
        color_map_[rgbIndex] = rainbow1;
        reds[rgbIndex] =   (byte)rainbow1.getRed();
        greens[rgbIndex] = (byte)rainbow1.getGreen();
        blues[rgbIndex] =  (byte)rainbow1.getBlue();
      }
    }
    if (spec_.version == 1.4) {
      //Make the transparent pixel different from black.
      color_map_[0] = new Color(1, 1, 1); 
      reds[0] = 1;
      greens[0] = 1;
      blues[0] = 1;
      // Here is the black we use for our fractal, at colorMap location 1.
      color_map_[1] = new Color(0, 0, 0);
      reds[1] = 0;
      greens[1] = 0;
      blues[1] = 0;
    } else {
      // Here is the black we use for our fractal, at colorMap location 0.
      color_map_[255] = new Color(0, 0, 0); 
      reds[255] = 0;
      greens[255] = 0;
      blues[255] = 0;
    }
  }
      
  
  private void setColorMapFromRGB (byte [] reds, byte [] greens, byte [] blues)
    // Bytes are no longer unsigned, thus the asInt conversion.
  {
    for (int i = 0; i < reds.length; i++) {
      color_map_[i] = new Color(PosByte.asInt(reds[i]), PosByte.asInt(greens[i]),
                                PosByte.asInt(blues[i]));
    }
    updateColorModel();
  }

  public void set256ColorMapFromCanvas (JQCanvas sourceCanvas)
  {
    IndexColorModel model256 = sourceCanvas.colorModel256();
    model256.getReds(reds);
    model256.getGreens(greens);
    model256.getBlues(blues);
    setColorMapFromRGB(reds, greens, blues);
  }

/* ********************** Write related methods ************************* */

  public boolean imageExistsP ()
  {
    return (refresher_ != null);
  }
  
  public GIFEncoder createGIFEncoder () throws AWTException
  {
    return (new GIFEncoder(spec_.pix_width, spec_.pix_height, pixels(), colorModel256()));
  }
  
/* *********************** Other misc methods **************************** */

  public void stop ()
  {
    typeOfImagePreparation_ = COMPUTED; // Just in case we need to set it back.
    if (fractalProducer_ != null) {
      fractalProducer_.stop();
    }
  }

  public void setEverDrawnPFalse ()
  {
    ever_drawn_p = false;
  }
  
  private boolean trapImageProducerP (String which)
  {
    JQLooper looper = jQCentral_.findLoopObject(which);
    if (looper == null || jQCentral_.getTrapFrame() == null) {
      return false;
    }
    return looper.acceptsTrapGif();
  }
  
  private Q1ImageProducer createImageProducer (FractalSpec specArg)
    // Note that this is overloaded by the next method.
    // Returns a Q1ImageProducer or a derived class.  The base class produces
    // a normal fractal by varying pixels over a range about the center.  The
    // derived classes vary pixels over a range of the coefficient or the translate.
    // This is called by setParameters.
  {
    //System.out.println("JQCanvas.createImageProducer center=" + specArg.center.toString() + 
    //                   " coeff=" + specArg.coeff.toString() + " trans=" + specArg.trans.toString()); 
    if (trapImageProducerP(specArg.which))
      return new TrapImageProducer(specArg, this, jQCentral_.getTrapFrame());
    if (specArg.rangeOver == FractalSpec.CENTER)
      return new Q1ImageProducer(specArg, this);
    if (specArg.rangeOver == FractalSpec.COEFF)
      return (Q1ImageProducer)(new Q2ImageProducer(specArg, this));
    if (specArg.rangeOver == FractalSpec.TRANS)
      return (Q1ImageProducer)(new Q3ImageProducer(specArg, this));
    jQCentral_.message("JQCanvas:: specArg.rangeOver had illegal value" + specArg.rangeOver, true);
    specArg.rangeOver = FractalSpec.CENTER;
    return new Q1ImageProducer(specArg, this);
  }


  private Q1ImageProducer createImageProducer (Image refresher, FractalSpec specArg)
    // Note that this is overloaded by the previous method.
    // This is called by createDummyProducer.
  {
    if (trapImageProducerP(specArg.which))
      return new TrapImageProducer(refresher, specArg, this, jQCentral_.getTrapFrame());
    if (specArg.rangeOver == FractalSpec.CENTER)
      return new Q1ImageProducer(refresher, specArg, this);
    if (specArg.rangeOver == FractalSpec.COEFF)
      return (Q1ImageProducer)(new Q2ImageProducer(refresher, specArg, this));
    if (specArg.rangeOver == FractalSpec.TRANS)
      return (Q1ImageProducer)(new Q3ImageProducer(refresher, specArg, this));
    jQCentral_.message("JQCanvas:: specArg.rangeOver had illegal value" + specArg.rangeOver, true);
    specArg.rangeOver = FractalSpec.CENTER;
    return new Q1ImageProducer(refresher, specArg, this);
  }

  public void setParameters (FractalSpec specArg)
     // This sets the internal parameters of this fractal.
     // CALLER SHOULD MAKE SURE ALL PARAMETERS ARE VALID!!!
  {
    //System.out.println("JQCanvas.setParameters center=" + specArg.center.toString() + 
    //                   " coeff=" +
    //                   specArg.coeff.toString() + " trans=" + specArg.trans.toString()); 
    // check for a change
    parameter_change_p_ = !spec_.equal_p(specArg) || !ever_drawn_p ||
                          other_change_requires_redraw_p_;
    other_change_requires_redraw_p_ = false;
    int old_pix_width = spec_.pix_width;
    int old_pix_height = spec_.pix_height;
    int old_rangeOver = spec_.rangeOver;
    boolean old_colorsIs256_p = spec_.colorsIs256_p;
 
    // If any change then set new spec_
    
    if (parameter_change_p_) {
      spec_.assign_from(specArg);
      refresher_ = null; // So we will create a new image in the run method.
      
      // Caller makes sure vectors are normal and orthogonal.
      
      if (spec_.pix_width < min_pix_width_)
        spec_.pix_width = min_pix_width_;
      if (spec_.pix_height < min_pix_width_)
        spec_.pix_height = min_pix_width_;
        
      Rectangle canvas_size = getBounds();
//      if (spec_.pix_width > canvas_size.width)
//        spec_.pix_width = canvas_size.width;
//      if (spec_.pix_height > canvas_size.height)
//        spec_.pix_height = canvas_size.height;
        
      if (old_pix_width != spec_.pix_width || 
          old_pix_height != spec_.pix_height || 
          old_rangeOver != spec_.rangeOver) {
        // We will make one of the correct size below.
        fractalProducer_ = null;
      }
      
      boolean which_ok_p = jQCentral_.findLoopObject(spec_.which) != null;
      if (!which_ok_p)
        spec_.which = "Loop22";

      if (old_colorsIs256_p != spec_.colorsIs256_p) {
        updateColorModel();
        fractalProducer_ = null;
      }
      
      if (fractalProducer_ != null &&
          fractalProducer_.trapImageProducerP() != trapImageProducerP(spec_.which)) {
        // If this is changing we need a new ImageProducer.
        fractalProducer_ = null;
      }
      
      if (fractalProducer_ == null) {
        System.gc();
        fractalProducer_ = createImageProducer(spec_);
      }
      
      if (spec_.pts_per_pixel >= 9) {
        spec_.pts_per_pixel = 9;
      } else if (spec_.pts_per_pixel >= 4) {
        spec_.pts_per_pixel = 4;
      } else if (spec_.pts_per_pixel >= 2) {
        spec_.pts_per_pixel = 2;
      } else {
        spec_.pts_per_pixel = 1;
      }  

      h_offset_ = (int)Math.max(0, (canvas_size.width - spec_.pix_width) / 2);
      v_offset_ = (int)Math.max(0, (canvas_size.height - spec_.pix_height) / 2);
    }
//    jQCentral_.message("JQCanvas.setParameters spec_.center.iToString()=" +
//                        spec_.center.iToString(), false);
    spec_.correctScale(); // Use no more precision than necessary in center, coeff and trans.
//    upload_parameters(); // Show the user our corrections to his FractalSpec.
    fractalProducer_.setParameters(spec_, parameter_change_p_);
  }
  
  public JQCentral jQCentral ()
    // Called by the Image Producers in one place.
  {
    return jQCentral_;
  }
  
//  public void upload_parameters ()
//  {
//    jQCentral_.setParameters(spec_);
//  }
  
//  public FractalSpec get_parameters ()
//  {
//    return spec_;
//  }

  
  /* ******************** RUN AND PAINT methods *********************** */


  public synchronized void threaded_paint_fractal ()
  {
    // System.out.println("JQCanvas.threaded_paint_fractal entry");
    if (drawing_p()) {
      jQCentral_.message("Draw on this canvas already in progress.", false);
      return;
    }
    //fractalProducer_.setDrawingFractalInterlock_p(true);
    spec_.version = 2.0; // Mark it as JavaQuat 2.0.
    runner_ = new Thread(this, "image process " + who_am_i_);
    runner_.setPriority(Thread.MIN_PRIORITY);
//    jQCentral_.message("JQCanvas.threaded_paint_fractal about to start.", false);
    runner_.start();
  }
  
  public boolean drawing_p ()
  {
    if (fractalProducer_ != null &&
        fractalProducer_.drawingFractalInterlock_p() )
      return true;
    return false;
  }
  
  public void run ()
  {
    // In order to see the redraw, we need to clear the image and yield to that process.
    //System.out.println("JQCanvas.run process name = " + Thread.currentThread().getName());
    Graphics gr = getGraphics();
    if (gr != null)
      gr.clearRect(0, 0, getSize().width, getSize().height);
    
    int last_row = fractalProducer_.spec_.pix_height - 1 - (fractalProducer_.spec_.hifi_p ? 0 : 1);
    boolean justDidAdvance = false;
    if (refresher_ == null || fractalProducer_.last_row_drawn_ < last_row) {
      if (jQCentral_.loggingP()) {
        jQCentral_.writeToLog(spec_);
      }
      synchronized(this){
        if (drawing_p()) {
          jQCentral_.message("Draw on this canvas already in progress..", false);
          return;
        }
        fractalProducer_.setDrawingFractalInterlock_p(true);
      }
      fractalProducer_.drawTheFractal();
      // this thread waits for third thread to finish.  Then it calls for the Image. 
      while (fractalProducer_.drawingFractalInterlock_p()) {
        try {
          Thread.sleep(2000); // Two seconds;
        } catch (Throwable exception1) {}
      }
      // On Macintosh MRJ2.01 this mistakenly executes in primary gui listener thread
      // so if we let this call draw the image there would be no response to STOP button.
      // Therefore we call drawTheFractal above and then wait.
      refresher_ = createImage(fractalProducer_);
      if (fractalProducer_.lastRowDrawnP()) {
        jQFrame_.autoAdvance(spec_); //If not doing a sequence nothing happens.
      }
    } else {
      // This happens frequently with the first frame of a sequence when it
      //  was just drawn before the sequence was started.  Need to advance.
      jQFrame_.autoAdvance(spec_); //If not doing a sequence nothing happens.
    }
    repaint();
  }
    
  public void paint (Graphics g)
  {
    if (!ever_drawn_p && !parameter_change_p_) { //For before the first draw.
      return;
    }
    
    // We should be able to get meaningful bounds() now.
    Rectangle canvas_size = getBounds();
    h_offset_ = (int)Math.max(0, (canvas_size.width - spec_.pix_width) / 2);
    v_offset_ = (int)Math.max(0, (canvas_size.height - spec_.pix_height) / 2);
    
    if (refresher_ != null) {
//      System.out.println("Fractal.paint refresher_ = " + refresher_);
      if (fractalProducer_ != null) {
        int last_row = fractalProducer_.spec_.pix_height - 1 - (fractalProducer_.spec_.hifi_p ? 0 : 1);
        if (fractalProducer_.last_row_drawn_ == last_row && 
            stop_requires_new_image) {
          // If we've stopped, then when the picture finishes we need a new image from the producer
          // in order to get the last lines drawn.
          stop_requires_new_image = false;
          refresher_ = createImage(fractalProducer_);
        }
        g.drawImage(refresher_, h_offset_, v_offset_, this);
      }
      /* jQCentral_.main_background, was next to last arg in line above*/ 
      ever_drawn_p  = true;
      if (fractalProducer_ == null) {
        g.drawImage(refresher_, h_offset_, v_offset_, this); // Experimenting with this line or next one.
        createDummyProducer();
      }
    }
    if (drawBox_p) {
      drawBox(g);
    } else if (drawCross_p) {
      drawCross(g);
    }
  }

  public void drawBoxFlag (int lox, int loy, int hix, int hiy)
  {
    drawBox_p = true;
    box_low_x = lox + h_offset_;
    box_low_y = loy + v_offset_;
    box_hi_x = hix + h_offset_;
    box_hi_y = hiy + v_offset_;
    repaint();
  }
  
  
  private void drawBox (Graphics g)
  {
    g.setColor(Color.white);
    g.drawLine(box_low_x, box_low_y, box_hi_x, box_low_y); // low y
    g.drawLine(box_low_x, box_hi_y, box_hi_x, box_hi_y);   // hi y
    g.drawLine(box_low_x, box_low_y, box_low_x, box_hi_y); // low x
    g.drawLine(box_hi_x, box_low_y, box_hi_x, box_hi_y);   // hi x
    drawBox_p = false;
  }

  
  public void drawCrossFlag (int x, int y)
  {
    drawCross_p = true;
    box_low_x = x + h_offset_;
    box_low_y = y + v_offset_;
    repaint();
  } 
  
  private void drawCross (Graphics g)
  {
    g.setColor(Color.white);
    int pwh = spec_.pix_width + h_offset_;
    int pwv = spec_.pix_height + v_offset_;
    g.drawLine(box_low_x, Math.max(0, box_low_y - 4), box_low_x, Math.max(0, box_low_y - 2)); // top
    g.drawLine(box_low_x, Math.min(pwv, box_low_y + 4), box_low_x, Math.min(pwv, box_low_y + 2));   // bottom
    g.drawLine(Math.max(0, box_low_x - 4), box_low_y, Math.max(0, box_low_x - 2), box_low_y); // left
    g.drawLine(Math.min(pwh, box_low_x + 4), box_low_y, Math.min(pwh, box_low_x + 2), box_low_y);   // right
    drawCross_p = false;
  }
  
  public boolean keepCurrentDrawLine (int x1, int x2, int y, Color color)
    // Called from fractalProducer_ as it draws in order to keep us currnt.
  {
//    System.out.println("JQCanvas.keepCurrentDrawLine colorIndex = " + colorIndex);
    Graphics ggg = getGraphics();
    if (null == ggg) {
      // This could happen in JQ 1.x, it seems, when we switch to/from 1 big canvas while drawing.
      return false;
    }
    ggg.setColor(color);
    ggg.drawLine(h_offset_ + x1, v_offset_ + y, h_offset_ + x2, v_offset_ + y);
    return true;
  }
   
  public Image refresher ()
  {
    return refresher_;
  }
  
  
  /* ******************** MOUSE METHODS *********************** */
  
  public void mouseClicked(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}
  
  public void mousePressed(MouseEvent evt)
  {
    drag_anchor_x = evt.getX();
    drag_anchor_y = evt.getY();
  }
  
  public void mouseReleased(MouseEvent evt)
  {
    if (jQFrame_.doingSequenceP()) {
      return;
    }
  
    jQCentral_.setActiveCanvas(this); //Transfers colormap
    int x = evt.getX();
    int y = evt.getY();
    
    if (drag_anchor_x < 0) {// Check that we've started the drag in the canvas.
      drag_anchor_x = -1; // Mark that we've finished.
      return;
    }
    if (!ever_drawn_p)  {// Check that there's a picture there to enlarge.
      drag_anchor_x = -1; // Mark that we've finished.
      return;
    }
    
    FractalSpec new_parms = new FractalSpec(spec_);
    
    // SIZE/RANGE CALCULATIONS
    int h_new_size = (int)Math.abs(x - drag_anchor_x);
    int v_new_size = (int)Math.abs(y - drag_anchor_y);
    boolean dragged_p = h_new_size > 2 && v_new_size > 2;
    Dimension hVRatio = jQCentral_.getHVRatio();
    int center_x = - h_offset_ - 1; // -1 is empirical, i.e. it looks right.
    int center_y = - v_offset_ - 1;
    if (!RatioDialog.offP(hVRatio)) {
      int width_mult = Math.round(((float)h_new_size) / hVRatio.width);
      int height_mult = Math.round(((float)v_new_size) / hVRatio.height);
      int max_mult = Math.max(width_mult, height_mult);
      h_new_size = max_mult * hVRatio.width;
      v_new_size = max_mult * hVRatio.height;
      if (x > drag_anchor_x) {
        center_x += drag_anchor_x +  h_new_size / 2;
      } else {
        center_x += drag_anchor_x - h_new_size / 2;
      }
      if (y > drag_anchor_y) {
        center_y += drag_anchor_y + v_new_size / 2;
      } else {
        center_y += drag_anchor_y - v_new_size / 2;
      }
    } else { 
      // CENTER CALCULATIONS
       //new center within current panel.
      center_x += ((x + drag_anchor_x) / 2);
      center_y += ((y + drag_anchor_y) / 2);
    }
    
    int h_inter_pixel_spaces = spec_.pix_width * spec_.panels - 1;
    int v_inter_pixel_spaces = spec_.pix_height * spec_.panels - 1;
    int pix_rt_of_abs_center = center_x + h_inter_pixel_spaces / -2 + spec_.panel_x * spec_.pix_width;
    int pix_above_abs_center = - center_y + v_inter_pixel_spaces / 2 - spec_.panel_y * spec_.pix_height;
    if (new_parms.rangeOver == FractalSpec.COEFF) {
      new_parms.coeff.plus_equal(fractalProducer_.incx.times_scalar(pix_rt_of_abs_center));
      new_parms.coeff.plus_equal(fractalProducer_.incy.times_scalar(pix_above_abs_center)); //Y is down.
    } else if (new_parms.rangeOver == FractalSpec.TRANS) {
      new_parms.trans.plus_equal(fractalProducer_.incx.times_scalar(pix_rt_of_abs_center));
      new_parms.trans.plus_equal(fractalProducer_.incy.times_scalar(pix_above_abs_center)); //Y is down.
    } else { // we will assume new_parms.rangeOver == FractalSpec.CENTER for safety.
      new_parms.center.plus_equal(fractalProducer_.incx.times_scalar(pix_rt_of_abs_center));
      new_parms.center.plus_equal(fractalProducer_.incy.times_scalar(pix_above_abs_center)); //Y is down.
    }
    new_parms.correctScale();

    drag_anchor_x = -1; // Mark that we've finished with the anchor.

    if (dragged_p) { // Set range and draw box if user really made drag.  Else keep range.
      // COMPUTE NEW WIDTH AND HEIGHT.
//      FractalSpec parmsWindowSpec = new FractalSpec();
//      jQCentral_.parmsToSpec(parmsWindowSpec);
      double parmWidth = jQCentral_.parmFramePixelsWide(); //Get area user entered in case of change.
      double parmHeight = jQCentral_.parmFramePixelsHigh();
      if (parmWidth <= 0 || parmHeight <= 0) {
        parmWidth = spec_.pix_width;
        parmHeight = spec_.pix_height;
      }
      if (!RatioDialog.offP(hVRatio)) {
        // Fixed ratio is on.
        // Round the multiple, we want the ratio between h_new_size and v_new_size in the final numbers.
        double inflationD = Math.sqrt((parmWidth * parmHeight)/(hVRatio.width * (float)hVRatio.height));
        new_parms.pix_width = hVRatio.width * (int)Math.round(inflationD);
        new_parms.pix_height = hVRatio.height * (int)Math.round(inflationD);
      } else {
        // Round the product, we'd rather keep the area as nearly constant as possible.
        double inflationD = Math.sqrt((parmWidth * parmHeight)/(h_new_size * (float)v_new_size));
        new_parms.pix_width = (int)Math.round(h_new_size * inflationD);
        new_parms.pix_height = (int)Math.round(v_new_size * inflationD);
      }

      // COMPUTE BOX
      int h_half_size = h_new_size / 2;
      int v_half_size = v_new_size / 2;
      drawBoxFlag(center_x - h_half_size, center_y - v_half_size,
                  center_x + h_half_size, center_y + v_half_size);
      new_parms.range = ((h_new_size * new_parms.range)/ h_inter_pixel_spaces);
      jQCentral_.setParameters(new_parms);
      jQCentral_.setParmFramePanelFields(1, 0, 0);
    } else { // Not dragged.
      // Set up cross drawing, give point to JQCentral, maybe give index to color Map Editor.
      drawCrossFlag(center_x, center_y); // Remove one below
      int parmFramePanels = jQCentral_.parmFramePanels();
      boolean pointUsed_p = jQCentral_.receivePoint(this, new_parms);  // $$$$$$
      
      if (pointUsed_p) {
        jQCentral_.message("JQCanvas.mouseReleased pointUsed_p", false);
        if (parmFramePanels == spec_.panels) {
          // # of panels unchanged: the user just clicked on a panel and probably would like to draw it.
          jQCentral_.setParmFramePanelFields(1, 0, 0);
          jQCentral_.setParmFrameRange(spec_.range/spec_.panels);
        } else if (parmFramePanels > 1) {
          // User has set different number of panels and would like to know panel values.
          jQCentral_.receivePoint(this, spec_); //Don't want new center.
          double fractionRightOfLeftEdge = 0.5 + pix_rt_of_abs_center/(double)(spec_.pix_width * spec_.panels);
          double fractionDownFromTop = 0.5 - pix_above_abs_center/(double)(spec_.pix_height * spec_.panels);
          int x_panel = (int)(fractionRightOfLeftEdge * parmFramePanels);
          int y_panel = (int)(fractionDownFromTop * parmFramePanels);
          jQCentral_.setParmFramePanelFields(parmFramePanels, x_panel, y_panel);
        }
      } else if (jQCentral_.activeCanvasP(this)) {
        jQCentral_.message("JQCanvas.mouseReleased pointUsed_p FALSE", false);
        if (center_x < 0) center_x = 0;
        if (center_y < 0) center_y = 0;
        if (center_x >= spec_.pix_width) center_x = spec_.pix_width - 1;
        if (center_y >= spec_.pix_height) center_y = spec_.pix_height - 1;
        int color_map_index = fractalProducer_.colorIndexAt(center_x, center_y); //These args OK.
        if (color_map_index >= 0 && color_map_index <= colors - 1) { 
          Color c = getColorAt(color_map_index);
          jQCentral_.setFocusOfColorMapEditor(color_map_index, c, true);
        }
      }
    }
  }


  public void startProductionFinished ()
    // This is the callback for when the image is created.
  {
    jQFrame_.autoAdvance(spec_); //If we're not doing a sequence nothing happens.
    jQCentral_.message("Image complete.", false);  //300
  }
  
  // -------------------------- Loading Images  --------------------------------
  
  public void whichBits (int flags)
  {
    String abort = ((flags & ABORT) != 0) ? " ABORT" : "";
    String allbits = ((flags & ALLBITS) != 0) ? " ALLBITS" : "";
    String error = ((flags & ERROR) != 0) ? " ERROR" : "";
    String framebits = ((flags & FRAMEBITS) != 0) ? " FRAMEBITS" : "";
    String height = ((flags & HEIGHT) != 0) ? " HEIGHT" : "";
    String properties = ((flags & PROPERTIES) != 0) ? " PROPERTIES" : "";
    String somebits = ((flags & SOMEBITS) != 0) ? " SOMEBITS" : "";
    String width = ((flags & WIDTH) != 0) ? " WIDTH" : "";
    System.out.println("JQCanvas::whichBits on update:" + abort + allbits + error + framebits + height +
                        properties + somebits + width);
  }


  public boolean imageUpdate (Image  img, int  infoflags,
                              int  x, int  y, int  width, int  height)
  {
//    System.out.println("JQCanvas imageUpdate entry.");
    super.imageUpdate(img, infoflags, x, y, width, height);
    
//    whichBits(infoflags);
    
    if ((infoflags & ALLBITS) > 0) {
      switch (typeOfImagePreparation_) {
        case COMPUTED:
          break;
        case LOADgIF:
          finishLoadGIF();
          break;
        case LOADpNG:
          break;
        default:
          break;
      }
      return false; // No more information needed.
    } 
    return true; // Still want to be notified when image is complete.
  }  
  
  public boolean loadGIF(String pathname, FractalSpec args, RGBxSizeSequence RGBxSizeSequence)
    // This is called from the JQFrame.loadArgsThenGIF, this version for pathname
  {
//    System.out.println("JQCanvas loadGIF(string) entry 1.");
    spec_ = args;
    if (true) {
      int bound = (int)RGBxSizeSequence.byteTable.length / 3;
      if (bound < colors)
        jQCentral_.message("JQCanvas::loadGIF only got " + bound + " colors.", true);
      for (int i = 0; i < bound; i++) {
        int ix3 = i * 3;
        reds[i] = RGBxSizeSequence.byteTable[ix3];
        greens[i] = RGBxSizeSequence.byteTable[ix3 + 1];
        blues[i] = RGBxSizeSequence.byteTable[ix3 + 2];
      }
      setColorMapFromRGB(reds, greens, blues);
      updateColorModel();
    }
    
    typeOfImagePreparation_ = LOADgIF;
    refresher_ = getToolkit().getImage(pathname);
    boolean prepared_p = prepareImage(refresher_, this); //Loads image & calls imageUpdate.
    if (prepared_p) {
      //If loaded returns true, else startProductionFinished is called.
      finishLoadGIF();
    }
    return true;
  }
  
  public boolean loadGIF(URL url, FractalSpec args, RGBxSizeSequence RGBxSizeSequence)
    // This is called from the ImageSaver, this version for URL
  {
//    System.out.println("JQCanvas loadGIF(url) entry.");
    spec_ = args;
    int bound = (int)RGBxSizeSequence.byteTable.length / 3;
    if (bound < colors)
      jQCentral_.message("JQCanvas::loadGIF only got " + bound + " colors.", true);
    // We restore 0th color map entry for now to retrieve 0th pixel.
    for (int i = 0; i < bound; i++) {
      int ix3 = i * 3;
      reds[i] = RGBxSizeSequence.byteTable[ix3];
      greens[i] = RGBxSizeSequence.byteTable[ix3 + 1];
      blues[i] = RGBxSizeSequence.byteTable[ix3 + 2];
    }
    setColorMapFromRGB(reds, greens, blues);
    updateColorModel();
    
    typeOfImagePreparation_ = LOADgIF;
    refresher_ = getToolkit().getImage(url);
    if (prepareImage(refresher_, this)) {   //Loads image & calls imageUpdate unless true
      //If loaded returns true, else startProductionFinished is called.
      finishLoadGIF();
    }
    return true;
  }
  
  private void finishLoadGIF ()
    // This is called after the GIF image, refresher, is finished loading.
  {
//    System.out.println("JQCanvas finishLoadGIF entry.");
    typeOfImagePreparation_ = COMPUTED;
    parameter_change_p_ = false;
    ever_drawn_p = true;
    stop_requires_new_image = false; // So we don't try to compute one in paint().
//    Rectangle canvas_size = bounds();
//    h_offset_ = (int)Math.max(0, (canvas_size.width - spec_.pix_width) / 2);
//    v_offset_ = (int)Math.max(0, (canvas_size.height - spec_.pix_height) / 2);
   
    // Check sizes.
    int w = refresher_.getWidth(this);
    int h = refresher_.getHeight(this);
    if (h != spec_.pix_height || w != spec_.pix_width) {
      jQCentral_.message("JQCanvas quitting: h = " + h + ", w = " + w + 
                         ", arg.pix_width = " + spec_.pix_width +
                         ", arg.pix_height = " + spec_.pix_height, true);
       ever_drawn_p = false;
       fractalProducer_ = null;
       return;
    } 
    jQCentral_.updateColorMapEd(false, this);
    updateColorModel();
    fractalProducer_ = null; // Tells paint() to call createDummyProducer.
    repaint(); // So pixels can be grabbed, Q1ImageProducer constructor called here.
  }
  
  private void createDummyProducer ()
    // One cannot grab pixels from an image before it is drawn.  Since paint is called in 
    // its own process and does not notify anyone when it is finished, this is called from
    // the paint method when fractalProducer_ is null.
  {
//    System.out.println("JQCanvas createDummyProducer entry.");
    System.gc();
    fractalProducer_ = createImageProducer(refresher_, spec_);
    // fractalProducer_ has restored all the pixel values at this point by calling
    // colorArrayToIndexArray below.  Now restore unused value to transp pixel.
    if (spec_.version == 1.4) {
      reds[0] = greens[0] = blues[0] = (byte)1;
    }
    setColorMapFromRGB(reds, greens, blues); // Othwise we're out of sync.
    updateColorModel();
    
    // Experiment on ColorModel
    if (false) {
      Class icmClass = null;
      try {
        icmClass = Class.forName("IndexColorModel");
      } catch (Exception e) { icmClass = null;}
      if (icmClass == getColorModel().getClass()) {
        IndexColorModel myColorModel = (IndexColorModel)getColorModel();  // ******** UNSAFE *********
        int size = (int)Math.min(colors, myColorModel.getMapSize());
        if (colors != size || reds == null) {
          reds = new byte[size];
          greens =  new byte[size];
          blues = new byte[size];
        }
        myColorModel.getReds(reds);
        myColorModel.getGreens(greens);
        myColorModel.getBlues(blues);
        setColorMapFromRGB(reds, greens, blues); // sets the variable color_map_
        updateColorModel(); // sets the variable colorModel.
      }
    }
    
    typeOfImagePreparation_ = COMPUTED; 
  }
  
  public void colorArrayToIndexArray (int [] colors, byte [] indices)
    // finds each color of COLORS in our reds, greens, blues arrays and sets 
    // corresponding index of INDICES.  Both parameters are pixel sets.
    // This is used when we read in a GIF image.  COLORS comes from the PixelGrabber
  {    
    int index = 0;
    int searchStart = 0;
    int errors = 0;
    if (spec_.version == 1.4) {
      // Find the color.  Don't want to finish loop.  Skip 0 since phoney anyway.
      indices[0] = 0;
      index = 1;
      searchStart = 1;
    }
    for (; index < colors.length; index++) {
      byte r = (byte)((colors[index] >> 16) & 0xFF);
      byte g = (byte)((colors[index] >> 8) & 0xFF);
      byte b = (byte)((colors[index]) & 0xFF);
      int search;  // We don't want the transparent 0 index, so start search at 1.
      for (search = searchStart; search < reds.length; ++search) {
        if (reds[search] == r && greens[search] == g && blues[search] == b) {
          indices[index] = (byte)search;
          break;
        }
      }
      if (search >= reds.length) {
        if (++errors <= 30) {
          jQCentral_.message("JQCanvas::colorArrayToIndexArray failed to find (" + r + ", " +
                              g + ", " + b + ") at index=" +  index, true);
        }
        if (spec_.version == 1.4) {
          indices[index] = 1; //Paint it black.
        } else {
          indices[index] = (byte)-1; //Paint it black.  -1 is 255 as a byte.
        }
      }
    }
    if (errors > 30) {
      jQCentral_.message("JQCanvas::colorArrayToIndexArray failed... suppressed " +
                         (errors - 30) + " times.", true);
    }
  }
  
  class TransImage implements Transferable
  {
    public DataFlavor[] getTransferDataFlavors()
    {
       /*Returns an array of DataFlavor objects indicating the flavors the data can be provided in. The array should be ordered according to preference for providing the
       data (from most richly descriptive to least descriptive). 

       Returns: 
              an array of data flavors in which this data can be transferred */
       DataFlavor[] answer = new DataFlavor[0];
       return answer;
    }
   

    public boolean isDataFlavorSupported(DataFlavor flavor)
    {
       /*Returns whether or not the specified data flavor is supported for this object. 

       Parameters: 
              flavor - the requested flavor for the data 
       Returns: 
              boolean indicating wjether or not the data flavor is supported */ 
       return false;
    } 

    public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException
    {
       /*Returns an object which represents the data to be transferred. The class of the object returned is defined by the representation class of the flavor. 

       Parameters: 
              flavor - the requested flavor for the data 
       Throws: IOException 
              if the data is no longer available in the requested flavor. 
       Throws: UnsupportedFlavorException 
              if the requested data flavor is not supported. 
       See Also: 
              getRepresentationClass */
       return refresher_;
    }
  }
}

