package JavaQuat20;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import JavaQuat20.GIFDecoderPkg.*;
import JavaQuat20.GIFEncoderPkg.*;
import java.awt.image.IndexColorModel;
import java.awt.image.DirectColorModel;
import java.io.*;
import java.net.URL;
import java.net.*;
import JavaQuat20.FractalLoops.*;

class JQFrame extends Frame
{
  final static String titleString_ = "JQ Drawing Window ";
  static String lastComment_ = "";

  private JQCentral jQCentral_;
  private JQFrame self_;
  private JQCanvas canvas_;
  private FractalSpec spec_;
  private FractalSpec savedSequencePauseSpec_;
  private int myFrameNumber_;
  private Vector loopCheckBoxes_ = new Vector(35);
  private Dimension hVRatio_ = null;
  private boolean seqInProgressP_ = false;
  private boolean seqPausedP_ = false;
  private boolean seqPaused4ScreenCaptureP_ = false;
  private int lastCommandWasAdjustCREDIT_ = 0; // stops frame activation from setting above false.
  private SeqDialog seqDialog_ = null;
  private SeqSpec seqSpec_ = null;  //Shared with the SeqDialog for this frame.
  
  private int seqFrameNumber_; // For Step/StepAgain in Sequence menu.

    
  /* ------ Settings Menu Items ------ */
  private CheckboxMenuItem fastMI_;
  private CheckboxMenuItem c256MI_;
  private CheckboxMenuItem Q1MI_;
  private CheckboxMenuItem Q2MI_;
  private CheckboxMenuItem Q3MI_;
  private CheckboxMenuItem ppp1MI_;
  private CheckboxMenuItem ppp2MI_;
  private CheckboxMenuItem ppp4MI_;
  private CheckboxMenuItem ppp9MI_;
  private CheckboxMenuItem keepColorsMI_; 
  
  /* ------ Loop Menu Items & Subitems ------ */
  private MenuItem LoopLabelMI_; 
  private MenuItem startSeqMI_;
 
  private CheckboxMenuItem userCommentMI_;
  private CheckboxMenuItem write2URLMI_;
  private CheckboxMenuItem strSeqMI_;
  private CheckboxMenuItem endSeqMI_;
  
  private CheckboxMenuItem loggingMI_;
  private CheckboxMenuItem colorMapFromTrapP_;

  
  public JQFrame (JQCentral parent, int width, int height)
  {
    super(titleString_ + parent.frameCount());
    myFrameNumber_ = parent.incFrameCount();
    jQCentral_ = parent;
    spec_ = new FractalSpec();
    hVRatio_ = jQCentral_.getHVRatio();
    canvas_ = new JQCanvas(jQCentral_, this, myFrameNumber_, spec_);
    seqSpec_ = new SeqSpec();
    
    self_ = this;
    
    setLayout(new BorderLayout(0, 0));
    add("Center", canvas_);
    
    setSize(width, height);
    
    //addMouseListener(new MyMouseLNR());
    addWindowListener(new MyWindowLNR());
    
    MenuBar menuBar = new MenuBar();
    setMenuBar(menuBar);
    
    /* --------------- File Menu ---------------- */
    Menu fileMenu = new Menu("File", true); // true for tearoff.
    
    MenuItem newFrameMI = new MenuItem("New Window");
    newFrameMI.addActionListener(new newFrameLNR());
    fileMenu.add(newFrameMI);
    
    fileMenu.addSeparator();
    
    MenuItem openFileMI = new MenuItem("Open JQ Image...");
    openFileMI.addActionListener(new openFileLNR());
    fileMenu.add(openFileMI);
    
    MenuItem openParmMI = new MenuItem("Open JQ Parms...");
    openParmMI.addActionListener(new openParmLNR());
    fileMenu.add(openParmMI);
    
    MenuItem openLogMI = new MenuItem("Open Log Entry...");
    openLogMI.addActionListener(new openLogLNR());
    fileMenu.add(openLogMI);
    
    MenuItem trapFileMI = new MenuItem("Open Trap GIF...");
    trapFileMI.addActionListener(new trapFileLNR());
    fileMenu.add(trapFileMI);
    
    fileMenu.addSeparator();
    
    MenuItem saveAsFileMI = new MenuItem("Save Image...");
    saveAsFileMI.addActionListener(new saveAsFileLNR());
    fileMenu.add(saveAsFileMI);
    
    MenuItem saveParamsFileMI = new MenuItem("Save Params...");
    saveParamsFileMI.addActionListener(new saveArgsFLNR());
    fileMenu.add(saveParamsFileMI);
    
    fileMenu.addSeparator();
    
    loggingMI_ = new CheckboxMenuItem("Logging");
    loggingMI_.addItemListener(new loggingLNR());
    loggingMI_.setState(false);
    fileMenu.add(loggingMI_);
    
    write2URLMI_ = new CheckboxMenuItem("IO to URL?");
    write2URLMI_.addItemListener(new write2URLLNR());
    write2URLMI_.setState(false);
    fileMenu.add(write2URLMI_);
    
    colorMapFromTrapP_ = new CheckboxMenuItem("Set Colors From Trap");
    colorMapFromTrapP_.addItemListener(new ClrFmTrap());
    colorMapFromTrapP_.setState(true);
    fileMenu.add(colorMapFromTrapP_);
      
    userCommentMI_ = new CheckboxMenuItem("Comment on Save?");
    userCommentMI_.addItemListener(new usrComLNR());
    userCommentMI_.setState(false);
    fileMenu.add(userCommentMI_);
    
    fileMenu.addSeparator();
    
    //MenuItem gcMI = new MenuItem("Collect Garbage");
    //gcMI.addActionListener(new gcLNR());
    //fileMenu.add(gcMI);

    MenuItem closeMI = new MenuItem("Close");
    closeMI.addActionListener(new closeLNR());
    fileMenu.add(closeMI);

    MenuItem quitMI = new MenuItem("Quit");
    quitMI.addActionListener(new quitLNR());
    fileMenu.add(quitMI);
    
    CheckboxMenuItem debugMI= new CheckboxMenuItem("Debug");
    debugMI.addItemListener(new debugLNR());
    debugMI.setState(false);
    //fileMenu.add(debugMI);   // COMMENT OUT THIS LINE FOR USE BY OTHERS.
    
    // newFrameLNR openFileLNR openUrlLNR saveAsFileLNR saveParamsLNR closeLNR quitLNR
    
    /* --------------- Draw Menu ---------------- */
    Menu drawMenu = new Menu("Draw", true); // true for tearoff.
    
    MenuItem startDrawMI = new MenuItem("Start");
    startDrawMI.addActionListener(new startDrawLNR());
    drawMenu.add(startDrawMI);
    
    MenuItem stopDrawMI = new MenuItem("Pause");
    stopDrawMI.addActionListener(new stopDrawLNR());
    drawMenu.add(stopDrawMI);
    
    MenuItem clipBdMI = new MenuItem("Copy");
    clipBdMI.addActionListener(new ClipLNR());
    //drawMenu.add(clipBdMI);
    
    // startDrawLNR stopDrawLNR ClipLNR
    
     /* --------------- Windows Menu ---------------- */
    Menu windowsMenu = new Menu("Windows", true); // true for tearoff.
    
    MenuItem parmsMI = new MenuItem("Parameters Window");
    parmsMI.addActionListener(new parmsLNR());
    windowsMenu.add(parmsMI);
    
    MenuItem colorsMI = new MenuItem("Color Editor Window");
    colorsMI.addActionListener(new colorsLNR());
    windowsMenu.add(colorsMI);
        
    MenuItem seqMI = new MenuItem("Sequence Setup Window");
    seqMI.addActionListener(new seqLNR());
    windowsMenu.add(seqMI);
    
    MenuItem msgMI = new MenuItem("Messages Window");
    msgMI.addActionListener(new msgLNR());
    windowsMenu.add(msgMI);
    
    MenuItem trapMI = new MenuItem("Trap GIF Window");
    trapMI.addActionListener(new TrapLNR());
    windowsMenu.add(trapMI);
    
     /* --------------- Settings Menu ---------------- */
    Menu settingsMenu = new Menu("Settings", true); // true for tearoff.
    
    fastMI_ = new CheckboxMenuItem("Draw Fast?");
    //fastMI_.addItemListener(new fastLNR());
    fastMI_.setState(true);
    settingsMenu.add(fastMI_);
    
    c256MI_ = new CheckboxMenuItem("256 Colors?");
    //c256MI_.addItemListener(new c256LNR());
    c256MI_.setState(true);
    settingsMenu.add(c256MI_);
    
    settingsMenu.addSeparator();
    
    Menu scanOnMenu = new Menu("Scan On", true); // true for tearoff.
    settingsMenu.add(scanOnMenu);
    
      Q1MI_ = new CheckboxMenuItem("Q1");
      Q1MI_.addItemListener(new soQ1LNR());
      Q1MI_.setState(true);
      scanOnMenu.add(Q1MI_);
      
      Q2MI_ = new CheckboxMenuItem("Q2");
      Q2MI_.addItemListener(new soQ2LNR());
      Q2MI_.setState(false);
      scanOnMenu.add(Q2MI_);
      
      Q3MI_ = new CheckboxMenuItem("Q3");
      Q3MI_.addItemListener(new soQ3LNR());
      Q3MI_.setState(false);
      scanOnMenu.add(Q3MI_);
    
    Menu ptsPerPixelMenu = new Menu("Pts/Pixel", true); // true for tearoff.
    settingsMenu.add(ptsPerPixelMenu);
    
      ppp1MI_ = new CheckboxMenuItem("1");
      ppp1MI_.addItemListener(new ppp1LNR());
      ppp1MI_.setState(true);
      ptsPerPixelMenu.add(ppp1MI_);
      
      ppp2MI_ = new CheckboxMenuItem("2");
      ppp2MI_.addItemListener(new ppp2LNR());
      ppp2MI_.setState(false);
      ptsPerPixelMenu.add(ppp2MI_);
      
      ppp4MI_ = new CheckboxMenuItem("4");
      ppp4MI_.addItemListener(new ppp4LNR());
      ppp4MI_.setState(false);
      ptsPerPixelMenu.add(ppp4MI_);
    
      ppp9MI_ = new CheckboxMenuItem("9");
      ppp9MI_.addItemListener(new ppp9LNR());
      ppp9MI_.setState(false);
      ptsPerPixelMenu.add(ppp9MI_);
      
    MenuItem hVRatioMI = new MenuItem("W/H Ratio...");
    hVRatioMI.addActionListener(new HVRatioLNR());
    settingsMenu.add(hVRatioMI);
    
    settingsMenu.addSeparator();

//    keepColorsMI_ = new CheckboxMenuItem("Transfer Color Map");
//    keepColorsMI_.addItemListener(new keepColorsLNR());
//    keepColorsMI_.setState(false);
//    settingsMenu.add(keepColorsMI_);
        
    MenuItem displayParmsMI = new MenuItem("Show Arguments");
    displayParmsMI.addActionListener(new displayParmsLNR());
    settingsMenu.add(displayParmsMI);
    
     /* --------------- Loop Menu ---------------- */
     
    Menu loopMenu = new Menu("Loop", true); // true for tearoff.
    
    LoopLabelMI_ = new MenuItem("CpxPower2");
    loopMenu.add(LoopLabelMI_);

    loopMenu.addSeparator();

    Menu currentMenu = new Menu("Complex Powers (CPX)", true); // true for tearoff.
    loopMenu.add(currentMenu);
    
      CheckboxMenuItem defaultLoop = newLoopCheckbox("CpxPower2");
      defaultLoop.setState(true);
      currentMenu.add(defaultLoop);
      currentMenu.add(newLoopCheckbox("CpxPower3"));
      currentMenu.add(newLoopCheckbox("CpxPower4"));
      currentMenu.add(newLoopCheckbox("Cpx8"));
      currentMenu.add(newLoopCheckbox("CpxN"));
      currentMenu.add(newLoopCheckbox("Cpx2TrapGIF"));
      currentMenu.add(newLoopCheckbox("Cpx3TrapGIF"));
      currentMenu.add(newLoopCheckbox("CpxNTrapGIF"));
      currentMenu.add(newLoopCheckbox("Cpx2InvTrapGIF"));
      
    currentMenu = new Menu("Phoenix", true); // true for tearoff.
    loopMenu.add(currentMenu);
    
      currentMenu.add(newLoopCheckbox("PhoenixDepth"));
      currentMenu.add(newLoopCheckbox("Phoenix"));
      currentMenu.add(newLoopCheckbox("PhoenixTrapGIF"));
    
    currentMenu = new Menu("Quaternion (QAT)", true); // true for tearoff.
    loopMenu.add(currentMenu);
    
    currentMenu = new Menu("HyperComplex (HCX)", true); // true for tearoff.
    loopMenu.add(currentMenu);
    
      currentMenu.add(newLoopCheckbox("HcxPower2"));
    
    currentMenu = new Menu("Tetran (TET)", true); // true for tearoff.
    loopMenu.add(currentMenu);
    
      currentMenu.add(newLoopCheckbox("TetraPower2"));
      currentMenu.add(newLoopCheckbox("TetraPower3"));
      currentMenu.add(newLoopCheckbox("TetraPower4"));
      
    currentMenu = new Menu("Miscelaneous (MSC 3D)", true); // true for tearoff.
    loopMenu.add(currentMenu);
    
      currentMenu.add(newLoopCheckbox("Loop32"));
      currentMenu.add(newLoopCheckbox("Loop321"));
      currentMenu.add(newLoopCheckbox("Loop322"));
      currentMenu.add(newLoopCheckbox("Loop323"));
      currentMenu.add(newLoopCheckbox("Loop324"));
      currentMenu.add(newLoopCheckbox("Loop322k"));
      
    currentMenu = new Menu("Miscelaneous (MSC 4D)", true); // true for tearoff.
    loopMenu.add(currentMenu);
    
      currentMenu.add(newLoopCheckbox("LoopTetraR"));
      currentMenu.add(newLoopCheckbox("LoopTetraP"));
      currentMenu.add(newLoopCheckbox("LoopTetraN"));
      currentMenu.add(newLoopCheckbox("LoopTetraRN"));
      currentMenu.add(newLoopCheckbox("LoopTetraRP"));
      currentMenu.add(newLoopCheckbox("LoopTetraNP"));
      currentMenu.add(newLoopCheckbox("Loop429"));
      currentMenu.add(newLoopCheckbox("Loop42A"));
      currentMenu.add(newLoopCheckbox("Loop42B"));
      currentMenu.add(newLoopCheckbox("Loop42C"));
      currentMenu.add(newLoopCheckbox("Loop448"));
      
     /* --------------- Adjust Menu ---------------- */
  // EnlargeLNR EnlrgAgnLNR RightLNR RightAgnLNR UpLNR UpAgnLNR
  
    Menu adjustMenu = new Menu("Adjust", true); // true for tearoff.
    
    MenuItem enlargeMI_ = new MenuItem("Zoom...");
    enlargeMI_.addActionListener(new EnlargeLNR());
    adjustMenu.add(enlargeMI_);

    MenuItem enlrgAgnMI_ = new MenuItem("Zoom again");
    enlrgAgnMI_.addActionListener(new EnlrgAgnLNR());
    adjustMenu.add(enlrgAgnMI_);

    adjustMenu.addSeparator();

    MenuItem rightMI_ = new MenuItem("Right shift...");
    rightMI_.addActionListener(new RightLNR());
    adjustMenu.add(rightMI_);

    MenuItem rightAgnMI_ = new MenuItem("Right shift again");
    rightAgnMI_.addActionListener(new RightAgnLNR());
    adjustMenu.add(rightAgnMI_);

    MenuItem downMI_ = new MenuItem("Down shift...");
    downMI_.addActionListener(new DownLNR());
    adjustMenu.add(downMI_);

    MenuItem downAgnMI_ = new MenuItem("Down shift again");
    downAgnMI_.addActionListener(new DownAgnLNR());
    adjustMenu.add(downAgnMI_);

    adjustMenu.addSeparator();

    MenuItem onWMI_ = new MenuItem("W axis shift...");
    onWMI_.addActionListener(new OnWLNR());
    adjustMenu.add(onWMI_);

    MenuItem onWAgnMI_ = new MenuItem("W axis shift again");
    onWAgnMI_.addActionListener(new OnWAgnLNR());
    adjustMenu.add(onWAgnMI_);
    
 // enlargeMI_ enlrgAgnMI_ rightMI_ rightAgnMI_ downMI_ downAgnMI_ onWMI_ onWAgnMI_


    /* --------------- Axis Menu ---------------- */
    Menu axisMenu = new Menu("Axis", true); // true for tearoff.
    
    MenuItem swapHVMI = new MenuItem("H <=> V");
    swapHVMI.addActionListener(new swapHVLNR());
    axisMenu.add(swapHVMI);
    
    MenuItem rot90LeftMI = new MenuItem("Rot. 90 left");
    rot90LeftMI.addActionListener(new Rotate90LLNR());
    axisMenu.add(rot90LeftMI);
    
    MenuItem rot90RightMI = new MenuItem("Rot. 90 right");
    rot90RightMI.addActionListener(new Rotate90RLNR());
    axisMenu.add(rot90RightMI);
    
    axisMenu.addSeparator();
    
    MenuItem w2HMI = new MenuItem("W => H");
    w2HMI.addActionListener(new W2HLNR());
    axisMenu.add(w2HMI);

    MenuItem w2VMI = new MenuItem("W => V");
    w2VMI.addActionListener(new W2VLNR());
    axisMenu.add(w2VMI);

    axisMenu.addSeparator();
    
    MenuItem h2MemHMI = new MenuItem("H => Mem H");
    h2MemHMI.addActionListener(new H2MemHLNR());
    axisMenu.add(h2MemHMI);
    
    MenuItem v2MemVMI = new MenuItem("V => Mem V");
    v2MemVMI.addActionListener(new V2MemVLNR());
    axisMenu.add(v2MemVMI);
    
    MenuItem memH2HMI = new MenuItem("Mem H => H");
    memH2HMI.addActionListener(new MemH2HLNR());
    axisMenu.add(memH2HMI);
    
    MenuItem memV2VMI = new MenuItem("Mem V => V");
    memV2VMI.addActionListener(new MemV2VLNR());
    axisMenu.add(memV2VMI);
    
    axisMenu.addSeparator();
    
    MenuItem scaleVMI = new MenuItem("Scale V-axis");
    scaleVMI.addActionListener(new ScaleVLNR());
    axisMenu.add(scaleVMI);
    
    MenuItem mouseV2MI = new MenuItem("Mouse V-axis");
    mouseV2MI.addActionListener(new mouseV2LNR());
    axisMenu.add(mouseV2MI);
    
    MenuItem mouseH2MI = new MenuItem("Mouse H-axis");
    mouseH2MI.addActionListener(new mouseH2LNR());
    axisMenu.add(mouseH2MI);
    
     /* --------------- Sequence Menu ---------------- */
    
    Menu sequenceMenu = new Menu("Sequence", true); // true for tearoff.
    
    startSeqMI_ = new MenuItem("Start");
    startSeqMI_.addActionListener(new startSeqLNR());
    sequenceMenu.add(startSeqMI_);

   sequenceMenu.addSeparator();

    MenuItem seq2MI = new MenuItem("Sequence Setup...");
    seq2MI.addActionListener(new seqLNR());
    sequenceMenu.add(seq2MI); // Defined under window menu.
    
    sequenceMenu.addSeparator();

    CheckboxMenuItem reversePanelsMI = new CheckboxMenuItem("Reverse order panels.");
    reversePanelsMI.addItemListener(new RevPanelsLNR());
    reversePanelsMI.setState(false);
    sequenceMenu.add(reversePanelsMI);
        
    CheckboxMenuItem rowByRowMI = new CheckboxMenuItem("Row by row.");
    rowByRowMI.addItemListener(new RowByRowLNR());
    rowByRowMI.setState(true);
    sequenceMenu.add(rowByRowMI);
    
    sequenceMenu.addSeparator();

    strSeqMI_ = new CheckboxMenuItem("Set start to current image");
    strSeqMI_.addItemListener(new StrSeqLNR());
    strSeqMI_.setState(false);
    sequenceMenu.add(strSeqMI_);    

    endSeqMI_ = new CheckboxMenuItem("Set end to current image");
    endSeqMI_.addItemListener(new EndSeqLNR());
    endSeqMI_.setState(false);
    sequenceMenu.add(endSeqMI_);    

    MenuItem stepSeqMI_ = new MenuItem("Set Step...");
    stepSeqMI_.addActionListener(new StepLNR());
    sequenceMenu.add(stepSeqMI_);

//    MenuItem stepAgainSeqMI_ = new MenuItem("Set Step Again");
//    stepAgainSeqMI_.addActionListener(new StepAgnLNR());
//    sequenceMenu.add(stepAgainSeqMI_);
    
       /* --------- Speed Keys and MenuBar contents ----------- */

    strSeqMI_.setShortcut(new MenuShortcut(KeyEvent.VK_1));
    endSeqMI_.setShortcut(new MenuShortcut(KeyEvent.VK_2));
    startSeqMI_.setShortcut(new MenuShortcut(KeyEvent.VK_A));
    seq2MI.setShortcut(new MenuShortcut(KeyEvent.VK_B));
    startDrawMI.setShortcut(new MenuShortcut(KeyEvent.VK_D));
    colorsMI.setShortcut(new MenuShortcut(KeyEvent.VK_E));
    msgMI.setShortcut(new MenuShortcut(KeyEvent.VK_M));
    newFrameMI.setShortcut(new MenuShortcut(KeyEvent.VK_N));
    //stepSeqMI_.setShortcut(new MenuShortcut(KeyEvent.VK_N, true));
    openFileMI.setShortcut(new MenuShortcut(KeyEvent.VK_O));
    stopDrawMI.setShortcut(new MenuShortcut(KeyEvent.VK_P));
    quitMI.setShortcut(new MenuShortcut(KeyEvent.VK_Q));
    hVRatioMI.setShortcut(new MenuShortcut(KeyEvent.VK_R));
    saveAsFileMI.setShortcut(new MenuShortcut(KeyEvent.VK_S));
    trapFileMI.setShortcut(new MenuShortcut(KeyEvent.VK_T));
    closeMI.setShortcut(new MenuShortcut(KeyEvent.VK_W));
    stepSeqMI_.setShortcut(new MenuShortcut(KeyEvent.VK_ADD)); //Comes out as K, not +
    parmsMI.setShortcut(new MenuShortcut(KeyEvent.VK_EQUALS));

    
    menuBar.add(fileMenu);
    menuBar.add(drawMenu);
    menuBar.add(settingsMenu);
    menuBar.add(adjustMenu);
    menuBar.add(axisMenu);
    menuBar.add(windowsMenu);
    
    menuBar.add(sequenceMenu);
    menuBar.add(loopMenu);

    
      /* --------------- End of Menus ---------------- */
    
//    if (jQCentral_.frameCount() <= 6) {
//      jQCentral_.setParameters(canvas_, canvas_.spec());
//      setCheckBoxValues(canvas_.spec());
//    } else {
//      jQCentral_.continuePreviousParameters(this, spec_);
//    }
    
    setBackground(Color.lightGray);
    show();
    //setVisible(true);
  }
  
  private CheckboxMenuItem newLoopCheckbox (String name)
  {
    CheckboxMenuItem cmi = new CheckboxMenuItem(name);
    cmi.addItemListener(new loopLNR(name));
    cmi.setState(false);
    loopCheckBoxes_.addElement(cmi);
    return cmi;
  }
      
  /* ------------------- UTILITIES ------------------------- */
  
  public JQCanvas canvas ()
  {
    return canvas_;
  }
  
  public void draw ()
  { 
    if (seqInProgressP_) {
      jQCentral_.message("There is a sequence in progress.  You must finish or pause it first.",
                         true);
      return;
    }
    if (canvas_.drawing_p()) {
      jQCentral_.message("There is an image in progress.  You must finish or pause it first.",
                         true);
      return;
    }
    FractalSpec newSpec = new FractalSpec(); //Used to be copy of JQCentral.centralSpec
    if (jQCentral_.parmsToSpec(newSpec)) {
      getCheckBoxValues(newSpec);
      newSpec.which = jQCentral_.translateLoopName(newSpec.which);
      canvas_.setParameters(newSpec); // This sets spec_
      if (seqPausedP_) {
        seqPausedP_ = false;
        startSeqMI_.setLabel("Start");
      }
      canvas_.threaded_paint_fractal();
    } else {
      jQCentral_.message("Invalid parameters, please correct and try again.", true);
    }
  }

  public boolean doingSequenceP ()
    // This used so we ignore mouse clicks in canvas.
  {
    return seqInProgressP_ || seqPausedP_;
  }
  
//  public void activate ()
//  {
//    validate();
//    setVisible(true);
//    jQCentral_.setHVRatio(hVRatio_);
//    jQCentral_.setActiveCanvas(canvas_);
//  }

  
  public Dimension getHVRatio ()
  {
    return hVRatio_;
  }
  
  public String getUserComment ()
  {
    return (new TextDialog("Add Comments", this, lastComment_).getText());
  }

  
  private boolean writeGIF (boolean is_url_p, String pathname)
    // Return true if the write was successful.
  {
    boolean interlace_p = false;
    GIFEncoder gifEncoder = null;
    JQIO outUtility = new JQIO(jQCentral_, this);
    OutputStream stream1 = null;
    String userComment = "";
    
    // Get the pathname
    if (!canvas_.imageExistsP()) {
      jQCentral_.message("There is no image.", false);
      return false;
    }
    if (pathname == null) {
      if (is_url_p) {
        pathname = outUtility.urlStringFromDialog();
      } else {
        pathname = outUtility.pathFromDialog(false); // false => use write dialog
      }
    }
    if (pathname == null || pathname.length() == 0) {
      jQCentral_.message("Write GIF image - No action taken.", false);
      return false;
    }

    if (jQCentral_.userWantsToAddCOmmentP()) {
      userComment = getUserComment();
    }
        
    // Save TRAP info in comment if appropriate.
    JQLooper whichInstance = jQCentral_.findLoopObject(canvas_.spec().which);
    if (whichInstance != null && whichInstance.acceptsTrapGif() &&
        jQCentral_.trapPathname() != null) {
      userComment = ("\nTrap:" + jQCentral_.trapPathname() + "\n" + userComment);
    }

    try {
      stream1 = outUtility.toOutputStream(pathname, is_url_p, false);
      if (stream1 == null)
        throw new IOException("NOT FOUND: " + pathname);
      GIFEncoder gifWriter = canvas_.createGIFEncoder();
      FractalSpec alteredArgs = new FractalSpec(canvas_.spec());
      alteredArgs.colorsIs256_p = true;
      gifWriter.JavaQuatWrite(stream1, alteredArgs.toString(1, userComment), 
                              interlace_p, alteredArgs.version);
      //jQCentral_.message("Image Saved.  Format = " + alteredArgs.version, false);
      jQCentral_.message("Image Saved.", false);

    } catch (Throwable e) {
      jQCentral_.message("Error on stream or encode: " + e,true);
      jQCentral_.message(e.getMessage() == null ? e.toString() : e.getMessage(), true);
      return false;
    } finally {
      try {
        if (stream1 != null)
          stream1.flush();
          stream1.close();
      } catch (Throwable e) {}
    }
    return true;
  }
  
  public boolean writeParms (boolean is_url_p, String pathname)
  {
    JQIO outUtility = new JQIO(jQCentral_, this);
    OutputStream stream1 = null;
    String userComment = "";
    
    // Get the pathname
    if (!canvas_.imageExistsP()) {
      jQCentral_.message("There is no image for which to save arguments.", false);
      return false;
    }
    if (pathname == null) {
      if (is_url_p) {
        pathname = outUtility.urlStringFromDialog();
      } else {
        pathname = outUtility.pathFromDialog(false); // false => use write dialog
      }
    }
    if (pathname == null || pathname.length() == 0) {
      jQCentral_.message("Write arguments - No action taken.", false);
      return false;
    }
    
    if (jQCentral_.userWantsToAddCOmmentP()) {
      userComment = getUserComment();
    }

    try {
      stream1 = outUtility.toOutputStream(pathname, is_url_p, false);
      OutputStreamWriter strStream = new OutputStreamWriter(stream1);
      String answer = spec_.toString(1, userComment);
      strStream.write(answer, 0, answer.length());
      strStream.flush();
      strStream.close();
      jQCentral_.message("Parameters Saved.", false);

    } catch (Throwable e) {
      jQCentral_.message("Error on stream or encode: " + e,true);
      jQCentral_.message(e.getMessage() == null ? e.toString() : e.getMessage(), true);
      return false;
    } finally {
      try {
        if (stream1 != null)
          stream1.flush();
          stream1.close();
      } catch (Throwable e) {}
    }
    return true;  
  }
  
  public FractalSpec readParms (boolean is_url_p, int argsNum)
    // Loads an argument set and returns it, setting parmsFrame_.  argsNum is the
    // number of the argument set, -1 means just take the first.
  {
    String pathname = null;
    InputStream stream1 = null;
    FractalSpec answer = null;
    JQIO inUtility = new JQIO(jQCentral_, this);

    // Get the pathname if none was passed in.
    if (pathname == null) {
      if (is_url_p) {
        pathname = inUtility.urlStringFromDialog();
      } else {
        pathname = inUtility.pathFromDialog(true); // true => use read dialog
      }
    }
    if (pathname == null || pathname.length() == 0) {
      jQCentral_.message("No action taken.", false);
      return null;
    }

    try {
      stream1 = inUtility.toInputStream(pathname, is_url_p, false); // false => create stream.
      InputStreamReader streamReader = new InputStreamReader(stream1);
      JQArgReader parmReader = new JQArgReader();
      answer = parmReader.readArgFile(streamReader, argsNum, jQCentral_);
      answer = new FractalSpec(answer);
      if (answer == null) {
        if (argsNum == -1) {
          jQCentral_.message("Could not find an arg set.", false);
        } else {
          jQCentral_.message("Could not find log_id " + argsNum, false);
        }
      } else {
        if (argsNum == -1) {
          jQCentral_.message("First arg set loaded.", false);
        } else {
          jQCentral_.message("Arg set " + argsNum + " loaded.",false);
        }
      }
    } catch (Throwable e) {
      jQCentral_.message(e.getMessage() == null ? e.toString() : e.getMessage(), true);
    } finally {
      if (stream1 != null)
        try {
          stream1.close(); // In case we didn't close it above.
        } catch (Throwable e) {}
    }
    if (answer == null)
      return null;
      
    setCheckBoxValues(answer, false);
    jQCentral_.setParameters(answer);
    
    return answer;
  }

  public FractalSpec loadArgsThenGIF (boolean is_url_p, String pathname)
  {
    boolean SKIPIMAGE = true;
    GIFDecoder gifDecoder = null;
    InputStream stream1 = null;
    FractalSpec args = null;
    JQIO inUtility = new JQIO(jQCentral_, this);
    
    // Get the pathname if none was passed in.
    if (pathname == null) {
      if (is_url_p) {
        pathname = inUtility.urlStringFromDialog();
      } else {
        pathname = inUtility.pathFromDialog(true); // true => use read dialog
      }
    }
    if (pathname == null || pathname.length() == 0) {
      jQCentral_.message("No action taken.", false);
      return null;
    }

    // Here we try to open the stream and extract the arguments only.
    try {
      stream1 = inUtility.toInputStream(pathname, is_url_p, false); // false => create stream.
      if (stream1 == null)
        throw new IOException("NOT FOUND: " + pathname);
      gifDecoder = new GIFDecoder(stream1, SKIPIMAGE);
      String argsString = gifDecoder.getApplicationData("JavaQuat", "014");
      boolean version014P = true;
      if (argsString == null) {
        argsString = gifDecoder.getApplicationData("JavaQuat", "020");
        jQCentral_.message("Style 2.0 image read.", false);
        version014P = false;
      } else {
        jQCentral_.message("Style 1.4 image read.", false);
      }
      if (argsString == null) {
        throw new AWTException("Not recognized as a JavaQuat image.");
      }
      StringReader argsStream = new StringReader(argsString);
      
      JQArgReader reader = new JQArgReader();
      // parent.exitImageSaver gets passed argsReadFromFile_ on exitButton.
      args = new FractalSpec(reader.readArgFile(argsStream, -1, jQCentral_));
      if (version014P) {
        args.version = 1.4;
      }
    } catch (Throwable e) {
      jQCentral_.message((e.getMessage() == null ? e.toString() : e.getMessage()), false);
      // e.printStackTrace();
    } finally {
      if (stream1 != null)
        try {
          stream1.close(); // In case we didn't close it above.
        } catch (Throwable e) {}
    }  
    if (args == null)
      return null;
      
    spec_.assignFrom(args);
    setCheckBoxValues(spec_, false);
    jQCentral_.setParameters(spec_);
    
    // Try to adjust our size.
    int c_width = canvas_.getSize().width;
    int c_height = canvas_.getSize().height;
    int delta_width = spec_.pix_width - c_width;
    int delta_height = spec_.pix_height - c_height;
    if (delta_width > 0 || delta_height > 0) {
      int new_width = (delta_width > 0 ? getSize().width + delta_width : getSize().width);
      int new_height = (delta_height > 0 ? getSize().height + delta_height : getSize().height);
      setSize(new_width, new_height);
      validate();
    }
    
    // Here's where we load the image itself.
//    System.out.println("ImageSaver.loadArgsThenGIF args=" + args + " is_url_p=" +
//                       is_url_p);
    if (is_url_p) {
      try {
        URL url = new URL(pathname);
        canvas_.loadGIF(url, args, gifDecoder.globalRGBTable());
      } catch (Throwable e) {
        jQCentral_.message((e.getMessage() == null ? e.toString() : e.getMessage()), false);
      }
    } else {
      canvas_.loadGIF(pathname, args, gifDecoder.globalRGBTable());
    }
    jQCentral_.message("Image read: " + pathname, false);
    return args; 
  }

  private void setScanOnFalse ()
  {
    Q1MI_.setState(false);
    Q2MI_.setState(false);
    Q3MI_.setState(false);
  } 
  
  private void setPPPFalse ()
  {
    ppp1MI_.setState(false);
    ppp2MI_.setState(false);
    ppp4MI_.setState(false);
    ppp9MI_.setState(false);
  }
  
  public String looper ()
  {
    return spec_.which;
  }
  
  private void setLooper (String which1)
  {
    for(int i = 0; i < loopCheckBoxes_.size(); i++) {
      CheckboxMenuItem box = (CheckboxMenuItem)loopCheckBoxes_.elementAt(i);
      if (box.getLabel().equals(which1))
        box.setState(true);
      else
        box.setState(false);
    }
    LoopLabelMI_.setLabel(which1);
    //spec_.which = which1; - We don't want to change the canvas_ spec.
  }
  
  public void setCheckBoxValues (FractalSpec spec, boolean pass2Central_p)
  // This sets the CheckboxesMenuItems etc. to the proper values for this frame.
  {
    setScanOnFalse();
    if (spec.rangeOver == FractalSpec.CENTER)
      Q1MI_.setState(true);
    else if (spec.rangeOver == FractalSpec.COEFF)
      Q2MI_.setState(true);
    else if (spec.rangeOver == FractalSpec.TRANS)
      Q3MI_.setState(true);
      
    setPPPFalse();
    if (spec.pts_per_pixel == 1)
      ppp1MI_.setState(true);
    else if (spec.pts_per_pixel == 2)
      ppp2MI_.setState(true);
    else if (spec.pts_per_pixel == 4)
      ppp4MI_.setState(true);
    else if (spec.pts_per_pixel == 9)
      ppp9MI_.setState(true);
      
    fastMI_.setState(!spec.hifi_p);
    //System.out.println("JQFrame.setCheckBoxValues fastMI_ = " + fastMI_.getState());
    c256MI_.setState(spec.colorsIs256_p);
    setLooper(spec.which);
//    if (pass2Central_p) {
//      System.out.println("JQFrame.setCheckBoxValues if TRUE.");
//      jQCentral_.setCheckBoxValues4ParmFrame(spec);
//    }
  }

  public void getCheckBoxValues (FractalSpec spec)
  // This copies the CheckboxesMenuItems etc. to the argument spec.
  {
    if (Q1MI_.getState()) {
      spec.rangeOver = FractalSpec.CENTER;
    } else if (Q2MI_.getState()) {
      spec.rangeOver = FractalSpec.COEFF;
    } else if (Q3MI_.getState()) {
      spec.rangeOver = FractalSpec.TRANS;
    }
    
    if (ppp1MI_.getState()) {
      spec.pts_per_pixel = 1;
    } else if (ppp2MI_.getState()) {
      spec.pts_per_pixel = 2;
    } else if (ppp4MI_.getState()) {
      spec.pts_per_pixel = 4;
    } else if (ppp9MI_.getState()) {
      spec.pts_per_pixel = 9;
    } 
      
    spec.hifi_p = !fastMI_.getState();   
    spec.colorsIs256_p = c256MI_.getState();
    spec.which = LoopLabelMI_.getLabel();
  }
  
  public void setSeqSpec (SeqSpec arg, boolean start_too_p, boolean end_too_p)
    // This is what the seqDialog calls when the user clicks OK.
  {
    stopSequence();
    seqSpec_.assignFrom(arg, true, true);
    if (seqSpec_.startSpec() == null) {
      strSeqMI_.setState(false);
    } else {
      strSeqMI_.setState(true);
      jQCentral_.message("First frame in sequence is set.", false);
    }
    if (seqSpec_.endSpec() == null) {
      endSeqMI_.setState(false);
    } else {
      endSeqMI_.setState(true);
      jQCentral_.message("Last frame in sequence is set.", false);
    }
  }
  
  public void autoAdvance (FractalSpec spec)
    // This is called each time our canvas_ finishes a drawing.
  {
    boolean debug = false;
  
    FractalSpec advancee = new FractalSpec(spec);  //After advance needs to be different from spec
                                                   // so it knows to draw again. spec is from canvas_.
    if (seqInProgressP_) {
      // First do the write and see if it worked.
      if (seqSpec_.pause_p) {
        if (debug) System.out.println("JQFrame.autoAdvance entry: pt1");
        // Pause for screen capture.
        seqInProgressP_ = false;
        seqPausedP_ = true;
        seqPaused4ScreenCaptureP_ = true;
        if (seqSpec_.advance(advancee, jQCentral_)) {
          savedSequencePauseSpec_ = advancee;
          //canvas_.setParameters(advancee);
          jQCentral_.setParameters(advancee);
          startSeqMI_.setLabel("Resume");
        } else {
          startSeqMI_.setLabel("Start");
        }
      } else if (!writeGIF(seqSpec_.destinationURLP(), seqSpec_.pathname(spec))) {
        if (debug) System.out.println("JQFrame.autoAdvance entry: pt2");
        // NORMALLY the test writes the file and we skip to next clause.
        //See startSeqLNR.
        seqInProgressP_ = false;     // This causes half restart.
        seqPausedP_ = false;
        startSeqMI_.setLabel("Start");
        seqSpec_.setStartSpec(null); // This causes half restart. 2 halves make a whole.
        // let them write to same place, don't: seqSpec_.enabled_p = false;
        jQCentral_.message("Sequence write failed - stopping", true);
      } else if (seqSpec_.advance(advancee, jQCentral_)) {
        if (debug) System.out.println("JQFrame.autoAdvance entry: pt3");
        // This first line sets canvas to new values. 
        canvas_.setParameters(advancee);
//            jQCentral_.message("JQFrame.autoAdvance 2 canvas advanced.",false);
        jQCentral_.setParameters(advancee);
//            jQCentral_.message("JQFrame.autoAdvance 3 parm-frame advanced.",false);
        canvas_.threaded_paint_fractal();
      } else {
        if (debug) System.out.println("JQFrame.autoAdvance entry: pt4");
        // We did not advance and so reached the end of the sequence.
        seqInProgressP_ = false;
        seqPausedP_ = false;
        startSeqMI_.setLabel("Start");
        seqSpec_.setStartSpec(null);
        seqSpec_.enabled_p = false;
      }
    }
  }
  
  private void stepAdvance (int steps)
    // Called from Sequence menu step listeners below to zoom, pivot or shift.
    // Moves the seqSpec to frame 'steps'
  {
    //seqSpec_.setSteps(steps);
    FractalSpec advancee = new FractalSpec();
    // First get any values not saved in the sequenceSpec.
    if (!jQCentral_.parmsToSpec(advancee)) {
      jQCentral_.message("Unacceptable initial parameter frame values.", true);
    // Second, get the settings of the sequenceSpec.
    } else if (seqSpec_.manualStep(advancee, steps, jQCentral_)) {
      //System.out.println("JQFrame.stepAdvance 1 advancee.range = " + advancee.center.r);
      //canvas_.setParameters(advancee);
      savedSequencePauseSpec_ = advancee;
      jQCentral_.setParameters(advancee);
      //canvas_.set_other_change_requires_redraw_p(true);
      //System.out.println("JQFrame.stepAdvance 2 canvas.range = " + canvas_.spec().center.r);
    } else {
      jQCentral_.message("Unable to go to frame " + steps, true);
    }
  }
  
    // adjustEnlarge adjustRight adjustUp
    
  private FractalSpec getAdjustBaseSpec ()
  {
    FractalSpec spec;
    spec = new FractalSpec();
    if (!jQCentral_.parmsToSpec(spec)) { // Try to get values from parmFrame
      jQCentral_.message("Taking value from image. Parameters window value is invalid.", 
                         true);
      spec.assignFrom(canvas_.spec());   // but if we cannot, use canvas values.
    }
    return spec;
  }

  public void adjustEnlarge (int percent)
  {
    FractalSpec spec = getAdjustBaseSpec();
    //double perPixel = spec.range / spec.pix_width;
    //double newRange = perPixel * Math.max(1,(spec.pix_width - pixels));
    double newRange = (percent >= 0) ? 
                      (spec.range * 100) / (100 + percent) :
                      (spec.range / 100) * (100 + -percent);
    spec.range = newRange;
    jQCentral_.setParameters(spec);
  }
  
  public void adjustCenterRight (int pixels)
  {
    FractalSpec spec = getAdjustBaseSpec();
    double perPixel = spec.range / spec.pix_width;
    //System.out.println("JQFrame.adjustRight perPixel = " + perPixel);
    if (spec.rangeOver == FractalSpec.CENTER) {
      spec.center.plus_equal(spec.vx.times_scalar(pixels * perPixel).bigQuaternionValue());
    } else if (spec.rangeOver == FractalSpec.COEFF) {
      spec.coeff.plus_equal(spec.vx.times_scalar(pixels * perPixel).bigQuaternionValue());
    } else if (spec.rangeOver == FractalSpec.TRANS) {
      spec.trans.plus_equal(spec.vx.times_scalar(pixels * perPixel).bigQuaternionValue());
    }
    jQCentral_.setParameters(spec);
  }
  
  public void adjustCenterUp (int pixels)
  {
    FractalSpec spec = getAdjustBaseSpec();
    double perPixel = spec.range / spec.pix_width;
    if (spec.rangeOver == FractalSpec.CENTER) {
      spec.center.plus_equal(spec.vy.times_scalar(pixels * perPixel).bigQuaternionValue());
    } else if (spec.rangeOver == FractalSpec.COEFF) {
      spec.coeff.plus_equal(spec.vy.times_scalar(pixels * perPixel).bigQuaternionValue());
    } else if (spec.rangeOver == FractalSpec.TRANS) {
      spec.trans.plus_equal(spec.vy.times_scalar(pixels * perPixel).bigQuaternionValue());
    }
    jQCentral_.setParameters(spec);
  }
  
  public void adjustCenterOnW (int pixels)
  {
    FractalSpec spec = getAdjustBaseSpec();
    double perPixel = spec.range / spec.pix_width;
    FastQuaternion wAxis = jQCentral_.parmsWAxis();
    if (spec.rangeOver == FractalSpec.CENTER) {
      spec.center.plus_equal(wAxis.times_scalar(pixels * perPixel).bigQuaternionValue());
    } else if (spec.rangeOver == FractalSpec.COEFF) {
      spec.coeff.plus_equal(wAxis.times_scalar(pixels * perPixel).bigQuaternionValue());
    } else if (spec.rangeOver == FractalSpec.TRANS) {
      spec.trans.plus_equal(wAxis.times_scalar(pixels * perPixel).bigQuaternionValue());
    }
    jQCentral_.setParameters(spec);
  }
  
 
      
  // ------------------ MouseListener & WindowListener for this frame ----------------------
  
  /* class MyMouseLNR implements MouseListener
  {
    public  void mouseClicked(MouseEvent e)
    {
      System.out.println("MyMouseLNR.mouseClicked");
    }

    public  void mousePressed(MouseEvent e)
    {
      System.out.println("MyMouseLNR.mousePressed");
    }

    public  void mouseReleased(MouseEvent e)
    {
      System.out.println("MyMouseLNR.mouseReleased");
    }

    public  void mouseEntered(MouseEvent e)
    {
      System.out.println("MyMouseLNR.mouseEntered");
    }

    public  void mouseExited(MouseEvent e)
    {
      System.out.println("MyMouseLNR.mouseExited");
    }
  } */
  
  class MyWindowLNR implements WindowListener
  {
    public void windowOpened (WindowEvent e)
    {
//      jQCentral_.message("JQFrame$MyWindowLNR.windowOpened", false);
    }

    public void windowClosing (WindowEvent e)
    {
//      jQCentral_.message("JQFrame$MyWindowLNR.windowClosing", false);
      jQCentral_.removeFrame(self_);
    }

    public void windowClosed (WindowEvent e)
    {
//      jQCentral_.message("JQFrame$MyWindowLNR.windowClosed", false);
    }

    public void windowIconified (WindowEvent e)
    {
//      jQCentral_.message("JQFrame$MyWindowLNR.windowIconified", false);
    }

    public void windowDeiconified (WindowEvent e)
    {
//      jQCentral_.message("JQFrame$MyWindowLNR.windowDeiconified", true);
    }

    public void windowActivated (WindowEvent e)
    {
      jQCentral_.activateMe(self_);
    }

    public void windowDeactivated (WindowEvent e)
    {      
      // jQCentral_.message("JQFrame$MyWindowLNR.windowDeactivated", false);
    }
  }
 
  // ------------------ Action Listeners for Menus ----------------------
  
  // -- File Menu -- openFileLNR openUrlLNR saveAsFileLNR saveParamsLNR closeLNR quitLNR



  class newFrameLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.makeNewFrame(self_);
    }
  }
  
  public void openImage ()
  {
    FractalSpec argsReadFromFile_ = loadArgsThenGIF(jQCentral_.write2URLP(), null);
    if (argsReadFromFile_ == null) {
      jQCentral_.message("JavaQuat image couldn't be read.", false);
    } else {
      jQCentral_.message("Image arg set loaded.", false);
    }
  }

  class openFileLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      self_.openImage();
    }
  }
  
  class openParmLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      self_.readParms(jQCentral_.write2URLP(), -1);
    }
  }
  
  class openLogLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent event)
    {
      IntDialog numberBox = new IntDialog("Enter log entry number", self_, "-1");
      try {
        int entry = numberBox.getValue();
        self_.readParms(jQCentral_.write2URLP(), entry);
      } catch (NumberFormatException exc) {
        jQCentral_.message("Error in reading entry number.",true);
        jQCentral_.message(exc.getMessage() == null ? exc.toString() : exc.getMessage(), true);
      }
    }
  }
  
  class trapFileLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.openTrapImage();
    }
  }
  
  public void saveImage()
  {
    writeGIF(jQCentral_.write2URLP(), null);
  }

  class saveAsFileLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      writeGIF(jQCentral_.write2URLP(), null);
    }
  }

  class write2URLLNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED)
        jQCentral_.setWrite2URLP(false);
      else if (event.getStateChange() == ItemEvent.SELECTED)
        jQCentral_.setWrite2URLP(true);
      else 
        jQCentral_.message("JQFrame.write2URLLNR unknown state.", false);
    }
  }
  
  class ClrFmTrap implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        //colorMapFromTrapP_.setState(false);
        jQCentral_.setColorMapFromTrap(false);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
        //colorMapFromTrapP_.setState(true);
        jQCentral_.setColorMapFromTrap(true);
      } else {
        jQCentral_.message("JQFrame.soQ1LNR unknown state.", false);
      }
    }
  }
  
  class loggingLNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        jQCentral_.toggleLogging(false);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
        if (!jQCentral_.toggleLogging(true)) {
          loggingMI_.setState(false); //Show that user cancelled.
        }
      } else {
        jQCentral_.message("JQFrame.loggingLNR logging unknown state.", false);
      }
    }
  }
  
  class saveArgsFLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      writeParms(jQCentral_.write2URLP(), null);
    }
  }
  
  class usrComLNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED)
        jQCentral_.setUserWantsToAddCOmmentP(false);
      else if (event.getStateChange() == ItemEvent.SELECTED)
        jQCentral_.setUserWantsToAddCOmmentP(true);
      else 
        jQCentral_.message("JQFrame.usrComLNR unknown state.", false);
    }
  }
  
  class quitLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.timeToQuit();
    }
  }

  class closeLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.removeFrame(self_);
    }
  }

  class debugLNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED)
        jQCentral_.setDebugP(false);
      else if (event.getStateChange() == ItemEvent.SELECTED)
        jQCentral_.setDebugP(true);
      else 
        jQCentral_.message("JQFrame.usrComLNR unknown state.", false);
    }
  }
  
  class gcLNR implements ActionListener
    // This was needed when the MRJ garbage collector didn't work right.
  {
    public void actionPerformed (ActionEvent e)
    {
      for(int i = 0; i < 10; i++) {
        // MRJ 2.1 doesn't do the whole thing before quitting.
        System.gc();
      }
    }
  }

  // ---------- Windows Menu ----------------- parmsLNR colorsLNR msgLNR

  class parmsLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.showParmFrameFor(self_);
    }
  }

  class colorsLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.updateColorMapEd(true, canvas());
    }
  }
    
  class msgLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.setMessagesVisible(true);
    }
  }
  
  class TrapLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.showTrapFrame();
    }
  }
  
  // -------------- Draw Menu ---------------- startDrawLNR stopDrawLNR

  class startDrawLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      self_.draw();
    }
  }
  
  class stopDrawLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      canvas_.stop();
      if (seqInProgressP_) {
        seqInProgressP_ = false;
        seqPausedP_ = true;
        startSeqMI_.setLabel("Resume");
      }
    }
  }

  class ClipLNR implements ActionListener
    // Puts the image in the SystemClipBoard.
  {
    public void actionPerformed (ActionEvent e)
    {
      canvas_.copyImageToSystemClipboard();
    }
  }
  
  // --------------- Settings Menu ------------------ fastLNR c256LNR
  
/*  class fastLNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED)
        jQCentral_.setHifi_p(true);
      else if (event.getStateChange() == ItemEvent.SELECTED)
        jQCentral_.setHifi_p(false);
      else 
        jQCentral_.message("JQFrame.fastLNR unknown state.", false);
    }
  }
  
  class c256LNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED)
        jQCentral_.setColorsIs256_p(false);
      else if (event.getStateChange() == ItemEvent.SELECTED)
        jQCentral_.setColorsIs256_p(true);
      else 
        jQCentral_.message("JQFrame.c256LNR unknown state.", false);
    }
  }
  */
  
  class soQ1LNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        Q1MI_.setState(true);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
//        jQCentral_.setRangeOver(FractalSpec.CENTER);
        setScanOnFalse();
        Q1MI_.setState(true);
      } else {
        jQCentral_.message("JQFrame.soQ1LNR unknown state.", false);
      }
    }
  }
  
  class soQ2LNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        Q2MI_.setState(true);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
//        jQCentral_.setRangeOver(FractalSpec.COEFF);
        setScanOnFalse();
        Q2MI_.setState(true);
      } else {
        jQCentral_.message("JQFrame.soQ2LNR unknown state.", false);
      }
    }
  }
  
  class soQ3LNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        Q3MI_.setState(true);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
//        jQCentral_.setRangeOver(FractalSpec.TRANS);
        setScanOnFalse();
        Q3MI_.setState(true);
      } else {
        jQCentral_.message("JQFrame.soQ3LNR unknown state.", false);
      }
    }
  }
  
  class ppp1LNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
//      jQCentral_.message("JQFrame$ppp1LNR entry.", false);
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        ppp1MI_.setState(true);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
//        jQCentral_.setPtsPerPixel(1);
        setPPPFalse();
        ppp1MI_.setState(true);
      } else {
        jQCentral_.message("JQFrame.ppp1LNR  unknown state.", false);
      }
    }
  }
  
  class ppp2LNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
//      jQCentral_.message("JQFrame$ppp2LNR entry.", false);
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        ppp2MI_.setState(true);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
//        jQCentral_.setPtsPerPixel(2);
        setPPPFalse();
        ppp2MI_.setState(true);
      } else { 
        jQCentral_.message("JQFrame.ppp2LNR  unknown state.", false);
      }
    }
  }
  
  class ppp4LNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        ppp4MI_.setState(true);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
//        jQCentral_.setPtsPerPixel(4);
        setPPPFalse();
        ppp4MI_.setState(true);
      } else {
        jQCentral_.message("JQFrame.ppp4LNR  unknown state.", false);
      }
    }
  }
  
  class ppp9LNR implements ItemListener 
  {
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        ppp9MI_.setState(true);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
//        jQCentral_.setPtsPerPixel(9);
        setPPPFalse();
        ppp9MI_.setState(true);
      } else { 
        jQCentral_.message("JQFrame.ppp9LNR  unknown state.", false);
      }
    }
  }

  class HVRatioLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      RatioDialog dialog = new RatioDialog("Width/Height Ratio", self_, hVRatio_);
      hVRatio_ = dialog.getDimension();
      jQCentral_.setHVRatio(hVRatio_);
    }
  }
  
//  class keepColorsLNR implements ItemListener 
//  {
//    public void itemStateChanged(ItemEvent event)
//    {
//      if (event.getStateChange() == ItemEvent.DESELECTED) {
//        jQCentral_.setRetainPreviousColorMapOnActivationP(false);
//      } else if (event.getStateChange() == ItemEvent.SELECTED) {
//        jQCentral_.setRetainPreviousColorMapOnActivationP(true);
//      } else 
//        jQCentral_.message("JQFrame.keepColorsLNR unknown state.", false);
//    }
//  }
  
  class displayParmsLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.setParameters(canvas_.spec());
      setCheckBoxValues(canvas_.spec(), false);
      jQCentral_.message("Format " + canvas_.spec().version, false);
    }
  }
      
  
     /* ----------------- Loop Menu ------------------ */
      // loopLNR
      
  class loopLNR implements ItemListener 
  {
    String loopName_;
    
    public loopLNR (String name)
    {
      loopName_ = name;
    }
    
    public void itemStateChanged(ItemEvent event)
    {
      if (event.getStateChange() == ItemEvent.SELECTED) {
        setLooper(loopName_);
        jQCentral_.setLoooper4ParmFrame(loopName_);
      }
    }
  }
  
     /* --------------- Adjust Menu ---------------- */
  
  class EnlargeLNR  implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      IntDialog dialog = new IntDialog("Zoom Percentage", self_, 
                                       String.valueOf(jQCentral_.enlargePixels()) );
      try {
        jQCentral_.setEnlargePixels(dialog.getValue());
        adjustEnlarge(jQCentral_.enlargePixels());
      } catch (NumberFormatException except) {
        jQCentral_.message("Illegal percentage entered: " + except,true);
        jQCentral_.message(except.getMessage() == null ? 
                           except.toString() : except.getMessage(), true);
      }
    }
  }
    
  class EnlrgAgnLNR  implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      adjustEnlarge(jQCentral_.enlargePixels());
    }
  }
  
  class RightLNR  implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      IntDialog dialog = new IntDialog("Shift Right pixels", self_, 
                                       String.valueOf(jQCentral_.rightPixels()) );
      try {
        jQCentral_.setRightPixels(dialog.getValue());
        adjustCenterRight(-jQCentral_.rightPixels());
      } catch (NumberFormatException except) {
        jQCentral_.message("Illegal # of pixels: " + except,true);
        jQCentral_.message(except.getMessage() == null ? 
                           except.toString() : except.getMessage(), true);
      }
    }
  }
    
  class RightAgnLNR  implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      adjustCenterRight(jQCentral_.rightPixels());
    }
  }
  
  class DownLNR  implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      IntDialog dialog = new IntDialog("Shift Down pixels", self_, 
                                       String.valueOf(jQCentral_.downPixels()) );
      try {
        jQCentral_.setDownPixels(dialog.getValue());
        adjustCenterUp(jQCentral_.downPixels());
      } catch (NumberFormatException except) {
        jQCentral_.message("Illegal # of pixels: " + except,true);
        jQCentral_.message(except.getMessage() == null ?
                           except.toString() : except.getMessage(), true);
      }
    }
  }
    
  class DownAgnLNR  implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      adjustCenterUp(jQCentral_.downPixels());
    }
  }    
     
  class OnWLNR  implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      IntDialog dialog = new IntDialog("Shift pixels on W", self_, 
                                       String.valueOf(jQCentral_.onWPixels()));
      try {
        jQCentral_.setOnWPixels(dialog.getValue());
        adjustCenterOnW(jQCentral_.onWPixels());
      } catch (NumberFormatException except) {
        jQCentral_.message("Illegal # of pixels: " + except,true);
        jQCentral_.message(except.getMessage() == null ?
                           except.toString() : except.getMessage(), true);
      }
    }
  }
    
  class OnWAgnLNR  implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      adjustCenterOnW(jQCentral_.onWPixels());
    }
  }    
     
  /* --------------- Axes Menu ---------------------- */
     
  class swapHVLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.swapHV();  
    }
  }
  
  class Rotate90RLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.rotate90R();
    }
  }

  class Rotate90LLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.rotate90L();
    }
  }

  class ScaleVLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      IntDialog dialog = new IntDialog("Scale V Axis by...", self_, "-1" );
      try {
        jQCentral_.scaleVAxis(dialog.getDoubleValue());
      } catch (NumberFormatException except) {
        jQCentral_.message("Illegal V scale factor: " + except,true);
        jQCentral_.message(except.getMessage() == null ? 
                           except.toString() : except.getMessage(), true);
      }
    }
  }

  class W2HLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.w2H();
    }
  }
  
  class W2VLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.w2V();
    }
  }

  class H2MemHLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.h2MemH();
    }
  }

  class V2MemVLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.v2MemV();
    }
  }
  
  class MemH2HLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.memH2H();
    }
  }

  class MemV2VLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.memV2V();
    }
  }

  class mouseH2LNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.mouseNewHAxis();
    }
  }

  class mouseV2LNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.mouseNewVAxis();
    }
  }
  
  /*  ------------------ Sequence Menu ---------------- */
    
  private void stopSequence ()
  {
    seqFrameNumber_ = 0;
    seqInProgressP_ = false;
    seqPausedP_ = false;
    seqPaused4ScreenCaptureP_ = false;
    seqSpec_.setStartSpec(null);
    startSeqMI_.setLabel("Start");
    canvas_.stop();
    //canvas_.setEverDrawnPFalse();
  }
    
  class startSeqLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
      // See autoAdvance at about line 981.
    {
      if (!seqInProgressP_) {
        FractalSpec spec1 = new FractalSpec(spec_); // Gets the basic values in.
        if (seqSpec_.startSpec() == null) {
          if (jQCentral_.parmsToSpec(spec1)) {        // Changes made for parm window values.
            getCheckBoxValues(spec1);
            seqSpec_.setStartSpec(spec1);
          } else {
            seqSpec_.setStartSpec(null);
            jQCentral_.message("Sequence is not properly specified: no start\nspec or " +
                                "valid parameter window values.", true);
            return;
          }
        } else if (!seqPausedP_ && jQCentral_.parmsToSpec(spec1)) {
          // Set the panels regardless of the start spec of the sequence if we are beginning.
          seqSpec_.startSpec().panels = spec1.panels;
          seqSpec_.startSpec().panel_x = spec1.panel_x;
          seqSpec_.startSpec().panel_y = spec1.panel_y;
        }
        if (!seqSpec_.readyP(jQCentral_)) {
          jQCentral_.message("Sequence is not properly specified: Sequence spec not ready.",
                             true);
          return;
        }
        // Start up an auto advance sequence.
        if (seqPaused4ScreenCaptureP_) {
          canvas_.setParameters(savedSequencePauseSpec_);
          seqPaused4ScreenCaptureP_ = false;
        } else if (!seqPausedP_) {
          canvas_.setParameters(seqSpec_.startSpec()); // this does extra stuff.
          seqSpec_.setToFrame(0); // Was seqFrameNumber_ rather than 0.  But we're at start.
        } else {
          jQCentral_.message("Continuing from pause in sequence.", false);
        }
        seqInProgressP_ = true;
        seqPausedP_ = false;
        startSeqMI_.setLabel("Stop");
        canvas_.threaded_paint_fractal();
      } else {
        stopSequence();
      }
    }
  }

  class StepLNR  implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      IntDialog dialog = new IntDialog("Set Sequence Step", self_, 
                                       String.valueOf(seqFrameNumber_));
      try {
        seqPaused4ScreenCaptureP_ = true;
        seqInProgressP_ = false;
        seqPausedP_ = true;
        seqFrameNumber_ = dialog.getValue();
        stepAdvance(seqFrameNumber_ -1);
        startSeqMI_.setLabel("Start at " + seqFrameNumber_);
      } catch (NumberFormatException except) {
        jQCentral_.message("Error giving steps: " + except,true);
        jQCentral_.message(except.getMessage() == null ?
                           except.toString() : except.getMessage(), true);
      }
    }
  }
    
  class seqLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      if (seqDialog_ == null) {
        seqDialog_ = new SeqDialog("Sequence Setup", self_, seqSpec_, jQCentral_);
      } else {
        seqDialog_.prepForDisplay();
        seqDialog_.setVisible(true);
      }
      seqFrameNumber_ = 0;
    }
  }
  
  class RevPanelsLNR implements ItemListener
    // If doing a 3x3 set of panels, in parameters window set panels=3, x panel=3,
    //  y panel = 3 and set reverse and you will get (2,2), (1,2), (0,2), (2,1)...
    //  (0,0).  This is geed when using 2 machines, to go from both ends of a lot
    // of panels.
  {
    public void itemStateChanged (ItemEvent event)
    { 
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        seqSpec_.advancePanelsBackwards_p = false;
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
        seqSpec_.advancePanelsBackwards_p = true;
      }
    }
  }

  class RowByRowLNR implements ItemListener
    // If doing a 3x3 set of panels, in parameters window set panels=3, x panel=3,
    //  y panel = 3 and set reverse and you will get (2,2), (1,2), (0,2), (2,1)...
    //  (0,0).  This is geed when using 2 machines, to go from both ends of a lot
    // of panels.
  {
    public void itemStateChanged (ItemEvent event)
    { 
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        seqSpec_.advanceRowByRow_p = false;
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
        seqSpec_.advanceRowByRow_p = true;
      }
    }
  }

  class StrSeqLNR implements ItemListener
  {
    public void itemStateChanged (ItemEvent event)
    { 
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        seqSpec_.setStartSpec(null);
        seqSpec_.enabled_p = true;
        jQCentral_.message("First frame in sequence is cleared.", false);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
        seqSpec_.setStartSpec(canvas_.spec());
        seqSpec_.enabled_p = true;
        jQCentral_.message("First frame in sequence is set.", false);
      }
    }
  }

  class EndSeqLNR implements ItemListener
  {
    public void itemStateChanged (ItemEvent event)
    { 
      if (event.getStateChange() == ItemEvent.DESELECTED) {
        seqSpec_.setEndSpec(null);
        jQCentral_.message("Last frame in sequence is cleared.", false);
      } else if (event.getStateChange() == ItemEvent.SELECTED) {
        seqSpec_.setEndSpec(canvas_.spec());
        seqSpec_.enabled_p = true;
        jQCentral_.message("Last frame in sequence is set.", false);
      }
    }
  }


    // startSeqLNR pasVsWrLNR
  
}
