package JavaQuat20;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.awt.image.IndexColorModel;
import java.awt.image.DirectColorModel;
import java.io.*;
import java.net.URL;
import java.net.*;

class JQIO
{
  Messenger messenger_ = null;
  Frame frame_ = null;
  
  public JQIO (Messenger messengerArg, Frame frameArg)
  {
    messenger_ = messengerArg;
    frame_ = frameArg;
  }
  
  public void setClient (Frame frameArg)
  {
    frame_ = frameArg;
  }
  
  public String pathFromDialog (boolean read_p)
  {
     FileDialog fd;
     if (read_p) {
       fd = new FileDialog(frame_, null, FileDialog.LOAD);
     } else {
       fd = new FileDialog(frame_, null, FileDialog.SAVE);
     }
     try {
       fd.show();
     } catch (Throwable e) {
       return null;
     }
     //Allow for variation in Java implementations
     String dir = fd.getDirectory();
     if (dir == null) {
       // User hit cancel probably.
       return null;
     }
     boolean haslash = dir.charAt(dir.length() - 1) == '/';
     String pathname_ = (dir + (haslash ? "" : "/") + fd.getFile());
     return pathname_;
  }
  
  public String urlStringFromDialog ()
  {
    UrlDialog ud = new UrlDialog("Enter URL", frame_);
    try {
      ud.show();
    } catch (Throwable e) {
      return null;
    }
    //Allow for variation in Java implementations
    return ud.getText();
  }
  
  public String urlStringFromDialog (String initialString)
  {
    UrlDialog ud = new UrlDialog("Enter URL", frame_);
    ud.setText(initialString);
    try {
      ud.show();
    } catch (Throwable e) {
      return null;
    }
    //Allow for variation in Java implementations
    return ud.getText();
  }
  
  public InputStream toInputStream (String pathname1, boolean is_url_p, boolean testOnly_p)
  {
     InputStream stream1 = null;
     try {
       if (!is_url_p) {
         File testExistance = new File(pathname1);
         if (!testExistance.exists()) {
           throw new IOException("File DOES NOT EXIST - no read done.");
//         } else if (!testExistance.canWrite()) {
//           throw new IOException("File cannot be written to.");
         }
         if (testOnly_p)
           stream1 = null;
         else
           stream1 = new FileInputStream(pathname1);
       } else {
         URL url = new URL(pathname1);
         stream1 = url.openStream();
       }
     } catch (MalformedURLException e) {
       messenger_.message("Malformed URL - try again.", false);
       stream1 = null;
     } catch (Throwable e) {
       messenger_.message((e.getMessage() == null ? e.toString() : e.getMessage()), false);
       stream1 = null;
     } finally {
       return stream1;
     }
  }

  public OutputStream toOutputStream (String pathname1, boolean is_url_p, boolean testOnly_p)
  {
     OutputStream stream1 = null;
     try {
       if (!is_url_p) {
         File testExistance = new File(pathname1);
         if (testExistance.exists()) {
           // Rename it to an unused backup version.  We could check here for filename length.
           // Since Mac only allows 31 characters.
           File tester = null;
           int count = 1;
           while ((tester = (new File(pathname1 + ".~" + count + "~"))).exists()) {
             if (count++ > 1000) {
               throw new IOException("Excessive backup versions exist - no write done.");
             }
           }
           testExistance.renameTo(tester);
           testExistance = new File(pathname1);
           if (testExistance.exists()) {
             throw new IOException("Already Exists or invalid - no write done.");
           }
         }
         if (testOnly_p)
           stream1 = null;
         else
           stream1 = new FileOutputStream(pathname1);
       } else {
         URL url = new URL(pathname1);
         URLConnection conn = url.openConnection();
         conn.setDoOutput(true);
         stream1 = conn.getOutputStream();
       }
     } catch (Throwable e) {
       messenger_.message(("JQIO.toOutputStream " + e.getMessage() == null ? e.toString() : e.getMessage()), false);
       stream1 = null;
     } finally {
       return stream1;
     }
  }

  public OutputStream toAppendStream (String pathname1) 
  {
     FileInputStream oldStream = null;
     FileOutputStream newStream = null;
     try {
       // Make sure temp file doesn't exist.
       String tempPathname1 = pathname1 + ".~temp~";
       File tempFile = new File(tempPathname1);
       if (tempFile.exists()) {
         tempFile.delete();
       }
       File oldFile = new File(pathname1);
       if (oldFile.exists()) {
         oldFile.renameTo(tempFile);        
         
         oldFile = new File(tempPathname1);
         oldStream = new FileInputStream(oldFile);
         File newFile = new File(pathname1);
         newStream = new FileOutputStream(newFile);
         int byte1;
         while ((byte1 = oldStream.read()) != -1) {
           newStream.write(byte1 & 0xFF);
         }
       } else {
         File newFile = new File(pathname1);
         newStream = new FileOutputStream(newFile);
       }
     } catch (Throwable e) {
       messenger_.message(("JQIO.toAppendStream " + e.getMessage() == null ? e.toString() : e.getMessage()), false);
       return null;
     } finally {
       try {
         oldStream.close();
       } catch (Throwable e) {
         messenger_.message(("JQIO.toAppendStream " + e.getMessage() == null ? e.toString() : e.getMessage()), false);
       }
       return newStream;
     }
  }
}
