package JavaQuat20;

import java.math.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;

class LabeledTextField extends Panel implements KeyListener, TextListener
{
  public static final int IS_INT = 0;
  public static final int IS_DOUBLE = 1;
  public static final int IS_BIG = 2;
  public static final int IS_STRING = 3;
  private int valueType = IS_DOUBLE;

  public static final int SIDE_BY_SIDE_1 = 0;
  public static final int SIDE_BY_SIDE_2 = 1;
  public static final int VERTICAL = 2;
  public static final int NO_LABEL = 3;
  private int styleType = VERTICAL;
    
  int int_value;
  double double_value;
  BigDecimal bigdecimal_value;
  

  boolean is_ok_p = false;
  boolean check_needed_p = true;
  Label the_label = null;
  private TextField text_ = null;
  private Vector myActionListeners;
  private Vector myTextListeners;
  
  public LabeledTextField (String label_string, int valueTypeArg, int styleTypeArg)
  {
    valueType = valueTypeArg;
    styleType = styleTypeArg;
    if (styleTypeArg == SIDE_BY_SIDE_1) {
        setLayout(new BorderLayout(2, 0));
        text_ = new TextField("0", 18);
        text_.addKeyListener(this);
        text_.addTextListener(this);
        the_label = new Label(" ",Label.CENTER);
        // text.setBackground(Color.white);
        add("Center", the_label);
        add("East", text_);
    } else if (styleTypeArg == SIDE_BY_SIDE_2) {
      // this was where we colored the fields in the 1.7 version of JQ.
        setLayout(new GridLayout(1, 2, 0, 0));
        text_ = new TextField("1", 9);
        text_.addKeyListener(this);
        text_.addTextListener(this);
        // text_.setBackground(colortext);
        the_label = new Label(label_string, Label.CENTER);
        // the_label.setBackground(colorlabel);
        add(the_label);
        add(text_);
    } else if (styleTypeArg == VERTICAL) {
        setLayout(new BorderLayout(2, 2));
        text_ = new TextField("0", 18);
        text_.addKeyListener(this);
        text_.addTextListener(this);
        the_label = new Label(label_string, Label.CENTER);
        add("Center", the_label);
        add("South", text_);
    } else if (styleTypeArg == NO_LABEL) {
        setLayout(new BorderLayout(2, 2));
        text_ = new TextField("0", 18);
        text_.addKeyListener(this);
        text_.addTextListener(this);
        // the_label = new Label(label_string, Label.CENTER);
        // add("Center", the_label);
        add("Center", text_);
    } else {
      System.err.println("LabeledTextField() - Illegal styleTypeArg.");
    }
    myActionListeners = new Vector(1);
    myTextListeners = new Vector(1);
    valueOk_p();
  }
  
  
  public void setText (String new_value)
  {
    text_.setText(new_value);
  }
  
  public String getText ()
  {
    return text_.getText();
  }
  
  public boolean isBlank_p ()
  {
    return (text_.getText().trim().length() == 0);
  }
  
  public boolean zeroIfBlank ()
  {
    if (isBlank_p()) {
      text_.setText("0");
      return true;
    }
    return false;
  }
  
  public void set_label (String new_label)
  {
    if (styleType != NO_LABEL)
      the_label.setText(new_label);
  }
  
  public void setFont (Font f)
  {
    text_.setFont(f);
    if (the_label != null)
      the_label.setFont(f);
    super.setFont(f);
  }
  
  private boolean intConvertible_p (String string1)
    // If string1 can be converted to a int, returns true and sets double_value.
  {
    try {
      Integer value_wrapped = Integer.valueOf(string1);
      int_value = value_wrapped.intValue();
      return true;
    } catch (NumberFormatException n) {
      return false;
    }
  }

  private boolean doubleConvertible_p (String string1)
    // If string1 can be converted to a double, returns true and sets double_value.
  {
    try {
      Double value_wrapped = Double.valueOf(string1);
      double_value = value_wrapped.doubleValue();
      return true;
    } catch (NumberFormatException except) {
//      System.err.println("LabeledTextField.doubleConvertible_p " + except);
      return false;
    }
  }
  
  private boolean bigdecimalConvertible_p (String string1)
    // If string1 can be converted to a BigDecimal, returns true and sets bigdecimal_value.
  {
    try {
      bigdecimal_value = new BigDecimal(string1);
      return true;
    } catch (NumberFormatException except) {
//      System.err.println("LabeledTextField.bigdecimalConvertible_p " + except);
      return false;
    }
  }


  public boolean valueOk_p (String string1)
  {
    boolean ok_p = false;
    if (valueType == IS_INT) {
        if (string1.length() > 0 && intConvertible_p(string1)) {
          ok_p = true;
        }
    } else if (valueType == IS_DOUBLE) {
        if (string1.length() > 0 && doubleConvertible_p(string1)) {
          ok_p = true;
        }
    } else if (valueType == IS_BIG) {
        if (string1.indexOf('e') >= 0 || string1.indexOf('E') >= 0) {
          string1 = removeExponent(string1);
          text_.setText(string1);
        }      
        if (string1.length() > 0 && bigdecimalConvertible_p(string1)) {
          ok_p = true;
        }
    } else if (valueType == IS_STRING) {
        if (string1.length() > 0 && Character.isJavaIdentifierStart(string1.charAt(0))) {
          ok_p = true;
        }
    } else {
      System.err.println("LabeledTextField valueOk_p - impossible valueType value.");
    }
    if (ok_p)
      text_.setForeground(Color.black);
    else
      text_.setForeground(Color.red);
    repaint();
    return ok_p;
  } 
   
  public boolean valueOk_p ()
    // If string1 can be converted to an int, returns true and sets int_value, colors red or black.
  {
    return valueOk_p(text_.getText());
  }

  private String removeExponent (String string1)
    // If string1 can be converted to an bigdecimal, returns true and sets bigdecimal_value, colors red or black.
  {
    int eIndex = string1.indexOf('e');
    if (eIndex < 0)
      eIndex = string1.indexOf('E');
    String exponentStr = string1.substring(eIndex + 1);
    String mantissaStr = string1.substring(0, eIndex);
    int exponent;
    BigDecimal mantissa;
    try {
      exponent = Integer.valueOf(exponentStr).intValue();
      mantissa = new BigDecimal(mantissaStr);
    } catch (NumberFormatException ex) {
      System.err.println("Labeled_Text_Field.removeExponent error: " + ex);
      return "0.0";
    }

    return mantissa.movePointRight(exponent).toString();
  }
  
  public int intValue ()
  {
    if (valueOk_p())
      return int_value;
    return 0;
  }
  
  public double doubleValue ()
  {
    if (valueOk_p())
      return double_value;
    return 0;
  }  

  
  public BigDecimal bigdecimalValue ()
  {
    if (valueOk_p())
      return bigdecimal_value;
    return new BigDecimal("0");
  }
    
  // Use getText to get the string value.
 
  public void textValueChanged (TextEvent event)
  {
    valueOk_p(); //Called to set color.
    
    /*
    if (valueOk_p()) { //Called to set color.
      Enumeration iterator1;
      synchronized(this) {
        iterator1 = myTextListeners.elements();
      }

      try {
        while (iterator1.hasMoreElements()) {
          ((TextListener)iterator1.nextElement()).textValueChanged(event);
        }
      } catch ( NoSuchElementException exception) {
        System.err.println("The impossible has happened in Labeled_Text_Field.textValueChanged. " +
                           exception);
      }
    }      */
  }
  
  public synchronized void addKeyListener(KeyListener listener)
  {
    text_.addKeyListener(listener);
  }
  
  public synchronized void addTextListener(TextListener listener)
    // listener's textValueChanged is passed a TextEvent when the text changes.
    // The big advantage with this is that the listener can see what the new text is.
  {
    text_.addTextListener(listener);
    //myTextListeners.addElement(listener);
  }
  
  public synchronized void removeTextListener(TextListener listener)
    // listener's textValueChanged is passed a TextEvent when the text changes.
    // The big advantage with this is that the listener can see what the new text is.
  {
    text_.removeTextListener(listener);
    //myTextListeners.removeElement(listener);
  }
  
  // Action Listener Stuff
  
  public synchronized void addActionListener(ActionListener listener)
  {
    myActionListeners.addElement(listener);
  }
  
  public void keyPressed(KeyEvent e) {}
  public void keyReleased(KeyEvent e) {}
 
  public void keyTyped(KeyEvent event)
    // This issues an ActionEvent whose source is 'this' when Return is typed
    // to all myActionListeners registered with addActionListener.
  {
  
    if (event.getKeyChar() != '\n') {
      return;
    }
    Enumeration iterator1;
    synchronized(this) {
      iterator1 = myActionListeners.elements();
    }
    ActionEvent actionEvent1 = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, "Return Typed");

    try {
      while (iterator1.hasMoreElements()) {
        ((ActionListener)iterator1.nextElement()).actionPerformed(actionEvent1);
      }
    } catch ( NoSuchElementException exception) {
      System.err.println("The impossible has happened in Labeled_Text_Field.keyTyped.  " +
                         exception);
    }
  }
}  
