package JavaQuat20;

import java.awt.*;
import java.awt.image.*;
import java.util.*;
import JavaQuat20.FractalLoops.*;

class Q1ImageProducer implements ImageProducer, Runnable 
{
  static final int bitsPerByte = 8;
  static final int bitsPerInt = 32;
  double NON_BLACKS = 255.0;  // This is 254 for version 1.4, and 255 thereafter.

  JQCanvas parent;
  Vector consumers = new Vector();
  Hashtable properties = new Hashtable();
  int width;
  int height;
  byte [] pixels_; // Array of indices into colorMap
  int [] intPixels_; // Array of packed RGB values for DirectColorMap use.

  //Drawing control
  protected boolean stop_p_ = false;
  protected boolean draw_in_progress_p_ = false;
  boolean drawingFractalInterlock_p_ = false;
  boolean parameter_change_p_ = false;
  boolean ever_drawn_p = false;
  int last_row_drawn_ = -1;
  
  // Fractal Parameters
  public FractalSpec spec_;
  FastQuaternion specPoint = null;
  FastQuaternion specCoeff = null;
  FastQuaternion specTrans = null;
  
  // Fractal drawing helpers.
  static int min_pix_width_ = 9; // Doubles as min_pix_height_
  static int guide_row_count = 5;
  BigQuaternion lox;
  BigQuaternion hiy;
  BigQuaternion incx;
  BigQuaternion incy;
  FastQuaternion loxFast;
  FastQuaternion hiyFast;
  FastQuaternion incxFast;
  FastQuaternion incyFast;
  double clump;
  
  Array2 guide; // = new Array2(guide_row_count,spec_.pix_width);
  JQLooper loop_function;
  
  FastQuaternion yFast = null;
  FastQuaternion xFast = null;
  FastQuaternion leftFast = null;
  FastQuaternion pointFast = null;
  FastQuaternion coeffFast = null;
  FastQuaternion transFast = null;
  

  Q1ImageProducer(FractalSpec arg, JQCanvas p) 
    // This is for normal creation when we are going to compute a fractal
  {
    parent = p;
    spec_ = new FractalSpec(arg); // This lets us create one guide for life of this.
                                 // Caller gets new F_I_Producer with new size.
    specPoint = spec_.center.fastQuaternionValue();
    specCoeff = spec_.coeff.fastQuaternionValue();
    specTrans = spec_.trans.fastQuaternionValue();

    if (spec_.version == 1.4) {
      NON_BLACKS = 254;
    }
    width = spec_.pix_width;
    height = spec_.pix_height;
    
    yFast = new FastQuaternion();
    xFast = new FastQuaternion();
    leftFast = new FastQuaternion();
    pointFast = new FastQuaternion();
    coeffFast = new FastQuaternion();
    transFast = new FastQuaternion();


    pixels_ = new byte[width * height]; // We always want these around for GIF writing.
    for (int i= 0; i < pixels_.length; i++) {
      pixels_[i] = 0;
    }
    if (!spec_.colorsIs256_p) {
      intPixels_ = new int[width * height];
      for (int i= 0; i < intPixels_.length; i++) {
        intPixels_[i] = 0;
      }
    }
    guide = new Array2(guide_row_count, width);
    
    lox = new BigQuaternion();
    hiy = new BigQuaternion();
    incx = new BigQuaternion();
    incy = new BigQuaternion();
    loxFast = new FastQuaternion();
    hiyFast = new FastQuaternion();
    incxFast = new FastQuaternion();
    incyFast = new FastQuaternion();
  }
  
   
  Q1ImageProducer(Image refresher, FractalSpec arg, JQCanvas p)
    // This is for dummy creation, when we have loaded an image file, GIF or PNG.
  {
    parent = p;
    spec_ = new FractalSpec(arg); // This lets us create one guide for life of this.
                                 // Caller gets new F_I_Producer with new size.
    specPoint = spec_.center.fastQuaternionValue();
    specCoeff = spec_.coeff.fastQuaternionValue();
    specTrans = spec_.trans.fastQuaternionValue();

    if (spec_.version == 1.4) {
      NON_BLACKS = 254;
    }
    width = spec_.pix_width;
    height = spec_.pix_height;
    
    yFast = new FastQuaternion();
    xFast = new FastQuaternion();
    leftFast = new FastQuaternion();
    pointFast = new FastQuaternion();
    coeffFast = new FastQuaternion();
    transFast = new FastQuaternion();

    if (spec_.colorsIs256_p)
      intPixels_ = new int[width * height];
    try {
      PixelGrabber pg = new PixelGrabber(refresher, 0, 0, width, height, intPixels_, 0, width);
      if(!pg.grabPixels(10000)) // Wait for up to 10 sec.
        parent.message("Q1ImageProducer failed to grab pixels from image after 10 seconds.", false);

      // Check for errors.
      if ((pg.status() & ImageObserver.ABORT) != 0) {
        parent.message("Error while fetching image", true);
      }
    } catch (Exception e) {
      parent.message("Error while reading image", true);
      e.printStackTrace();
    } catch (Throwable e) {
      // I don't know this should be here, just being careful.
      parent.message("Throw-type Error while reading image", true);
    }
    
    if (spec_.colorsIs256_p) {
      pixels_ = new byte[width * height];
      parent.colorArrayToIndexArray(intPixels_, pixels_);
      intPixels_ = null;
      System.gc();
    }

    // Let's create this just in case of some mistake.
    guide = new Array2(guide_row_count, width);
    
    lox = new BigQuaternion();
    hiy = new BigQuaternion();
    incx = new BigQuaternion();
    incy = new BigQuaternion();
    loxFast = new FastQuaternion();
    hiyFast = new FastQuaternion();
    incxFast = new FastQuaternion();
    incyFast = new FastQuaternion();
    computeLoxEtc(); // Takes care of the above and clump also.
    
    stop_p_ = false;
    draw_in_progress_p_ = false;
    parameter_change_p_ = false;
    ever_drawn_p = true;
    last_row_drawn_ = height - 1;
  }
 
  
  int intify (byte b)
  {
    return (b < 0) ? 256 + b : b;
  }
  
  int reverseLoops256 (int depth, double loops)
  {
    if (spec_.version == 1.4) {
      return depth - (int)loops;
    }
    return Math.min(depth, (int)loops);
  }
  
  double reverseLoopsMillions (int depth, double loops)
  {
    if (spec_.version == 1.4) {
      return (double)depth - loops;
    }
    return Math.min((double)depth, loops);
  }
  
  public boolean trapImageProducerP ()
  {
    return false;
  }

  public void setDrawingFractalInterlock_p (boolean arg1)
  {
    drawingFractalInterlock_p_ = arg1;
  }
  
  public boolean drawingFractalInterlock_p ()
  {
    return drawingFractalInterlock_p_;
  }
  
  /*  ******************* pixel accessors for routines that write files **** */
  
  public byte[] pixels ()
  {
    return pixels_;
  }
  
  public int[] intPixels ()
  {
    return intPixels_;
  }
 
  public void remapPixelIndices (int [] pixelMap)
    //This is called from the colorMapEditor when colors have been sorted,
    // by way of JQCentral and the JQCanvas.
  {
    for (int i = 0; i < pixels_.length; i++) {
      int oldIndex = pixels_[i];
      if (oldIndex < 0) {
        oldIndex += 256;
      }
      int newIndex = pixelMap[oldIndex];
      pixels_[i] = (byte)((newIndex > 127) ? (newIndex - 256) : newIndex);
    }
  }
  

  
  /*  ******************* Anti-Alias / Anti-Speckle ************************ */
  
  /* private double checkLoops (double loops, int depth)
  {
    if (loops < 0) {
      System.err.println("Loop error, result was < 0.");
      return 0;
    }
    if (loops > depth) {
      System.err.println("Loop error, result was > depth.");
      return depth;
    }
    return loops;
  } */

  public double call_loop_function(FastQuaternion point_in, FastQuaternion coeff, 
                                   FastQuaternion trans, int depth)
  {
    if (spec_.pts_per_pixel == 1) {
      return loop_function.loop(point_in, coeff, trans, depth);
    }
    FastQuaternion pointA = new FastQuaternion(point_in);
    Vector answers = new Vector(9);
    double loops = 0;
    if (spec_.pts_per_pixel == 2) {  // Currently pts_per_pixel is never 2.
      FastQuaternion halfx = incxFast.times_scalar(0.5);
      FastQuaternion halfy = incyFast.times_scalar(0.5);
      loops = loop_function.loop(pointA, coeff, trans, depth);
      answers.addElement(new Double(loops));
      pointA.plus_equal(halfx).plus_equal(halfy);
      loops = loop_function.loop(pointA, coeff, trans, depth);
      answers.addElement(new Double(loops));
    } else if (spec_.pts_per_pixel <= 4) {
      FastQuaternion halfx = incxFast.times_scalar(0.5);
      FastQuaternion halfy = incyFast.times_scalar(0.5);
      FastQuaternion point2 = pointA.plus(halfx);
      for (int i = 0; ; i++) {
        loops = loop_function.loop(pointA, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.loop(point2, coeff, trans, depth);
        answers.addElement(new Double(loops));
        if (i == 1) {
          break;
        }
        pointA.plus_equal(halfy);
        point2.plus_equal(halfy);
      }
    } else if (spec_.pts_per_pixel > 4) {
      FastQuaternion thirdx = incxFast.times_scalar(0.333333333333);
      FastQuaternion thirdy = incyFast.times_scalar(0.333333333333);
      FastQuaternion point2 = pointA.plus(thirdx);
      FastQuaternion point3 = point2.plus(thirdx);
      for (int i = 0; ; i++) {
        loops = loop_function.loop(pointA, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.loop(point2, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.loop(point3, coeff, trans, depth);
        answers.addElement(new Double(loops));
        if (i == 2) {
          break;
        }
        pointA.plus_equal(thirdy);
        point2.plus_equal(thirdy);
        point3.plus_equal(thirdy);
      }      
    } 
    int numAnswers = answers.size();
    if (numAnswers <= 0) {
      parent.message("Q1ImageProducer::call_loop_function error: " + 
                     spec_.pts_per_pixel, true);
      return (double)loop_function.loop(pointA, coeff, trans, depth);;
    }
    // Find the average of answers.
    loops = 0;
    for (int j = 0; j < numAnswers; j++) {
      loops += ((Double)answers.elementAt(j)).doubleValue();
    }
    double average = loops / numAnswers;
    // Find answer closest to average and return it.
    loops = ((Double)answers.firstElement()).doubleValue();
    for (int k = 1; k < numAnswers; k++) {
      if (Math.abs(((Double)answers.elementAt(k)).doubleValue() - average) < 
          Math.abs(loops - average)) {
        loops = ((Double)answers.elementAt(k)).doubleValue();
      }
    }
    if (spec_.colorsIs256_p) {
      return Math.round(loops);
    }
    return loops;
  }
  
  public double call_big_loop_function(BigQuaternion point_in, BigQuaternion coeff, 
                                       BigQuaternion trans, int depth)
  {
    if (spec_.pts_per_pixel <= 1) {
      return loop_function.bigLoop(point_in, coeff, trans, depth);
    }
    BigQuaternion pointA = new BigQuaternion(point_in);
    Vector answers = new Vector(9);
    double loops = 0;
    if (spec_.pts_per_pixel == 2) {  // Currently pts_per_pixel is never 2.
      BigQuaternion halfx = incx.times_scalar(0.5);
      BigQuaternion halfy = incy.times_scalar(0.5);
      loops = loop_function.bigLoop(pointA, coeff, trans, depth);
      answers.addElement(new Double(loops));
      pointA.plus_equal(halfx).plus_equal(halfy);
      loops = loop_function.bigLoop(pointA, coeff, trans, depth);
      answers.addElement(new Double(loops));
    }
    if (spec_.pts_per_pixel <= 4) {
      BigQuaternion halfx = incx.times_scalar(0.5);
      BigQuaternion halfy = incy.times_scalar(0.5);
      BigQuaternion point2 = pointA.plus(halfx);
      for (int i = 0; ; i++) {
        loops = loop_function.bigLoop(pointA, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.bigLoop(point2, coeff, trans, depth);
        answers.addElement(new Double(loops));
        if (i == 1) {
          break;
        }
        pointA.plus_equal(halfy);
        point2.plus_equal(halfy);
      }
    }
    if (spec_.pts_per_pixel > 4) {
      BigQuaternion thirdx = incx.times_scalar(0.333333333333);
      BigQuaternion thirdy = incy.times_scalar(0.333333333333);
      BigQuaternion point2 = pointA.plus(thirdx);
      BigQuaternion point3 = point2.plus(thirdx);
      for (int i = 0; ; i++) {
        loops = loop_function.bigLoop(pointA, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.bigLoop(point2, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.bigLoop(point3, coeff, trans, depth);
        answers.addElement(new Double(loops));
        if (i == 2) {
          break;
        }
        pointA.plus_equal(thirdy);
        point2.plus_equal(thirdy);
        point3.plus_equal(thirdy);
      }      
    }
    int numAnswers = answers.size();
    if (numAnswers <= 0) {
      parent.message("Q1ImageProducer::call_big_loop_function error: " + 
                     spec_.pts_per_pixel, true);
      return (double)loop_function.bigLoop(pointA, coeff, trans, depth);
    }
    // Find the average of answers.
    loops = 0;
    for (int j = 0; j < numAnswers; j++) {
      loops += ((Double)answers.elementAt(j)).doubleValue();
    }
    double average = loops / numAnswers;
    // Find answer closest to average and return it.
    loops = ((Double)answers.firstElement()).doubleValue();
    for (int k = 1; k < numAnswers; k++) {
      if (Math.abs(((Double)answers.elementAt(k)).doubleValue() - average) < 
          Math.abs(loops - average)) {
        loops = ((Double)answers.elementAt(k)).doubleValue();
      }
    }
    if (spec_.colorsIs256_p) {
      return Math.round(loops);
    }
    return loops;
  }
  
  /*  ******************* Drawing Primitives ************************ */
  
  // *******   256 Color version for IndexColorModel. ******* 
  double pixout(int x, int y, int loops, int gdidx, boolean recordd)
    //loops is actually max-loops minus actual loops, so 0 means non-divergence and to be black.
  {
    //System.out.println("Q1ImageProducer::pixout ent");
    int depth = spec_.depth;
    int ldivc;
    boolean loops_is_ColorIndex = loop_function.outputsColorIndexP();
    if (loops_is_ColorIndex) {
      ldivc = loops;
    } else if (loops == depth) {
      ldivc = depth;
    } else {
      ldivc = (int)Math.min((int)NON_BLACKS * spec_.color_cycles,
                            (int)Math.floor(loops / clump)); //Only loop==depth is max value
    }
    // Record the real value
    if (recordd) {
      guide.aset(gdidx, x, ldivc);
    }
    
      // modify: 0 goes to 1 because in colormap 0 is sometimes transparent, 1 is black.
      //         all else goes to 2 through 255
    int colorIndex = 1; // We need to return ldivc so we need another variable.
     if (loops_is_ColorIndex) {
       colorIndex = (int)Math.min(loops, 255);
     } else if (loops == depth) {
        colorIndex = 255;
     } else {
        colorIndex = (int)(ldivc % NON_BLACKS);  // this is mod 255, so biggest is 254
     }
    
    // Draw the point.
    int i = x + y * width;
    pixels_[i] = (byte)colorIndex;

    // In the following line the returned value indicates whether the graphics instance is 
    // present.
    // So if the graphics instance is gone and we were not planning to stop, plan to stop.
    Color indexColor = parent.getColorAt(colorIndex);
    if (!parent.keepCurrentDrawLine(x, x, y, indexColor) && !stop_p_)
      stop_p_ = true;
    //System.out.println("Q1ImageProducer::pixout ex");
    return ldivc;
  }


  void frDrawline(int xleft, int xright, int yndx, int ldivc)
    // 256 color version
  {
    //System.out.println("Q1ImageProducer::frDrawline ent");
    // ldivc is a number recorded by pixout, so we pick up the computations for
    // drawing from there
    
    boolean loops_is_ColorIndex = loop_function.outputsColorIndexP();
    int colorIndex = 1; // this is to parallel the variable naming in pixOut.
     if (loops_is_ColorIndex) {
       colorIndex = ldivc;
     } else if (spec_.version == 1.4) {
      // modify: 0 goes to 1 because 0 is sometimes transparent, 1 is black.
      //         all else goes to 2 through 255
      if (ldivc != 0) {
        colorIndex = ((ldivc - 1) % (int)NON_BLACKS) + 2;
      }
    } else {
      if (ldivc == spec_.depth) {
        colorIndex = 255;
      } else {
        colorIndex = (int)(ldivc % NON_BLACKS);  // this is mod 255, so biggest is 254
      }
    }
    
    // Draw the horizontal line.
    int end = xright + yndx * width;
    for (int i = xleft + yndx * width; i <= end; i++) {
      pixels_[i] = (byte)colorIndex;
    }    
    
    Color indexColor = parent.getColorAt(colorIndex);
    parent.keepCurrentDrawLine(xleft, xright, yndx, indexColor);
    //System.out.println("Q1ImageProducer::frDrawline ex");
  }

  // *******   Full Color version for DirectColorModel. ******* 
  
  private double findFullColorIndex (double ldivc)
    // Returns a number in the range 1 thru 255.
  {
    double remainder = 0;
    if (ldivc != 0) {
      remainder = Math.IEEEremainder(ldivc, NON_BLACKS); // Removes whole color_cycles.
      if (remainder < 0)
        remainder += NON_BLACKS;
      else if (remainder == -0.0)
        remainder = 0;
      if (remainder == 0)
        remainder = NON_BLACKS;
    }
    return remainder + 1;
  }
  
  private Color interpolatedColor (double fullColorIndex)
    // Does a HSB interpolation between values of the colorModel256's values on either side 
    // of fullColorIndex.  Understands wrap around of those 254 colors.
    // fullColorIndex is in the range 1 thru 255.  1 represents black (i.e. divergence).
  {
    int lowIndex;
    int highIndex = ((int)Math.ceil(fullColorIndex));
    if (fullColorIndex <= NON_BLACKS || highIndex != 2) {
      lowIndex = ((int)Math.floor(fullColorIndex));
    } else {
      lowIndex = (int)NON_BLACKS + 1;  //Interpolate around the cycle, not with black.
    }
   
    IndexColorModel map256 = parent.colorModel256();
    int r1 = map256.getRed(lowIndex);
    int g1 = map256.getGreen(lowIndex);
    int b1 = map256.getBlue(lowIndex);
    int r2 = map256.getRed(highIndex);
    int g2 = map256.getGreen(highIndex);
    int b2 = map256.getBlue(highIndex);
    float[] hsb1 = Color.RGBtoHSB(r1, g1, b1, (new float[3]));
    float[] hsb2 = Color.RGBtoHSB(r2, g2, b2, (new float[3]));
    double interpol = Math.IEEEremainder(fullColorIndex, 1.0);
    if (interpol < 0)
      interpol += 1;
    else if (interpol == -0.0)
      interpol = 0.0;
    float newh = (float)Math.max(0.0, Math.min(1.0, hsb1[0] + ((hsb2[0] - hsb1[0]) * interpol)));
    float news = (float)Math.max(0.0, Math.min(1.0, hsb1[1] + ((hsb2[1] - hsb1[1]) * interpol)));
    float newb = (float)Math.max(0.0, Math.min(1.0, hsb1[2] + ((hsb2[2] - hsb1[2]) * interpol)));
    return Color.getHSBColor(newh, news, newb);
  }
  
  double pixout(int x, int y, double loops, int gdidx, boolean recordd)
    //loops is actually max-loops minus actual loops, so 0 means non-divergence and to be black.
    // This version takes a loops count that is a float, for DirectColorModel use.
  {
    if (spec_.colorsIs256_p) {
      return pixout(x, y, (int)loops, gdidx, recordd);
    }
    
    double ldivc = 0;
    if (loops != 0) {
      ldivc = Math.min(NON_BLACKS * spec_.color_cycles, loops / clump);
    } 
    // ldivc is in the range 0 through 254 * color_cycles.
    
    // Record the real value
    if (recordd) {
      guide.aset(gdidx, x, ldivc);
    }
    
    double colorIndex = findFullColorIndex(ldivc); 
    // colorIndex is in the range 1 to 255.
    Color newColor = interpolatedColor(colorIndex);
    
    int r3 = newColor.getRed();
    int g3 = newColor.getGreen();
    int b3 = newColor.getBlue();
    
    // Draw the point.
    int i = x + y * width;
    intPixels_[i] = (r3 << 16) + (g3 << 8) + b3;
    pixels_[i] = (byte)Math.round(colorIndex);
    
//    System.out.println("Q1ImageProducer::pixout loops=" + loops + " intPixels_[" + i + "]=" 
//        + intPixels_[i] + " r3=" + r3 + " g3=" + g3 + " b3=" + b3 + " colorIndx=" + colorIndex);
    
    // In the following line the returned value indicates whether the graphics instance is 
    // present.  So if the graphics instance is gone plan to stop.
    if (!parent.keepCurrentDrawLine(x, x, y, newColor))
      stop_p_ = true;
    return ldivc;
  }


  void frDrawline(int xleft, int xright, int yndx, double ldivc)
    // Full color version.
  {
    if (spec_.colorsIs256_p) {
      frDrawline(xleft, xright, yndx, (int)ldivc);
      return;
    }

    // ldivc is a number recorded by pixout, so we pick up the computations for
    // drawing from there
    
    double colorIndex = findFullColorIndex(ldivc); 
    Color newColor = interpolatedColor(colorIndex);
    
    int r3 = newColor.getRed();
    int g3 = newColor.getGreen();
    int b3 = newColor.getBlue();
    int pixel_value = (r3 << 16) + (g3 << 8) + b3;
    
    // Draw the horizontal line.
    int end = xright + yndx * width;
    for (int i = xleft + yndx * width; i <= end; i++) {
      intPixels_[i] = pixel_value;
      pixels_[i] = (byte)Math.round(colorIndex);
    }    
    
    parent.keepCurrentDrawLine(xleft, xright, yndx, newColor);
  }
  
  void rowSimple(int yndx, int gdidx, boolean record_p)
  {
    double loops;

    if (loop_function.bigLoop_p()) {
      BigQuaternion y = hiy.minus(incy.times_scalar(yndx));
      for(int x = 0; x < spec_.pix_width; x++) {
        BigQuaternion pointA = lox.plus(incx.times_scalar(x)).plus(y);
        loops = call_big_loop_function(pointA, spec_.coeff, spec_.trans, spec_.depth);
        if (spec_.colorsIs256_p)
          pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdidx, record_p);
        else
          pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdidx, record_p);
        if (stop_p_)
          return; 
      }
    } else {
      pointFast.assign_from(incyFast).times_scalar_n(yndx); //Used as a temp here.
      yFast.assign_from(hiyFast).minus_equal(pointFast);
      leftFast.assign_from(loxFast).plus_equal(yFast);
      coeffFast.assign_from(specCoeff);
      transFast.assign_from(specTrans);
      for(int x = 0; x < spec_.pix_width; x++) {
        xFast.assign_from(incxFast).times_scalar_n(x); //We multiply it out for precision.
        pointFast.assign_from(leftFast).plus_equal(xFast);
        loops = call_loop_function(pointFast, coeffFast, transFast, spec_.depth);
        if (spec_.colorsIs256_p)
          pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdidx, record_p);
        else
          pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdidx, record_p);
        if (stop_p_)
          return;
      }
    }
    last_row_drawn_ = yndx;
    try {
      Thread.sleep(0, 1);
    } catch (Throwable excpt) {
    }
  }
  
  int guidedFillLeft(BigQuaternion y, int xleft, int x, int yndx,
                       int gdtop, int gdbot)
  {
    //System.out.println("Q1ImageProducer::guidedFillLeft ent");
    int gdfill = (int)Math.floor((gdtop + gdbot) / 2);
    boolean recordd = (boolean)(gdtop < (gdbot - 1));
    double loops;
    double ldivc;

    x++;
    if (loop_function.bigLoop_p()) {
      BigQuaternion pointA = lox.plus(incx.times_scalar(x)).plus(y);
      do {
        x--;
        pointA.minus_equal(incx);
        loops = call_big_loop_function(pointA, spec_.coeff, spec_.trans, spec_.depth);
        if (spec_.colorsIs256_p)
          ldivc = pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
        else
          ldivc = pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
        } while (!((x <= xleft) ||
                   ((ldivc == guide.aref(gdtop, x)) &&
                    (ldivc == guide.aref(gdbot, x)) )));
    } else {
      leftFast.assign_from(loxFast).plus_equal(y.fastQuaternionValue()); //pt change 1 of 3.
//      xFast.assign_from(incxFast).times_scalar_n(x); //Used as a temp here.
//      pointFast.assign_from(incxFast).times_scalar_n(x).plus_equal(loxFast).plus_equal(y.fastQuaternionValue());
      coeffFast.assign_from(specCoeff);
      transFast.assign_from(specTrans);
      do {
        x--;
        xFast.assign_from(incxFast).times_scalar_n(x); //We multiply it out each time for precision.
        pointFast.assign_from(leftFast).plus_equal(xFast);
        loops = call_loop_function(pointFast, coeffFast, transFast, spec_.depth);
        if (spec_.colorsIs256_p)
          ldivc = pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
        else
          ldivc = pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
        } while (!((x <= xleft) ||
                   ((ldivc == guide.aref(gdtop, x)) &&
                    (ldivc == guide.aref(gdbot, x)) )));
    }
    //System.out.println("Q1ImageProducer::guidedFillLeft ex");
    return x;
  }

  int guidedFillRight(BigQuaternion y, int x, int yndx, int gdtop, int gdbot, int xmax_plus_1)
  {
    //System.out.println("Q1ImageProducer::guidedFillRight ent");
    int gdfill = (int)Math.floor((gdtop + gdbot) / 2);
    boolean recordd = (gdtop < (gdbot - 1));
    boolean behind_ok;
    boolean ahead_ok;
    double loops;
    double ldivc;
    x--;

    if (loop_function.bigLoop_p()) {
      BigQuaternion pointA = lox.plus(incx.times_scalar(x)).plus(y);
      do {
        x++;
        pointA.plus_equal(incx);
        loops = call_big_loop_function(pointA, spec_.coeff, spec_.trans, spec_.depth);
        if (spec_.colorsIs256_p)  //+++++++++++++++++++
          ldivc = pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
        else
          ldivc = pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
        behind_ok = (x == 0) || ((ldivc == guide.aref(gdtop, x - 1)) &&
                                 (ldivc == guide.aref(gdbot, x - 1)) );
        ahead_ok = (x == xmax_plus_1 - 1) || ((ldivc == guide.aref(gdtop, x + 1)) &&
                                         (ldivc == guide.aref(gdbot, x + 1)) );
        } while ((x < xmax_plus_1 - 1) &&
                 !(behind_ok &&
                   (ldivc == guide.aref(gdtop, x)) &&
                   (ldivc == guide.aref(gdbot, x)) &&
                   ahead_ok ));
    } else {
      leftFast.assign_from(loxFast).plus_equal(y.fastQuaternionValue()); //pt change 1 of 3.
//      pointFast.assign_from(incxFast).times_scalar_n(x).plus_equal(loxFast).plus_equal(y.fastQuaternionValue());
      coeffFast.assign_from(specCoeff);
      transFast.assign_from(specTrans);
      do {
        x++;
        xFast.assign_from(incxFast).times_scalar_n(x); //We multiply it out each time for precision.
        pointFast.assign_from(leftFast).plus_equal(xFast);
//        pointFast.plus_equal(incxFast);
        loops = call_loop_function(pointFast, coeffFast, transFast, spec_.depth);
        if (spec_.colorsIs256_p)  //+++++++++++++++++++
          ldivc = pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
        else
          ldivc = pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
        behind_ok = (x == 0) || ((ldivc == guide.aref(gdtop, x - 1)) &&
                                 (ldivc == guide.aref(gdbot, x - 1)) );
        ahead_ok = (x == xmax_plus_1 - 1) || ((ldivc == guide.aref(gdtop, x + 1)) &&
                                         (ldivc == guide.aref(gdbot, x + 1)) );
        } while ((x < xmax_plus_1 - 1) &&
                 !(behind_ok &&
                   (ldivc == guide.aref(gdtop, x)) &&
                   (ldivc == guide.aref(gdbot, x)) &&
                   ahead_ok ));
    }
    //System.out.println("Q1ImageProducer::guidedFillRight ex");
    return x + 1; // x is first where everything matches but is plotted already.
  }
  
  
  double singlePoint(BigQuaternion y, int x, int yndx, int gdtop, int gdbot)
    // This method computes and draws a single point and returns the loop count.
  {
    //System.out.println("Q1ImageProducer::singlePoint ent");
    int gdfill = 0; // If we are not recording we don't care.
    boolean recordd = (gdtop < (gdbot - 1));
    if (recordd) {
      gdfill = (int)Math.floor((gdtop + gdbot) / 2);
    }
    double loops;
    if (loop_function.bigLoop_p()) {
      BigQuaternion pointA = lox.plus(incx.times_scalar(x)).plus(y);
      loops = call_big_loop_function(pointA, spec_.coeff, spec_.trans, spec_.depth);
    } else {
      pointFast.assign_from(incxFast).times_scalar_n(x).plus_equal(loxFast).plus_equal(y.fastQuaternionValue());
      coeffFast.assign_from(specCoeff);
      transFast.assign_from(specTrans);
//      FastQuaternion yFast = y.fastQuaternionValue();
//      FastQuaternion point = loxFast.plus(incxFast.times_scalar(x)).plus(yFast);
//      FastQuaternion coeffFast = spec_.coeff.fastQuaternionValue();
//      FastQuaternion transFast = spec_.trans.fastQuaternionValue();
      loops = call_loop_function(pointFast, coeffFast, transFast, spec_.depth);
    }
    if (spec_.colorsIs256_p)  //+++++++++++++++++++
      return pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
    //System.out.println("Q1ImageProducer::singlePoint ex almost");
    return pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
  }
  

  int getRightMax(int gdtop, int gdbot, int xleft, int xmax_plus_1)
  {
    //System.out.println("Q1ImageProducer::getRightMax ent");
    double value = guide.aref(gdtop, xleft - 1);
    while ((xleft < xmax_plus_1) &&
           (guide.aref(gdtop, xleft) == value) &&
           (guide.aref(gdtop, xleft) == guide.aref(gdbot, xleft)) ) {
          xleft++; }
    //System.out.println("Q1ImageProducer::getRightMax ex");
    return xleft - 1;
  }
  

  void rowFancy(int yndx, int gdtop, int gdbot, int xmin, int xmax_plus_1)
  {
    //System.out.println("Q1ImageProducer::rowFancy en");
    if (stop_p_)
      return; 
    boolean recordd = (gdtop < (gdbot - 1));
    int gdfill = (int)Math.floor((gdtop + gdbot) / 2);
    BigQuaternion y = hiy.minus(incy.times_scalar(yndx));
    double ldivc;
    int xright;
    int right_max;
    for(int xleft = xmin; xleft < xmax_plus_1;) {
      xleft = guidedFillRight(y, xleft, yndx, gdtop, gdbot, xmax_plus_1);
      if (xleft < xmax_plus_1) {
        //System.out.println("Q1ImageProducer::rowFancy pt1");
        right_max = getRightMax(gdtop, gdbot, xleft, xmax_plus_1);
        xright = guidedFillLeft(y, xleft, right_max, yndx, gdtop, gdbot);
        ldivc = guide.aref(gdtop, xleft); //was right_max
        if (xright - xleft < 3) { //If there is a short gap to fill in just do it
          //System.out.println("Q1ImageProducer::rowFancy pt2");
          frDrawline(xleft, xright - 1, yndx, ldivc);
          if (recordd) {
            //System.out.println("Q1ImageProducer::rowFancy pt3");
            for(int x = xleft; x < xright; x++) {
              guide.aset(gdfill, x, ldivc); }}
        } else { // Longer gap, test a point in the middle.
          //System.out.println("Q1ImageProducer::rowFancy pt4");
          int mid_x = (xleft + xright) / 2;
          double mid_ldivc = singlePoint(y, xleft, yndx, gdtop, gdbot);
          if (ldivc == mid_ldivc) { // Mid-point is the same, fill in the gap.
            //System.out.println("Q1ImageProducer::rowFancy pt5");
            frDrawline(xleft, xright - 1, yndx, ldivc);
            if (recordd) {
              for(int x = xleft; x < xright; x++) {
                guide.aset(gdfill, x, ldivc); }}
          } else { // Mid-point is different, do it in halves.
            //System.out.println("Q1ImageProducer::rowFancy pt6");
            rowFancy(yndx, gdtop, gdbot, xleft, mid_x);
            rowFancy(yndx, gdtop, gdbot, mid_x, xright);
          }
        }
        xleft = right_max;
      }
    }
    last_row_drawn_ = yndx;
    try {
      Thread.sleep(0, 1);
    } catch (Throwable excpt) {
    }
    //System.out.println("Q1ImageProducer::rowFancy ex");
  }

  void doEight(int min_yndx)
  {
    rowSimple((min_yndx + 8), 4, true);
    rowFancy((min_yndx + 4), 0, 4, 0 , spec_.pix_width);
    rowFancy((min_yndx + 2), 0, 2, 0 , spec_.pix_width);
    rowFancy((min_yndx + 6), 2, 4, 0 , spec_.pix_width);
    rowFancy((min_yndx + 1), 0, 1, 0 , spec_.pix_width);
    rowFancy((min_yndx + 3), 1, 2, 0 , spec_.pix_width);
    rowFancy((min_yndx + 5), 2, 3, 0 , spec_.pix_width);
    rowFancy((min_yndx + 7), 3, 4, 0 , spec_.pix_width);
  }
  
  protected void computeLoxEtc ()
  {
    /* int desired_scale = spec_.correctScale(); */
    //System.out.println("Q1ImageProducer.computeLoxEtc center=" + spec_.center.toString() + 
    //                   " coeff=" + spec_.coeff.toString() + " trans=" + spec_.trans.toString()); 
    int h_inter_pixel_spaces = spec_.pix_width * spec_.panels - 1;
    int v_inter_pixel_spaces = spec_.pix_height * spec_.panels - 1;
    incx.assign_from(spec_.vx.times_scalar(spec_.range / h_inter_pixel_spaces).bigQuaternionValue());
    incy.assign_from(spec_.vy.times_scalar(spec_.range / /* yes h */ h_inter_pixel_spaces).bigQuaternionValue());
      
    int lox_pixels_rt_of_mid = h_inter_pixel_spaces / -2 + spec_.panel_x * spec_.pix_width;
    lox.assign_from(spec_.center.plus(incx.times_scalar(lox_pixels_rt_of_mid)));    //A position.
    int hiy_pixels_above_mid = v_inter_pixel_spaces / 2 - spec_.panel_y * spec_.pix_height;
    hiy.assign_from(incy.times_scalar(hiy_pixels_above_mid));         // A vector, not a position.
    
    /* lox.setScaleLessOrEqual(desired_scale);
    hiy.setScaleLessOrEqual(desired_scale); */

    loxFast = lox.fastQuaternionValue();      
    hiyFast = hiy.fastQuaternionValue();      
    incxFast = incx.fastQuaternionValue();      
    incyFast = incy.fastQuaternionValue();
         
    clump = spec_.depth / (double)(spec_.color_cycles * NON_BLACKS);//0=transp,1=blk
    //System.out.println("Q1ImageProducer.computeLoxEtc lox=" + lox.toString() + 
    //                   " hiy=" + hiy.toString()); 
  } 
   
  protected void hifiScan ()
  {
//    System.out.println("QIImageProducer hifiScan entry.");
    if (draw_in_progress_p_ ) {
      return;
    }
    draw_in_progress_p_ = true;
    if (last_row_drawn_ == -1) { // THEN NEED TO INITIALIZE
      computeLoxEtc();
    } else {
//      g.drawImage(refresher, h_offset_, v_offset_, parent.parent.main_background, this);
    }
    int row = last_row_drawn_ + 1;
    for (; row < spec_.pix_height; row++) {
      rowSimple(row, 0, false);
      if (stop_p_) {
        last_row_drawn_ = row - 1;
        parent.set_stop_requires_new_image(true);
        break;
      }
    }
  }
  

  protected void scan ()
    // this routine gets us ready to draw groups of 8 rows.  For a new picture this means drawing row 0.
    //  If we have stopped in the middle of a drawing it means drawing whatever is lacking.
    // When ready, we call scan_8 to finishe the job.
  {
//    System.out.println("FIP.scan entry draw_in_progress_p_ = " + draw_in_progress_p_);
    if (draw_in_progress_p_ ) {
      return;
    }
    draw_in_progress_p_ = true;
    int last_floor_eight;

    if (last_row_drawn_ == -1) { // THEN NEED TO INITIALIZE
      computeLoxEtc();
      rowSimple(0, 0, true);
      last_floor_eight = 0;
    } else {
      // NEED TO MAKE UP THE REMAINING ROWS IN CURRENT SET OF 8.
      last_floor_eight = (int)Math.floor((last_row_drawn_ - 1)/ 8); // last finished GROUP floor 8
      int base_row = 8 * last_floor_eight;
      int last_mod_eight = last_row_drawn_  % 8;
      if (((last_row_drawn_ >= (spec_.pix_height - 2)) && last_row_drawn_ != spec_.pix_height - 1) ||
          stop_p_) { // We have drawn it all.
        return; // Nothing more to draw.
      }
      if (base_row + 8 <= spec_.pix_height) {
        switch (last_mod_eight) {
          case 0: rowFancy((base_row + 4), 0, 4, 0 , spec_.pix_width);
          case 4: rowFancy((base_row + 2), 0, 2, 0 , spec_.pix_width);
          case 2: rowFancy((base_row + 6), 2, 4, 0 , spec_.pix_width);
          case 6: rowFancy((base_row + 1), 0, 1, 0 , spec_.pix_width);
          case 1: rowFancy((base_row + 3), 1, 2, 0 , spec_.pix_width);
          case 3: rowFancy((base_row + 5), 2, 3, 0 , spec_.pix_width);
          case 5: rowFancy((base_row + 7), 3, 4, 0 , spec_.pix_width);
          default:  // If we just did 7, we're ready to start a new group of 8.
        }
        last_floor_eight++;
      }
    }
    if (!stop_p_) {
      scan_8(last_floor_eight);
    }
    if (stop_p_)
      parent.set_stop_requires_new_image(true);
  }
  
  void scan_8 (int which_eight)
  {
    // Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
    //System.out.println("Q1ImageProducer priority = " + Thread.currentThread().getPriority());
    //System.out.println("FIP.scan_8 process name = " + Thread.currentThread().getName());
    for (; which_eight < (int)Math.floor((spec_.pix_height - 1)/ 8); which_eight++) {
      doEight(which_eight * 8);
      if (stop_p_)
        return; 
      guide.copy_row(4, 0, spec_.pix_width);
    }
    
    // We allow any height, so finish off whatever portion af another 8 is left.
    int rowsLeft = spec_.pix_height - (which_eight * 8) - 1;
    switch (rowsLeft) {
      case 1:
        rowSimple(which_eight * 8 + 1, 0, false);
        break;
      case 2:
        rowSimple(which_eight * 8 + 2, 1, true);
        rowFancy(which_eight * 8 + 1, 0, 1, 0 , spec_.pix_width);
        break;
      case 3:
        rowSimple(which_eight * 8 + 2, 1, true);
        rowFancy(which_eight * 8 + 1, 0, 1, 0 , spec_.pix_width);
        rowSimple(which_eight * 8 + 3, 0, false);
        break;
      case 4:
        rowSimple(which_eight * 8 + 4, 2, true);
        rowFancy(which_eight * 8 + 2, 0, 2, 0 , spec_.pix_width);
        rowFancy(which_eight * 8 + 1, 0, 1, 0 , spec_.pix_width);
        rowFancy(which_eight * 8 + 3, 1, 2, 0 , spec_.pix_width);
        break;
      case 5:
        rowSimple(which_eight * 8 + 4, 2, true);
        rowFancy(which_eight * 8 + 2, 0, 2, 0 , spec_.pix_width);
        rowFancy(which_eight * 8 + 1, 0, 1, 0 , spec_.pix_width);
        rowFancy(which_eight * 8 + 3, 1, 2, 0 , spec_.pix_width);
        rowSimple(which_eight * 8 + 5, 0, false);
        break;
      case 6:
        rowSimple(which_eight * 8 + 4, 2, true);
        rowFancy(which_eight * 8 + 2, 0, 2, 0 , spec_.pix_width);
        rowFancy(which_eight * 8 + 1, 0, 1, 0 , spec_.pix_width);
        rowFancy(which_eight * 8 + 3, 1, 2, 0 , spec_.pix_width);
        rowSimple(which_eight * 8 + 6, 3, true);
        rowFancy(which_eight * 8 + 5, 2, 3, 0 , spec_.pix_width);
        break;
      case 7:
        rowSimple(which_eight * 8 + 4, 2, true);
        rowFancy(which_eight * 8 + 2, 0, 2, 0 , spec_.pix_width);
        rowFancy(which_eight * 8 + 1, 0, 1, 0 , spec_.pix_width);
        rowFancy(which_eight * 8 + 3, 1, 2, 0 , spec_.pix_width);
        rowSimple(which_eight * 8 + 6, 3, true);
        rowFancy(which_eight * 8 + 5, 2, 3, 0 , spec_.pix_width);
        rowSimple(which_eight * 8 + 7, 0, false);
        break;
      default: break;
    }
  }
  
  public boolean lastRowDrawnP ()
    // This is used in autoadvance to determine if we are to advance.  We may not want to
    // if the user did a 'pause' or 'stop'.
  {
    if (last_row_drawn_ == -1) {
      return true;
    }
    if (spec_.hifi_p || spec_.pix_height % 2 == 0) {
      return last_row_drawn_ == spec_.pix_height - 1;
    }
    return (last_row_drawn_ == (spec_.pix_height - 2));
  }

    
  protected synchronized void drawTheFractal () 
  {
    Thread runner_ = new Thread(this, "FIP " + parent.who_are_you());
    runner_.setPriority(Thread.MIN_PRIORITY);
    runner_.start();
  }

  // Convert the contents of the input file to image data.
//  protected synchronized void drawTheFractal () 
  public void run ()
  {
    //System.out.println("Q1ImageProducer.drawTheFractal ever_drawn_p = " + ever_drawn_p + 
    //                   " parm_ch_p = " + parameter_change_p_ + " stop_p_ = " + stop_p_);
    // System.out.println("FIP.run process name = " + Thread.currentThread().getName());
    if (!ever_drawn_p && !parameter_change_p_) { //For before the first draw.
      return;
    }
    if (parameter_change_p_) {
      last_row_drawn_ = -1; // Indicating start with row 0, the beginning.
    }
        
    parameter_change_p_ = false;
    stop_p_ = false;
    if (spec_.hifi_p)
      hifiScan();
    else
      scan();
    ever_drawn_p = true;
    stop_p_ = false;
    draw_in_progress_p_ = false;
    setDrawingFractalInterlock_p(false);
  } 

  /*  ******************* Misc. utilities ************************ */

  public int colorIndexAt (int x, int y)
    // Returns the index into the color map of 255 colors (we don't use 0).
  {
    int index = x + width * y;
    return intify(pixels_[x + width * y]);
  }

  public void stop ()
  {
    stop_p_ = true;
  }

  public void setParameters (FractalSpec arg, boolean change_p)
     // This sets the internal parameters of this fractal.
     // CALLER SHOULD MAKE SURE ALL PARAMETERS ARE VALID!!!
  {
    spec_ = arg;
    specPoint = spec_.center.fastQuaternionValue();
    specCoeff = spec_.coeff.fastQuaternionValue();
    specTrans = spec_.trans.fastQuaternionValue();

    parameter_change_p_ = change_p;
    ever_drawn_p = !change_p;
    loop_function = parent.jQCentral().findLoopObject(spec_.which);
  }
  
  /*  ******************* ImageProducer implementation help ******************* */

    protected void produce(ImageConsumer ic) 
    {
      //System.out.println("produce entry. colorModel is of type " + parent.colorModel.getClass());
      if (width < 0) {
        ic.imageComplete(ImageConsumer.IMAGEERROR);
        return;
      } 
      ic.setDimensions(width, height);
      ic.setProperties(properties);
      ic.setColorModel(parent.colorModel);
      ic.setHints(ImageConsumer.SINGLEPASS |
                  ImageConsumer.SINGLEFRAME);
      if (spec_.colorsIs256_p)
        ic.setPixels(0, 0, width, height, parent.colorModel, pixels_, 0, width);
      else
        ic.setPixels(0, 0, width, height, parent.colorModel, intPixels_, 0, width);
      ic.imageComplete(ImageConsumer.STATICIMAGEDONE);
    }
    
  /*  ******************* ImageProducer implementation ************************ */

  public synchronized void startProduction(ImageConsumer ic)
    // This is one of the two interface methods for ImageProducer, other than add,is,remove consumer.
    //  It guarantees that drawTheBox sets up the picture before produce is called
    //  to make settings in the ImageConsumer(s).
  {
    //System.out.println("Q1ImageProducer.startProduction entry");
    if (ic != null) {
      addConsumer(ic);
      int last_row = spec_.pix_height - 1 - (spec_.hifi_p ? 0 : 1);
      // We used to drawTheFractal here.  Now the JQCanvas initiates that.
      int temp;
      if ((temp = consumers.size()) == 0) {
        return; //Following loop does not work properly.
      }
      for (int i = consumers.size(); i>0; i--) {
        if (i - 1 >= consumers.size()) // synchronized does not work.
          continue;
        ic = (ImageConsumer)consumers.elementAt(i - 1);
        produce(ic);
        consumers.removeElement(ic);
      }
    }
    parent.startProductionFinished();
  }

  public void requestTopDownLeftRightResend(ImageConsumer ic) 
  {
//    System.out.println("Q1ImageProducer::requestTopDownLeftRightResend entry ic=" + ic);
    //System.out.println("FIP::requestTopDownLeftRightResend entry. colorModel is of type " +
    //                    parent.colorModel.getClass());
    if (spec_.colorsIs256_p ? (pixels_ != null) : (intPixels_ != null)) {
      //System.out.println("Q1ImageProducer::requestTopDownLeftRightResend IF entry");
      ic.setDimensions(width, height);
      ic.setProperties(properties);
      ic.setColorModel(parent.colorModel());
      ic.setHints(ImageConsumer.SINGLEPASS |
                  ImageConsumer.TOPDOWNLEFTRIGHT |
                  ImageConsumer.SINGLEFRAME);
      if (spec_.colorsIs256_p)
        ic.setPixels(0, 0, width, height, parent.colorModel, pixels_, 0, width);
      else
        ic.setPixels(0, 0, width, height, parent.colorModel, intPixels_, 0, width);
    }
  }

  public synchronized void addConsumer (ImageConsumer ic) 
  {
    if (ever_drawn_p) {
      produce(ic);
    } else if (!consumers.contains(ic)) {
      consumers.addElement(ic);
    }
  }

  public synchronized boolean isConsumer (ImageConsumer ic) 
  {
    return consumers.contains(ic);
  }

  public synchronized void removeConsumer (ImageConsumer ic) {
    consumers.removeElement(ic);
  }
    
}


