package JavaQuat20;

import java.awt.*;
import java.awt.image.*;
import java.util.*;
import JavaQuat20.FractalLoops.*;

class Q2ImageProducer extends Q1ImageProducer 
{
  Q2ImageProducer(FractalSpec arg, JQCanvas p) 
    // This is for normal creation when we are going to compute a fractal
  {
    super(arg, p);
  }
   
  Q2ImageProducer(Image refresher, FractalSpec arg, JQCanvas p)
    // This is for dummy creation, when we have loaded an image file, GIF or PNG.
  {
    super(refresher, arg, p);
  }
  
  public double call_loop_function(FastQuaternion point, FastQuaternion coeff_in,  //+++++++++++++++++++
                                   FastQuaternion trans, int depth)
  {
    if (spec_.pts_per_pixel == 1) {
      return (double)loop_function.loop(point, coeff_in, trans, depth);
    }
    FastQuaternion coeff = new FastQuaternion(coeff_in);
    Vector answers = new Vector(9);
    double loops = 0;
    if (spec_.pts_per_pixel == 2) {
      FastQuaternion halfx = incxFast.times_scalar(0.5);
      FastQuaternion halfy = incyFast.times_scalar(0.5);
      loops = loop_function.loop(point, coeff, trans, depth);
      answers.addElement(new Double(loops));
      coeff.plus_equal(halfx).plus_equal(halfy);
      loops = loop_function.loop(point, coeff, trans, depth);
      answers.addElement(new Double(loops));
    } else if (spec_.pts_per_pixel <= 4) {
      FastQuaternion halfx = incxFast.times_scalar(0.5);
      FastQuaternion halfy = incyFast.times_scalar(0.5);
      FastQuaternion coeff2 = coeff.plus(halfx);
      for (int i = 0; ; i++) {
        loops = loop_function.loop(point, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.loop(point, coeff2, trans, depth);
        answers.addElement(new Double(loops));
        if (i == 1) {
          break;
        }
        coeff.plus_equal(halfy);
        coeff2.plus_equal(halfy);
      }
    } else if (spec_.pts_per_pixel > 4) { //that would be 9
      FastQuaternion thirdx = incxFast.times_scalar(0.333333333333);
      FastQuaternion thirdy = incyFast.times_scalar(0.333333333333);
      FastQuaternion coeff2 = coeff.plus(thirdx);
      FastQuaternion coeff3 = coeff2.plus(thirdx);
      for (int i = 0; ; i++) {
        loops = loop_function.loop(point, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.loop(point, coeff2, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.loop(point, coeff3, trans, depth);
        answers.addElement(new Double(loops));
        if (i == 2) {
          break;
        }
        coeff.plus_equal(thirdy);
        coeff2.plus_equal(thirdy);
        coeff3.plus_equal(thirdy);
      }      
    } 
    int numAnswers = answers.size();
    if (numAnswers <= 0) {
      parent.message("Q2ImageProducer::call_loop_function error: " + 
                     spec_.pts_per_pixel, true);
      return (double)loop_function.loop(point, coeff, trans, depth);;
    }
    // Find the average of answers.
    loops = 0;
    for (int j = 0; j < numAnswers; j++) {
      loops += ((Double)answers.elementAt(j)).doubleValue();
    }
    double average = loops / numAnswers;
    // Find answer closest to average and return it.
    loops = ((Double)answers.firstElement()).doubleValue();
    for (int k = 1; k < numAnswers; k++) {
      if (Math.abs(((Double)answers.elementAt(k)).doubleValue() - average) < 
          Math.abs(loops - average)) {
        loops = ((Double)answers.elementAt(k)).doubleValue();
      }
    }
    if (spec_.colorsIs256_p) {
      return Math.round(loops);
    }
    return loops;
  }
  
  public double call_big_loop_function(BigQuaternion point, BigQuaternion coeff_in,  //+++++++++++++++++++
                                   BigQuaternion trans, int depth)
  {
    if (spec_.pts_per_pixel == 1) {
        return (double)loop_function.bigLoop(point, coeff_in, trans, depth);
    }
    BigQuaternion coeff = new BigQuaternion(coeff_in);
    Vector answers = new Vector(9);
    double loops = 0;
    if (spec_.pts_per_pixel == 2) {
      BigQuaternion halfx = incx.times_scalar(0.5);
      BigQuaternion halfy = incy.times_scalar(0.5);
      loops = loop_function.bigLoop(point, coeff, trans, depth);
      answers.addElement(new Double(loops));
      coeff.plus_equal(halfx).plus_equal(halfy);
      loops = loop_function.bigLoop(point, coeff, trans, depth);
      answers.addElement(new Double(loops));
    } else if (spec_.pts_per_pixel <= 4) {
      BigQuaternion halfx = incx.times_scalar(0.5);
      BigQuaternion halfy = incy.times_scalar(0.5);
      BigQuaternion coeff2 = coeff.plus(halfx);
      for (int i = 0; ; i++) {
        loops = loop_function.bigLoop(point, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.bigLoop(point, coeff2, trans, depth);
        answers.addElement(new Double(loops));
        if (i == 1) {
          break;
        }
        coeff.plus_equal(halfy);
        coeff2.plus_equal(halfy);
      }
    } else if (spec_.pts_per_pixel > 4) { //that would be 9
      BigQuaternion thirdx = incx.times_scalar(0.333333333333);
      BigQuaternion thirdy = incy.times_scalar(0.333333333333);
      BigQuaternion coeff2 = coeff.plus(thirdx);
      BigQuaternion coeff3 = coeff2.plus(thirdx);
      for (int i = 0; ; i++) {
        loops = loop_function.bigLoop(point, coeff, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.bigLoop(point, coeff2, trans, depth);
        answers.addElement(new Double(loops));
        loops = loop_function.bigLoop(point, coeff3, trans, depth);
        answers.addElement(new Double(loops));
        if (i == 2) {
          break;
        }
        coeff.plus_equal(thirdy);
        coeff2.plus_equal(thirdy);
        coeff3.plus_equal(thirdy);
      }      
    } 
    int numAnswers = answers.size();
    if (numAnswers <= 0) {
      parent.message("Q2ImageProducer::call_big_loop_function error: " + 
                     spec_.pts_per_pixel, true);
      return (double)loop_function.bigLoop(point, coeff, trans, depth);
    }
    // Find the average of answers.
    loops = 0;
    for (int j = 0; j < numAnswers; j++) {
      loops += ((Double)answers.elementAt(j)).doubleValue();
    }
    double average = loops / numAnswers;
    // Find answer closest to average and return it.
    loops = ((Double)answers.firstElement()).doubleValue();
    for (int k = 1; k < numAnswers; k++) {
      if (Math.abs(((Double)answers.elementAt(k)).doubleValue() - average) < 
          Math.abs(loops - average)) {
        loops = ((Double)answers.elementAt(k)).doubleValue();
      }
    }
    if (spec_.colorsIs256_p) {
      return Math.round(loops);
    }
    return loops;
  }
  
  /*  ******************* Drawing Primitives ************************ */

  void rowSimple(int yndx, int gdidx, boolean record_p)
  {
    double loops;  //+++++++++++++++++++

    if (loop_function.bigLoop_p()) {
      BigQuaternion y = hiy.minus(incy.times_scalar(yndx));
      for(int x = 0; x < spec_.pix_width; x++) {
        BigQuaternion point = lox.plus(incx.times_scalar(x)).plus(y);
        loops = call_big_loop_function(spec_.center, point, spec_.trans, spec_.depth); // $$$$$$
        if (spec_.colorsIs256_p)      //++++++++++++++++
          pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdidx, record_p);
        else
          pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdidx, record_p);
        if (stop_p_)
          return; 
      }
    } else {
      coeffFast.assign_from(incyFast).times_scalar_n(yndx); //Used as a temp here.
      yFast.assign_from(hiyFast).minus_equal(coeffFast);
      leftFast.assign_from(loxFast).plus_equal(yFast);
      coeffFast.assign_from(leftFast);
      pointFast.assign_from(specPoint);
      transFast.assign_from(specTrans);
      for(int x = 0; x < spec_.pix_width; x++) {
        xFast.assign_from(incxFast).times_scalar_n(x); //We multiply it out for precision.
        coeffFast.assign_from(leftFast).plus_equal(xFast);
        loops = call_loop_function(pointFast, coeffFast, transFast, spec_.depth);
        if (spec_.colorsIs256_p)
          pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdidx, record_p);
        else
          pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdidx, record_p);
        if (stop_p_)
          return; 
      }
    }
    last_row_drawn_ = yndx;
    try {
      Thread.sleep(0, 1);
    } catch (Throwable excpt) {
    }
  }
  
  int guidedFillLeft(BigQuaternion y, int xleft, int x, int yndx,
                       int gdtop, int gdbot)
  {
    int gdfill = (int)Math.floor((gdtop + gdbot) / 2);
    boolean recordd = (boolean)(gdtop < (gdbot - 1));
    double loops;       //++++++++++++++++++++++
    double ldivc;

    x++;
    if (loop_function.bigLoop_p()) {
      BigQuaternion point = lox.plus(incx.times_scalar(x)).plus(y);
      do {
        x--;
        point.minus_equal(incx);
        loops = call_big_loop_function(spec_.center, point, spec_.trans, spec_.depth); // $$$$$$
        if (spec_.colorsIs256_p)  //+++++++++++++++++++
          ldivc = pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
        else
          ldivc = pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
        } while (!((x <= xleft) ||
                   ((ldivc == guide.aref(gdtop, x)) &&
                    (ldivc == guide.aref(gdbot, x)) )));
    } else {
      pointFast.assign_from(specPoint);
      leftFast.assign_from(loxFast).plus_equal(y.fastQuaternionValue()); //pt change 1 of 3.
      transFast.assign_from(specTrans);
      do {
        x--;
        xFast.assign_from(incxFast).times_scalar_n(x); //We multiply it out each time for precision.
        coeffFast.assign_from(leftFast).plus_equal(xFast);
        loops = call_loop_function(pointFast, coeffFast, transFast, spec_.depth); // $$$$$$
        if (spec_.colorsIs256_p)  //+++++++++++++++++++
          ldivc = pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
        else
          ldivc = pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
        } while (!((x <= xleft) ||
                   ((ldivc == guide.aref(gdtop, x)) &&
                    (ldivc == guide.aref(gdbot, x)) )));
    }
    return x;
  }

  int guidedFillRight(BigQuaternion y, int x, int yndx, int gdtop, int gdbot, int xmax_plus_1)
  {
    int gdfill = (int)Math.floor((gdtop + gdbot) / 2);
    boolean recordd = (gdtop < (gdbot - 1));
    boolean behind_ok;
    boolean ahead_ok;
    double loops;        //+++++++++++++++++++
    double ldivc;
    x--;

    if (loop_function.bigLoop_p()) {
      BigQuaternion point = lox.plus(incx.times_scalar(x)).plus(y);
      do {
        x++;
        point.plus_equal(incx);
        loops = call_big_loop_function(spec_.center, point, spec_.trans, spec_.depth); // $$$$$$
        if (spec_.colorsIs256_p)  //+++++++++++++++++++
          ldivc = pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
        else
          ldivc = pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
        behind_ok = (x == 0) || ((ldivc == guide.aref(gdtop, x - 1)) &&
                                 (ldivc == guide.aref(gdbot, x - 1)) );
        ahead_ok = (x == xmax_plus_1 - 1) || ((ldivc == guide.aref(gdtop, x + 1)) &&
                                         (ldivc == guide.aref(gdbot, x + 1)) );
        } while ((x < xmax_plus_1 - 1) &&
                 !(behind_ok &&
                   (ldivc == guide.aref(gdtop, x)) &&
                   (ldivc == guide.aref(gdbot, x)) &&
                   ahead_ok ));
    } else {
      pointFast.assign_from(specPoint);
      leftFast.assign_from(loxFast).plus_equal(y.fastQuaternionValue());
      transFast.assign_from(specTrans);
      do {
        x++;
        xFast.assign_from(incxFast).times_scalar_n(x); //We multiply it out each time for precision.
        coeffFast.assign_from(leftFast).plus_equal(xFast);
        loops = call_loop_function(pointFast, coeffFast, transFast, spec_.depth); // $$$$$$
        if (spec_.colorsIs256_p)  //+++++++++++++++++++
          ldivc = pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
        else
          ldivc = pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
        behind_ok = (x == 0) || ((ldivc == guide.aref(gdtop, x - 1)) &&
                                 (ldivc == guide.aref(gdbot, x - 1)) );
        ahead_ok = (x == xmax_plus_1 - 1) || ((ldivc == guide.aref(gdtop, x + 1)) &&
                                         (ldivc == guide.aref(gdbot, x + 1)) );
        } while ((x < xmax_plus_1 - 1) &&
                 !(behind_ok &&
                   (ldivc == guide.aref(gdtop, x)) &&
                   (ldivc == guide.aref(gdbot, x)) &&
                   ahead_ok ));
    }
    return x + 1; // x is first where everything matches but is plotted already.
  }
  
  
  double singlePoint(BigQuaternion y, int x, int yndx, int gdtop, int gdbot) //++++++++++++++++
  {
    int gdfill = 0; // If we are not recording we don't care.
    boolean recordd = (gdtop < (gdbot - 1));
    if (recordd) {
      gdfill = (int)Math.floor((gdtop + gdbot) / 2);
    }
    double loops;
    if (loop_function.bigLoop_p()) {
      BigQuaternion point = lox.plus(incx.times_scalar(x)).plus(y);
      loops = call_big_loop_function(spec_.center, point, spec_.trans, spec_.depth); // $$$$$$
    } else {
      pointFast.assign_from(specPoint);
      coeffFast.assign_from(incxFast).times_scalar_n(x).plus_equal(loxFast).plus_equal(y.fastQuaternionValue());
      transFast.assign_from(specTrans);
//      FastQuaternion yFast = y.fastQuaternionValue();
//      FastQuaternion point = loxFast.plus(incxFast.times_scalar(x)).plus(yFast);
//      FastQuaternion pointFast = spec_.center.fastQuaternionValue();
//      FastQuaternion transFast = spec_.trans.fastQuaternionValue();
      loops = call_loop_function(pointFast, coeffFast, transFast, spec_.depth); // $$$$$$
    }
    if (spec_.colorsIs256_p)  //+++++++++++++++++++
      return pixout(x, yndx, reverseLoops256(spec_.depth , loops), gdfill, recordd);
    return pixout(x, yndx, reverseLoopsMillions(spec_.depth , loops), gdfill, recordd);
  }
  
  protected void computeLoxEtc ()
  {
    /* int desired_scale = spec_.correctScale(); */
    int h_inter_pixel_spaces = spec_.pix_width * spec_.panels - 1;
    int v_inter_pixel_spaces = spec_.pix_height * spec_.panels - 1;
    incx.assign_from(spec_.vx.times_scalar(spec_.range / h_inter_pixel_spaces).bigQuaternionValue());
    incy.assign_from(spec_.vy.times_scalar(spec_.range / /* yes h */ h_inter_pixel_spaces).bigQuaternionValue());
      
    int lox_pixels_rt_of_mid = h_inter_pixel_spaces / -2 + spec_.panel_x * spec_.pix_width;
    lox.assign_from(spec_.coeff.plus(incx.times_scalar(lox_pixels_rt_of_mid)));    //A position.  $$$$$$
    int hiy_pixels_above_mid = v_inter_pixel_spaces / 2 - spec_.panel_y * spec_.pix_height;
    hiy.assign_from(incy.times_scalar(hiy_pixels_above_mid));         // A vector, not a position.
    
    /* lox.setScaleLessOrEqual(desired_scale);
    hiy.setScaleLessOrEqual(desired_scale); */

    loxFast = lox.fastQuaternionValue();      
    hiyFast = hiy.fastQuaternionValue();      
    incxFast = incx.fastQuaternionValue();      
    incyFast = incy.fastQuaternionValue();      
      
    clump = spec_.depth / (double)(spec_.color_cycles * NON_BLACKS);//0=transp,1=blk
  }   

  protected void scan ()
    // this routine gets us ready to draw groups of 8 rows.  For a new picture this means drawing row 0.
    //  If we have stopped in the middle of a drawing it means drawing whatever is lacking.
    // When ready, we call scan_8 to finishe the job.
  {
//    System.out.println("FIP.scan entry draw_in_progress_p_ = " + draw_in_progress_p_);
    if (draw_in_progress_p_ ) {
      return;
    }
    draw_in_progress_p_ = true;
    int last_floor_eight;

    if (last_row_drawn_ == -1) { // THEN NEED TO INITIALIZE
      computeLoxEtc();
      rowSimple(0, 0, true);
      last_floor_eight = 0;
    } else {
      // NEED TO MAKE UP THE REMAINING ROWS IN CURRENT SET OF 8.
      last_floor_eight = (int)Math.floor((last_row_drawn_ - 1)/ 8); // last finished GROUP floor 8
      int base_row = 8 * last_floor_eight;
      int last_mod_eight = last_row_drawn_  % 8;
      if (((last_row_drawn_ >= (spec_.pix_height - 2)) && last_row_drawn_ != spec_.pix_height - 1) ||
          stop_p_) { // We have drawn it all.
        draw_in_progress_p_ = false;
        return; // Nothing more to draw.
      }
      if (base_row + 8 <= spec_.pix_height) {
        switch (last_mod_eight) {
          case 0: rowFancy((base_row + 4), 0, 4, 0 , spec_.pix_width);
          case 4: rowFancy((base_row + 2), 0, 2, 0 , spec_.pix_width);
          case 2: rowFancy((base_row + 6), 2, 4, 0 , spec_.pix_width);
          case 6: rowFancy((base_row + 1), 0, 1, 0 , spec_.pix_width);
          case 1: rowFancy((base_row + 3), 1, 2, 0 , spec_.pix_width);
          case 3: rowFancy((base_row + 5), 2, 3, 0 , spec_.pix_width);
          case 5: rowFancy((base_row + 7), 3, 4, 0 , spec_.pix_width);
          default:  // If we just did 7, we're ready to start a new group of 8.
        }
        last_floor_eight++;
      }
    }
    if (!stop_p_) {
      scan_8(last_floor_eight);
    }
    if (stop_p_)
      parent.set_stop_requires_new_image(true);
    draw_in_progress_p_ = false;
  }
}


