package JavaQuat20;

public class SeqSpec
 // This is the Zoom, Pivot, and Shift dialog.
{
  public static final int ZOOM = 0;
  public static final int PIVOT = 1;
  public static final int SHIFT = 2;
  
  public static final int NON_BLACKS = 254;
  
  public int choice = ZOOM;
  public boolean enabled_p = false;
  public boolean pause_p = false; // As opposed to write image and ocontinue.
  public boolean xFirstInPathname_p = false; // False means alpha sort by row then col.
  public boolean advancePanelsBackwards_p = false;
  public boolean advanceRowByRow_p = true;
  public String destination = "";
  public int frames = 1;
  public double degrees = 180;
  public double shift = 0;
  private int frameNumber_ = 0;
  private boolean horiz_p_ = false; // For pivot direction.
  
  private FractalSpec startSpec_ = null;
  private FractalSpec endSpec_ = null;
//  private int steps_ = 0;
  
  public SeqSpec ()
  {
  }
  
  public SeqSpec (SeqSpec arg, boolean start_too_p, boolean end_too_p)
  {
    choice = arg.choice;
    enabled_p = arg.enabled_p;
    pause_p = arg.pause_p; // As opposed to write image and ocontinue.
    destination = arg.destination;
    frames = arg.frames;
    degrees = arg.degrees;
    shift = arg.shift;
    if (start_too_p) {
      if (arg.startSpec() == null) {
        startSpec_ =  null;
      } else {
        if (startSpec_ == null)
          startSpec_ = new FractalSpec();
        startSpec_.assignFrom(arg.startSpec());
      }
    }
    if (end_too_p) {
      if (arg.endSpec() == null) {
        endSpec_ =  null;
      } else {
        if (endSpec_ == null)
          endSpec_ = new FractalSpec();
        endSpec_.assignFrom(arg.endSpec());
      }
    }
  }
  
  public void assignFrom (SeqSpec arg, boolean start_too_p, boolean end_too_p)
  {
    choice = arg.choice;
    enabled_p = arg.enabled_p;
    pause_p = arg.pause_p; // As opposed to write image and ocontinue.
    // these don't seem to be needed.
    //xFirstInPathname_p = arg.xFirstInPathname_p;
    //advancePanelsBackwards_p = arg.advancePanelsBackwards_p;
    //advanceRowByRow_p = arg.advanceRowByRow_p;
    destination = arg.destination;
    frames = arg.frames;
    degrees = arg.degrees;
    shift = arg.shift;
    horiz_p_ = arg.horiz_p_;
    frameNumber_ = arg.frameNumber_;
    
//    steps_ = arg.steps_;
    if (start_too_p) {
      if (arg.startSpec() == null) {
        startSpec_ =  null;
      } else {
        if (startSpec_ == null)
          startSpec_ = new FractalSpec();
        startSpec_.assignFrom(arg.startSpec());
      }
    }
    if (end_too_p) {
      if (arg.endSpec() == null) {
        endSpec_ =  null;
      } else {
        if (endSpec_ == null)
          endSpec_ = new FractalSpec();
        endSpec_.assignFrom(arg.endSpec());
      }
    }
  }
  
  public boolean horizontalPivotP ()
  {
    return horiz_p_;
  }
  
  public void setHorizontalPivotP (boolean arg)
  {
    horiz_p_ = arg;
  }
  
  public FractalSpec startSpec ()
  {
    return startSpec_;
  }
  
  public FractalSpec endSpec ()
  {
    return endSpec_;
  }
  
  public void setStartSpec (FractalSpec arg)
  {
    if (arg == null) {
      startSpec_ = null;
      return;
    }
    if (startSpec_ == null)
      startSpec_ = new FractalSpec();
    startSpec_.assignFrom(arg) ;
  }
  
  public void setEndSpec (FractalSpec arg)
  {
    if (arg == null) {
      endSpec_ = null;
      return;
    }
    if (endSpec_ == null)
      endSpec_ = new FractalSpec();
    endSpec_.assignFrom(arg);
  }
  
//  public void setSteps (int arg)
//  {
//    steps_ = arg;
//  }
  
  public boolean readyP (Messenger msg)
  { 
    // If no destination we should prompt for one with each frame.  Also we may
    // be pausing for screen capture and not need a pathname.
    if (!pause_p && destination.length() < 2) {
      msg.message("Need to specify destination pathname.", false);
      return false;
    }
    if (!enabled_p) {
      msg.message("Sequence is not initialized.", false);
      return false;
    }
    if (startSpec_ == null) {
      msg.message("First frame is not initialized.", false);
      return false;
    }
    if (frames <= 1 && startSpec_.panels <= 1) {
      msg.message("One sequence frame and one panel is not a sequence.", false);
      return false;
    }

    if (frames == 1) {
      return true; // From test above, we have panels.
    }
    if (choice == ZOOM) {
      if (endSpec_ == null)
        msg.message("No final frame chosen for ZOOM sequence.", false);
      return endSpec_ != null;
    } else if (choice == PIVOT) {
      if (degrees == 0)
        msg.message("Total degrees of PIVOT = 0, so no sequence.", false);
      return (degrees != 0);
    } else if (choice == SHIFT) {
      if (shift == 0)
        msg.message("Total SHIFT = 0, so no sequence.", false);
      return (shift != 0);
    }
    msg.message("Internal error in sequence type choice.", false);
    return false;
  }
  
  public void setToFrame (int arg)
  {
    frameNumber_ = arg;
  }
  
  public boolean destinationURLP ()
  {
    String lower = destination.toLowerCase();
    return (lower.startsWith("http://") ||
            lower.startsWith("ftp://") ||
            lower.startsWith("file://") );
  }
  
  private String ZeroPaddedIntString (int valueNow, int max)
  {
    StringBuffer buf = new StringBuffer(8);
    int max_digits = (int)Math.floor(Math.log(max) / Math.log(10.0));
    int digits_now;
    if (valueNow == 0) {
      digits_now = 0; //Both these numbers are really 1 short of the number of digits.
    } else if (valueNow < 0) {
      digits_now = 1 + (int)Math.floor(Math.log(-valueNow) / Math.log(10.0));
    } else {
      digits_now = (int)Math.floor(Math.log(valueNow) / Math.log(10.0));
    }
    for (int i = 0; i < max_digits - digits_now; i++) {
      buf.append('0');
    }
    buf.append(valueNow);
    return buf.toString();
  }
  
  public String pathname (FractalSpec arg)
  {
    StringBuffer buf = new StringBuffer(8);
    if (arg.panels > 1) { 
      if (xFirstInPathname_p) {     
        buf.append("x");
        buf.append(ZeroPaddedIntString(arg.panel_x, arg.panels));
        buf.append("-");
      }
      buf.append("y");
      buf.append(ZeroPaddedIntString(arg.panel_y, arg.panels));
      if (!xFirstInPathname_p) { 
        buf.append("-x");
        buf.append(ZeroPaddedIntString(arg.panel_x, arg.panels));
      }
    } 
    if (frames > 1) {
      if (choice == ZOOM) {
        buf.append("z");
      } else if (choice == PIVOT) {
        buf.append("p");
      } else if (choice == SHIFT) {
        buf.append("s");
      }
      buf.append(ZeroPaddedIntString(frameNumber_, frames));
    }
    buf.append(".gif");
    return (destination + buf.toString());
  }
  
  /* ------------------------ AutoAdvance Sequence methods ------------------- */
  
  public boolean advance (FractalSpec values, Messenger jQCentral)
    // This is for the auto advance.
  {
    boolean debug = false;
    
    if (startSpec_.panels <= 1 && !enabled_p) {
      if (debug) System.out.println("SeqSpec.advance exit 1f");
      return false;
    }
    if (advanceThePanel(values, jQCentral)) { // Returns false on panel (0, 0).
      if (debug) System.out.println("SeqSpec.advance exit 2t");
      return true;
    }
    frameNumber_++;
    if (frameNumber_ >= frames) {
//      System.out.println("SeqSpec.advance frameNumber_=" + frameNumber_ + " frames =" +
//                         frames);
      if (debug) System.out.println("SeqSpec.advance exit 3f");
      return false; // We are done.
    }
    if (choice == ZOOM && endSpec_ != null) {
      if (endSpec_ != null) {
        if (debug) System.out.println("SeqSpec.advance exit 4");
        return advanceTheZoom(values, jQCentral);
      }
      System.err.println("No final frame given for zoom sequence.");
      return false;
    } else if (choice == PIVOT) {
      if (degrees != 0) {
        if (debug) System.out.println("SeqSpec.advance exit 5");
        return advanceThePivot(values, jQCentral);
      }
      System.err.println("Zero degrees given for pivot, stopping.");
      return false;
    } else if (choice == SHIFT) {
      if (shift != 0) {
        if (debug) System.out.println("SeqSpec.advance exit 6");
        return advanceTheShift(values, jQCentral);
      }
      System.err.println("Zero degrees given for pivot, stopping.");
      return false;
    }
    jQCentral.message("Internal error in SeqSpec.advance: choice = " + choice, true);
    return false;
  }
  
   public boolean manualStep (FractalSpec values, int frameNumber, Messenger jQCentral)
    // This is for manual advance.  It skips the panel advance since this can be done by hand.
    // We also can advance without bounds, before 0 or after the last step.  We advance to
    // argument frameNumber.
  {
    //System.out.println("SeqSpec.manualStep entry.");
    if ((startSpec_ == null || startSpec_.panels <= 1) && !enabled_p) {
      return false;
    }
    frameNumber_ = frameNumber;
    values.assign_from(startSpec_);
    //System.out.println("SeqSpec.manualStep frameNumber_ set to " + frameNumber_);
    if (choice == ZOOM) {
      if (endSpec_ != null) {
        //System.out.println("SeqSpec.manualStep about to advanceThePanel.");
        return advanceTheZoom(values, jQCentral);
      }
      System.err.println("No final frame given for zoom sequence.");
      return false;
    } else if (choice == PIVOT) {
      if (degrees != 0) {
        return advanceThePivot(values, jQCentral);
      }
      System.err.println("Zero degrees given for pivot, stopping sequence.");
      return false;
    } else if (choice == SHIFT) {
      if (shift != 0) {
        return advanceTheShift(values, jQCentral);
      }
      System.err.println("Zero increment given for shift, stopping sequence.");
      return false;
    }
    jQCentral.message("Internal error in SeqSpec.advance: choice = " + choice, true);
    return false;
  }
  
  private boolean advanceThePanel (FractalSpec values, Messenger jQCentral)
    // Modifies the parameter and returns
    // Returns false if there are not multiple panels in the picture or we get to (0, 0).
  {
    if (startSpec_ == null || startSpec_.panels <= 1) {
      return false;
    }
    if (advanceRowByRow_p) {
      if (advancePanelsBackwards_p) {
        if (values.panel_x > 0) {
          values.panel_x--;
        } else if (values.panel_y > 0) {
          values.panel_y--;
          values.panel_x = startSpec_.panels - 1;
        } else { 
          values.panel_y = startSpec_.panels - 1;
          values.panel_x = startSpec_.panels - 1;
          return false; // We are done, or if zooming we are back to first panel.
        }
      } else {
        if (values.panel_x < startSpec_.panels - 1) {
          values.panel_x++;
        } else if (values.panel_y < startSpec_.panels - 1) {
          values.panel_y++;
          values.panel_x = 0;
        } else { 
          values.panel_y = 0;
          values.panel_x = 0;
          return false; // We are done, or if zooming we are back to first panel.
        }
      }
    } else { // Advancing Column by Column
      if (advancePanelsBackwards_p) {
        if (values.panel_y > 0) {
          values.panel_y--;
        } else if (values.panel_x > 0) {
          values.panel_x--;
          values.panel_y = startSpec_.panels - 1;
        } else { 
          values.panel_x = startSpec_.panels - 1;
          values.panel_y = startSpec_.panels - 1;
          return false; // We are done, or if zooming we are back to first panel.
        }
      } else {
        if (values.panel_y < startSpec_.panels - 1) {
          values.panel_y++;
        } else if (values.panel_x < startSpec_.panels - 1) {
          values.panel_x++;
          values.panel_y = 0;
        } else { 
          values.panel_x = 0;
          values.panel_y = 0;
          return false; // We are done, or if zooming we are back to first panel.
        }
      }
    }
    jQCentral.message("Beginning panel: vert = " + values.panel_y + " horiz = " + values.panel_x,
                      false);
    return true;
  }
  
  private boolean advanceTheZoom (FractalSpec current, Messenger jQCentral)
  {
    if (pause_p) {
      jQCentral.message("Frame " + frameNumber_ + " of " + frames + " complete.", false);
    } else {
      jQCentral.message("Drawing frame " + (frameNumber_  + 1) + " of " + frames, false);
    }
    boolean gotErrorP = false;
    if (startSpec_ == null) {
      gotErrorP = true;
      jQCentral.message("Sequence start frame has been lost.", true);
    }
    if (endSpec_ == null) {
      gotErrorP = true;
      jQCentral.message("Sequence end frame has been lost.", true);
    }
    if (gotErrorP) {
      enabled_p = false;
      return false;
    }
    long this_panel = frameNumber_;
    double ratio = (double)this_panel / (double)frames;

    // We will do linear movement through the coeff and translation
    if (startSpec_.rangeOver != FractalSpec.CENTER) {
      current.center.assign_from(startSpec_.center.plus(endSpec_.center.minus(startSpec_.center).times_scalar(ratio)));
    }
    if (startSpec_.rangeOver != FractalSpec.COEFF) {
      current.coeff.assign_from(startSpec_.coeff.plus(endSpec_.coeff.minus(startSpec_.coeff).times_scalar(ratio)));
    }
    if (startSpec_.rangeOver != FractalSpec.TRANS) {
      current.trans.assign_from(startSpec_.trans.plus(endSpec_.trans.minus(startSpec_.trans).times_scalar(ratio)));
    }
    current.depth  = (int)(startSpec_.depth + ratio * (endSpec_.depth - startSpec_.depth));
    
    current.color_cycles = (int)Math.min(endSpec_.color_cycles, current.depth / NON_BLACKS);
    
    // We rotate through the axes, which are already normalized
    // We apply ratio to the difference in angles, add to startSpec_ angle, 
    // then convert back from angles to axis projections.
    double bsAngle;
    double newAngle;
    bsAngle = Math.acos(startSpec_.vx.r);
    newAngle = bsAngle + (ratio * (Math.acos(endSpec_.vx.r) - bsAngle));
//    System.out.println("ratio=" + ratio + " newAngle=" + newAngle + " startSpec_Angle=" +
//        bsAngle);
    current.vx.r = Math.cos(newAngle);
//    System.out.println("Math.cos(newAngle) =" + Math.cos(newAngle));
    if (current.vx.r == -0) current.vx.r = 0; //Believe it or not cos(pi/2) = -0
//    System.out.println("current.vx.r =" + current.vx.r);
    bsAngle = Math.acos(startSpec_.vx.i);
    current.vx.i = Math.cos(bsAngle + (ratio * (Math.acos(endSpec_.vx.i) - bsAngle)));
    if (current.vx.i == -0) current.vx.i = 0;
    bsAngle = Math.acos(startSpec_.vx.j);
    current.vx.j = Math.cos(bsAngle + (ratio * (Math.acos(endSpec_.vx.j) - bsAngle)));
    if (current.vx.j == -0) current.vx.j = 0;
    bsAngle = Math.acos(startSpec_.vx.k);
    current.vx.k = Math.cos(bsAngle + (ratio * (Math.acos(endSpec_.vx.k) - bsAngle)));
    if (current.vx.k == -0) current.vx.k = 0;
    
    bsAngle = Math.acos(startSpec_.vy.r);
    current.vy.r = Math.cos(bsAngle + (ratio * (Math.acos(endSpec_.vy.r) - bsAngle)));
    if (current.vy.r == -0) current.vy.r = 0;
    bsAngle = Math.acos(startSpec_.vy.i);
    current.vy.i = Math.cos(bsAngle + (ratio * (Math.acos(endSpec_.vy.i) - bsAngle)));
    if (current.vy.i == -0) current.vy.i = 0;
    bsAngle = Math.acos(startSpec_.vy.j);
    current.vy.j = Math.cos(bsAngle + (ratio * (Math.acos(endSpec_.vy.j) - bsAngle)));
    if (current.vy.j == -0) current.vy.j = 0;
    bsAngle = Math.acos(startSpec_.vy.k);
    current.vy.k = Math.cos(bsAngle + (ratio * (Math.acos(endSpec_.vy.k) - bsAngle)));
    if (current.vy.k == -0) current.vy.k = 0;
   
    // These things may have to move in a logarithmic way if we're zooming
    // If we are not zooming the result is the same as for the simple case above.
    // SEE::NOTE1 AT BOTTOM OF FILE
    double log_trbr = Math.log(endSpec_.range / startSpec_.range);
    double range_ratio = Math.pow(Math.E, log_trbr * this_panel / frames);
    double total_range_movement = 0;
    double sofar_range_movement = 0;
    for (int i = 1; i <= frames; i++) {
      total_range_movement += Math.pow(Math.E, log_trbr * i / frames);
      if (i == this_panel) {
        sofar_range_movement = total_range_movement;
      }
    }
    double compensator = 1.0 / total_range_movement;
    double logRat = sofar_range_movement * compensator;
// Seems a bit too slow.
//    double LRcos = Math.cos(logRat * Math.PI / 2);
//    double LRest = 1 - LRcos;
    double LRsin = Math.sin(logRat * Math.PI / 2);
    double LRest = 1 - LRsin;
    
    BigQuaternion startSpec_Q;
    BigQuaternion currentQ;
    BigQuaternion endSpec_Q;
    if (startSpec_.rangeOver == FractalSpec.COEFF) {
      currentQ = current.coeff;
      startSpec_Q = startSpec_.coeff;
      endSpec_Q = endSpec_.coeff;
    } else if (startSpec_.rangeOver == FractalSpec.TRANS) {
      currentQ = current.trans;
      startSpec_Q = startSpec_.trans;
      endSpec_Q = endSpec_.trans;
    } else {
      currentQ = current.center;
      startSpec_Q = startSpec_.center;
      endSpec_Q = endSpec_.center;
    }
    if (startSpec_.range > endSpec_.range) {
      currentQ.assign_from(startSpec_Q.times_scalar(LRest).plus(endSpec_Q.times_scalar(LRsin)));
//      currentQ.r = startSpec_Q.r * LRest + endSpec_Q.r * LRsin;
    } else {
      currentQ.assign_from(startSpec_Q.times_scalar(LRsin).plus(endSpec_Q.times_scalar(LRest)));
    }
    current.range = startSpec_.range * range_ratio;
//    current.depth = startSpec_.depth +
//                    (int)Math.floor(logRat * (endSpec_.depth - startSpec_.depth));
      
    return true;
  }
  
  private boolean advanceThePivot(FractalSpec values, Messenger jQCentral)
    // Returns true if we advance, which we always do these days.
    // Horizontal pivot means vertical axis changes.
  {
    FastQuaternion base = (!horiz_p_ ? startSpec_.vx : startSpec_.vy);
    FastQuaternion target = startSpec_.vz;
    if (pause_p) {
      jQCentral.message("Frame " + frameNumber_ + " of " + frames + " complete.", false);
    } else {
      jQCentral.message("Drawing frame " + (frameNumber_  + 1) + " of " + frames, false);
    }
    double degreesAdvance = (frameNumber_ * degrees)/ frames;
    if (!horiz_p_) { // Horizontal pivot means vertical axis changes.
      values.vx.assign_from(FastQuaternion.rotate(degreesAdvance, base, target));
    } else {
      values.vy.assign_from(FastQuaternion.rotate(degreesAdvance, base, target));
    }
    return true;
  }
  
  private boolean advanceTheShift(FractalSpec values, Messenger jQCentral)
    // Returns true if we advance, which we always do these days.
  {
    FastQuaternion base = (horiz_p_ ? startSpec_.vx : startSpec_.vy);
    BigQuaternion addIn = 
      startSpec_.vz.times_scalar(frameNumber_ * shift).bigQuaternionValue();
    if (pause_p) {
      jQCentral.message("Frame " + frameNumber_ + " of " + frames + " complete.", false);
    } else {
      jQCentral.message("Drawing frame " + (frameNumber_  + 1) + " of " + frames, false);
    }
    if (startSpec_.rangeOver == FractalSpec.COEFF) {
      values.coeff.assign_from(startSpec_.coeff.plus(addIn));
    } else if (startSpec_.rangeOver == FractalSpec.TRANS) {
      values.trans.assign_from(startSpec_.trans.plus(addIn));
    } else {
      values.center.assign_from(startSpec_.center.plus(addIn));
    }
   return true;
  }
  
  
  
}
