package JavaQuat20;

import java.util.Vector;
import java.awt.*;

public class SortableVector extends Vector
{
  boolean debugP = false;
  int [] map_ = null;
  
  private void printRed (int start, int size)
  {
    if (debugP) {
      System.out.print(" From start=" + start + " for length=" + size);
      if (size > 10) {
        for (int i = start; i < start + size; i++) {
          System.out.print(" i=" + i + " r=" + ((Color)elementAt(i)).getRed());
        }
      } else {
        for (int i = start; i < start + size; i++) {
          System.out.print(" " + ((Color)elementAt(i)).getRed());
        }
      }
      System.out.println(".");
    }
  }
  
  
  
  public SortableVector ()
  {
    super();
  }
  
  public SortableVector (int initialCapacity)
  {
    super(initialCapacity);
  }
  
  public void qsort (Cmp comp)
  {
    qsortInternal(0, size() -1, comp);
  }
  
  public int [] mappedQsort (Cmp comp)
  {
    map_ = new int [size()];
    for (int i = 0; i < size(); i++) {
      map_[i] = i;
    }
    
    try {
      qsortInternal(0, size() -1, comp);
    } catch (Throwable e) {
      map_ = null;
      System.err.println("SortableVector.mappedQsort internal error.");
    }
    if (map_ != null) {
      // Set answer to the inverse of map_.
      int [] answer = new int [size()];
      for (int i = 0; i < size(); i++) {
        answer[map_[i]] = i;
      }
      map_ = null;
      return answer;
    }
    return null;
  }
  
  private void swap (int aa, int bb)
  { 
    if (debugP) {
      System.out.println("Swap " + ((Color)elementAt(aa)).getRed() + " at " + aa +
                         " and " + ((Color)elementAt(bb)).getRed() + " at " + bb);
    }
    if (map_ != null) {
      int tmp = map_[aa];
      map_[aa] = map_[bb];
      map_[bb] = tmp;
    }
    Object temp = elementAt(aa);
    setElementAt(elementAt(bb),aa);
    setElementAt(temp,bb);
  }
  
  public void qsortInternal (int left, int size, Cmp comp)
    // Quicksort is an unstable sort.  Left is the first element in the sort.
    // Size indicates the number of elements in the sort.  Comp satisfies the
    // Cmp interface in that it defines 'public int cmp (Object a, Object b)' which
    // returns 1 if the b should come before a and 0 or less otherwise.
  {
    if (size < 4) {
      if (debugP) {
        System.out.print("Entry < 4 - "); printRed(left, size);
      }
      if (size == 3) {
        int mid = left + 1;
        if (comp.cmp(elementAt(left), elementAt(mid)) > 0) {
          swap(left,mid);
        }
        if (comp.cmp(elementAt(mid), elementAt(mid + 1)) > 0) {
          swap(mid, mid+1);
        }
        if (comp.cmp(elementAt(left), elementAt(mid)) > 0) {
          swap(left,mid);
        }
      } else if (size == 2) {
        if (comp.cmp(elementAt(left), elementAt(left + 1)) > 0) {
          swap(left, left+1);
        }
      } 
      // size == 1.
      if (debugP) {
        System.out.print("Exit < 4 - "); printRed(left, size);
      }
    } else {
      if (debugP) {
        System.out.print("Entry - "); printRed(left, size);
      }
      Object pivot;
      int right = left + (size - 1);
      int mid = left + 1 * (size/2);
      // will point to the place with the middle value.
      if (comp.cmp(elementAt(left), elementAt(mid)) <= 0) {
        if (comp.cmp(elementAt(mid), elementAt(right)) <= 0)
          pivot = elementAt(mid);
        else if (comp.cmp(elementAt(right), elementAt(left)) <= 0) {
          if (comp.cmp(elementAt(mid), elementAt(left)) <= 0)
            pivot = elementAt(right);
          else
            pivot = elementAt(left);
        } else
          pivot = elementAt(right);
      } else {
        if (comp.cmp(elementAt(right), elementAt(mid)) <= 0)
          pivot = elementAt(mid);
        else if (comp.cmp(elementAt(right), elementAt(left)) <= 0)
          pivot = elementAt(right);
        else
          pivot = elementAt(left);
      }
      if (debugP) {
        System.out.println("Left=" + left + " mid=" + mid + " right=" + right + " pivot red=" +
                           ((Color)pivot).getRed());
      }
      // swap around pivot
      int i = left;
      int j = right;
      while (true) {
        while ((i != j) && (comp.cmp(elementAt(i), pivot) <= 0))
          i += 1;
        while ((i != j) && (comp.cmp(elementAt(j), pivot) > 0))
          j -= 1;
        if (i == j) break;
        swap(i, j);
      }
      if (debugP) {
         System.out.print(" i==j=" + i + " "); printRed(left, size);
      }
      int right_size = right - i;
      if (right_size > 0) { // This to get around data like (10 1 2 3 4 5 6 7 8 9)
        if (debugP) {
           System.out.println(" right_size > 0 ");
        }
        if (right_size > 4) { //For bad case 2 3 4 5 6....n 1
          if (debugP) {
             System.out.println(" right_size > 4 ");
          }
          int mid_right = i + (right_size/ 2);
          if (comp.cmp(elementAt(mid_right), elementAt(right)) > 0) { //For bad case 2 3 4 5 6....n 1
            swap(mid_right, right);
          }
        }
        qsortInternal(i, right_size + 1, comp);
        j = i - 1;
        // Big speed win if lots of elements are equal.  Garr
        while (j > left && comp.cmp(elementAt(i), elementAt(j)) <= 0) 
          j -= 1;
        if (debugP) {
           System.out.println(" left half - j=" + j);
        }
        qsortInternal(left, j - left + 1, comp);
        if (debugP) {
          System.out.print("Exit 2 - "); printRed(left, size);
        }
      } else {
        j = i - 1;
        // Big speed win if lots of elements are equal.  Garr
        if (debugP) {
           System.out.println(" right_size == 0 ");
        }
        while (j > left && comp.cmp(elementAt(i), elementAt(j)) <= 0) 
          j -= 1;
        int left_size = j - left;
        if (debugP) {
           System.out.println(" j=" + j + " left_size=" + left_size);
        }
        if (left_size > 4) {
          swap(left, left + (left_size/2) - 1);
        }
        qsortInternal(left, left_size + 1, comp);
        if (debugP) {
          System.out.print("Exit 3 - "); printRed(left, size);
        }
      }
    }
  }
}
