package JavaQuat20;

import java.awt.*;
import java.awt.image.*;
import java.util.*;

public class ThumbNailProducer
{
  final static public int size = 20;

  private Vector consumers_ = new Vector();
  private boolean hasPixels_p_ = false;
  private boolean hasColorModel_p_ = false;
  private boolean ready_p_ = false;
  private byte [] pixels_;
  private ColorModel colorModel_;
  private Hashtable properties_ = new Hashtable();

  
  public ThumbNailProducer ()
  {
     pixels_ = new byte[size * size]; // We always want these around for GIF writing.
     colorModel_ = null;
  }
  
  public void setPixel (int x, int y, byte color)
  {
    pixels_[x + y * size] = color;
    hasPixels_p_ = true;
    if (hasColorModel_p_)
      ready_p_ = true;
  }
  
  public void setColorModel (ColorModel arg)
  {
    colorModel_ = arg;
    hasColorModel_p_ = true;
    if (hasPixels_p_)
      ready_p_ = true;    
  }

  
  protected void produce(ImageConsumer ic) 
  {
    //System.out.println("produce entry. colorModel is of type " + parent.colorModel.getClass());
    if (!ready_p_) {
      System.err.println("ThumbNailProducer.produce but not ready.");
      ic.imageComplete(ImageConsumer.IMAGEERROR);
      return;
    } 
    ic.setDimensions(size, size);
    ic.setProperties(properties_);
    ic.setColorModel(colorModel_);
    ic.setHints(ImageConsumer.SINGLEPASS |
                ImageConsumer.SINGLEFRAME);
    ic.setPixels(0, 0, size, size, colorModel_, pixels_, 0, size);
    ic.imageComplete(ImageConsumer.STATICIMAGEDONE);
  }
    
  /*  ******************* ImageProducer implementation ************************ */

  public synchronized void startProduction(ImageConsumer ic)
    // This is one of the two interface methods for ImageProducer, other than add,is,remove consumer.
    //  It guarantees that drawTheBox sets up the picture before procude is called
    //  to make settings in the ImageConsumer(s).
  {
    //System.out.println("FractalImageProducer.startProduction entry");
    if (ic != null) {
      addConsumer(ic);
      if (!ready_p_) {
        System.err.println("ThumbNailProducer.startProduction but not ready.");
        return;
      } 
      if (consumers_.size() == 0) {
        return; //Following loop does not work properly.
      }
      for (int i = consumers_.size(); i>0; i--) {
        if (i - 1 >= consumers_.size()) // synchronized does not work.
          continue;
        ic = (ImageConsumer)consumers_.elementAt(i - 1);
        produce(ic);
        consumers_.removeElement(ic);
      }
    }
  }

  public void requestTopDownLeftRightResend(ImageConsumer ic) 
  {
    if (ready_p_) {
      //System.out.println("FractalImageProducer::requestTopDownLeftRightResend IF entry");
      ic.setDimensions(size, size);
      ic.setProperties(properties_);
      ic.setColorModel(colorModel_);
      ic.setHints(ImageConsumer.SINGLEPASS |
                  ImageConsumer.TOPDOWNLEFTRIGHT |
                  ImageConsumer.SINGLEFRAME);
      ic.setPixels(0, 0, size, size, colorModel_, pixels_, 0, size);
    }
  }

  public synchronized void addConsumer (ImageConsumer ic) 
  {
    if (hasPixels_p_) {
      produce(ic);
    } else if (!consumers_.contains(ic)) {
      consumers_.addElement(ic);
    }
  }

  public synchronized boolean isConsumer (ImageConsumer ic) 
  {
    return consumers_.contains(ic);
  }

  public synchronized void removeConsumer (ImageConsumer ic) {
    consumers_.removeElement(ic);
  }
  
}
