package JavaQuat20;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;
import JavaQuat20.GIFDecoderPkg.*;
import JavaQuat20.GIFEncoderPkg.*;
import java.io.*;
import java.net.URL;
import java.net.*;

public class TrapFrame extends Frame implements TransferCMap
{
  static final int bitsPerByte = 8;
  static final int indexedColors = 256;
  
  static int showWidth = 600;
  static int showHeight = 400;

  private JQCentral jQCentral_ = null;
  private TrapCanvas canvas_ = null;
  private Image refresher_ = null;
  private TrapFrame self_;
  private int width_ = 0;
  private int height_ = 0;
  private double radius_ = 0;
  private byte [] pixels_; // A picture as an Array of indices into colorMap.
  private int colorMapSize_ = 0;
  private byte [] reds_; // The red part of the colorMap.
  private byte [] greens_; // The green part of the colorMap.
  private byte [] blues_; // The blue part of the colorMap.
  private IndexColorModel colorModel256_ = null;
  private byte transparentPixel_ = 0;
  private boolean transparentPixelP_ = true;
  private boolean ready_ = false;
  private boolean colorsIs256p_ = true;
  private String pathname_ = null;
  private MenuItem trans0MI_ = null;

  private byte colorMapHit_ = 0;
  private int realCenterPix_ = 0;
  private int imagCenterPix_ = 0;
  private double realCenter_ = 0;
  private double imagCenter_ = 0;
  private double rangePerPixel_ = 1;
  private double rotation_ = 0;
  private double real2real_ = 1;
  private double imag2real_ = 0;
  private double real2imag_ = 0;
  private double imag2imag_ = 1;
  
  private int debugCount_ = 0;

  
  public TrapFrame (JQCentral parent)
  {
    self_ = this;
    jQCentral_ = parent;
    canvas_ = new TrapCanvas();
    setLayout(new BorderLayout(0, 0));
    add("Center", canvas_);
    addWindowListener(new MyWindowLNR());
    setResizable(true);
    reds_ = new byte[256];
    greens_ = new byte[256];
    blues_ = new byte[256];


    // -------------- Menu -----------------------------------
    MenuBar menuBar = new MenuBar();
    setMenuBar(menuBar);
    
    Menu fileMenu = new Menu("File", true); // true for tearoff.
    menuBar.add(fileMenu);
    
    MenuItem generalMI = new MenuItem("Open ...");
    generalMI.addActionListener(new trapFileLNR());
    generalMI.setShortcut(new MenuShortcut(KeyEvent.VK_T));
    fileMenu.add(generalMI);
        
    generalMI = new MenuItem("Save As...");
    generalMI.addActionListener(new saveLNR());
    generalMI.setShortcut(new MenuShortcut(KeyEvent.VK_S));
    fileMenu.add(generalMI);
    
    fileMenu.addSeparator();
    
    generalMI = new MenuItem("Edit Color Map");
    generalMI.addActionListener(new cMapLNR());
    generalMI.setShortcut(new MenuShortcut(KeyEvent.VK_E));
    fileMenu.add(generalMI);
    
    trans0MI_ = new MenuItem("Move Transp. to 0");
    trans0MI_.addActionListener(new trans0LNR());
    trans0MI_.setShortcut(new MenuShortcut(KeyEvent.VK_0));
    fileMenu.add(trans0MI_);
    
    fileMenu.addSeparator();
    
    generalMI = new MenuItem("Close");
    generalMI.addActionListener(new closeLNR());
    generalMI.setShortcut(new MenuShortcut(KeyEvent.VK_W));
    fileMenu.add(generalMI);
    
  }
  
  /* ********************************** Utility ************************** */
  
  public boolean readyP ()
  {
    return ready_;
  }
  
  public int width ()
  {
    return width_;
  }
  
  public int height ()
  {
    return height_;
  }
  
  public double radius ()
  {
    if (radius_ == 0) {
      radius_ = Math.sqrt(width_ * width_ + height_ * height_) / 2;
    }
    return radius_;
  }
  
  public String pathname ()
  {
    return pathname_;
  }
  
  public void setColorAtIndex (int index, Color c, boolean repaint_p)
    // This is called from JQCentral.
  {
    if (index < indexedColors) {
      reds_[index] = (byte)c.getRed();
      greens_[index] = (byte)c.getGreen();
      blues_[index] = (byte)c.getBlue();
      if (repaint_p) {
        updateColorModel();
        repaint();
      }
    }
  }
  
  public Color getColorAt (int i)
  {
    int r = reds_[i];
    if (r < 0) r += 256;
    int g = greens_[i];
    if (g < 0) g += 256;
    int b = blues_[i];
    if (b < 0) b += 256;
    return new Color(r, g, b);
  }

  public void transferColorMapTo (ColorMapOwner cmo)
   // tell Color Map Editor the color at each index.
   // This gets called from JQCentral.updateColorMap
  {
    for (int i = 0; i < reds_.length; i++) {
      cmo.setIndexAndColor(i, getColorAt(i), i == (reds_.length - 1));
    }
  }
  

  private void updateColorModel ()
  {
    colorModel256_ = new IndexColorModel(bitsPerByte, indexedColors, reds_, greens_, blues_);
  }
  
  public void paint (Graphics g)
  {
    if (pixels_ != null && colorModel256_ != null) {
      canvas_.showImage(pixels_, width_, height_, colorModel256_);
    }
  }
  
  private void swapPixelIndices(byte index1, byte index2)
  {
    for (int j = 0; j < pixels_.length; j++) {
      if (pixels_[j] == index2) {
        pixels_[j] = index1;
      } else if (pixels_[j] == index1) {
        pixels_[j] = index2;
      }
    }
  }

  private void swapRGBs (int i, int j)
    // Do not forget to call updateColorModel after calling this.
  {
    byte temp = reds_[i];
    reds_[i] = reds_[j];
    reds_[j] = temp;
    temp = greens_[i];
    greens_[i] = greens_[j];
    greens_[j] = temp;
    temp = blues_[i];
    blues_[i] = blues_[j];
    blues_[j] = temp;
  }
    
  //---------------------------------- top level --------------------------------------------
  public void openGIFImage (String pathname, boolean is_url_p, boolean copyColorMapToFractalP)
    // This uses my GIFDecoder to read in the image, thus saving memory compared to MRJ2.1.1
    // It is called by JQCentral in openTrapImage and loadInitialImage methods.
  {
    GIFDecoder gifDecoder = null;
    InputStream stream1 = null;
    FractalSpec args = null;
    JQIO inUtility = new JQIO(jQCentral_, this);
    
    if (jQCentral_.debugP()) {
      int two55 = 255;
      System.out.println("TrapFrame.openGIFImage (byte)255=" + (byte)two55);
    }
    
    if (pathname == null || pathname.length() == 0) {
      // Get the pathname if none was passed in.
      if (is_url_p) {
        pathname = inUtility.urlStringFromDialog(pathname_);
      } else {
        pathname = inUtility.pathFromDialog(true); // true => use read dialog
      }
    }

    if (pathname == null || pathname.length() == 0) {
      jQCentral_.message("No action taken.", false);
      return;
    }
    pathname_ = pathname;

    // Init 
    ready_ = false;
    width_ = -1;
    height_ = -1;
    boolean READIMAGE = false; // False means do not skip image.
    try {
      stream1 = inUtility.toInputStream(pathname, is_url_p, false); // false => create stream.
      if (stream1 == null)
        throw new IOException("NOT FOUND: " + pathname);

      gifDecoder = new GIFDecoder(stream1, READIMAGE);
      height_ = gifDecoder.getHeight();
      width_ = gifDecoder.getWidth();

      if (width_ >= 0 && height_ >= 0) {
        double wRatio = 600/(double)width_;
        double hRatio = 400/(double)height_;
        double ratio = Math.min(wRatio, hRatio);
        showWidth = Math.min(width_, (int)(ratio * width_));
        showHeight = Math.min(height_, (int)(ratio * height_));
        if (ratio < 1) {
          jQCentral_.message("Trap displayed at reduced size: " + (float)ratio, false);
        }
        setSize(showWidth, showHeight);
      } else {
        throw new IOException("ZERO SIZE: " + pathname);
      }
      pixels_ = gifDecoder.pixelsAsIndices();
      loadGIFColorMap(gifDecoder, copyColorMapToFractalP);
      canvas_.showImage(pixels_, width_, height_, colorModel256_);
      //validate();
      show();
      ready_ = true; // ready for use in a gif trap looper.
    } catch (Throwable e) {
      jQCentral_.message("TrapFrame.openGIFImage" +
                         (e.getMessage() == null ? e.toString() : e.getMessage()), true);
//      System.out.println("TrapFrame.openGIFImage: catch " + 
//                         (e.getMessage() == null ? e.toString() : e.getMessage()));
//      e.printStackTrace();
    } finally {
      if (stream1 != null)
        try {
          stream1.close(); // In case we didn't close it above.
        } catch (Throwable e) {}
    }  
  }
  
  public void remapPixelIndices (int [] pixelMap)
    //This is called from the colorMapEditor when colors have been sorted.
  {
    for (int i = 0; i < pixels_.length; i++) {
      int oldIndex = pixels_[i];
      if (oldIndex < 0) {
        oldIndex += 256;
      }
      int newIndex = pixelMap[oldIndex];
      pixels_[i] = (byte)((newIndex > 127) ? (newIndex - 256) : newIndex);
    }
    if (transparentPixelP_) {
      int transparentIntPixel = 
        transparentPixel_ < 0 ? transparentPixel_ + 256 : transparentPixel_;
      transparentPixel_ = (byte)pixelMap[transparentIntPixel];
    }
  }
  
  private void loadGIFColorMap (GIFDecoder gifDecoder, boolean copyColorMapToCanvasP)
    // This is similar to the ColorMapEditor's loadGIFColorMap.  Duplicated here since now (9-99)
    //  MRJ retursn a DirectColorModel from the PixelGrabber's getColorModel method, which is wrong.
    // This sets reds_, greens_, blues_, the transparency info, and colorMapSize_ .

    // Returns true if successful, false otherwise.
  {
    int transparentIntPixel = -1;
    RGBxSizeSequence colorTabel1 = gifDecoder.globalRGBTable();
    if ((transparentPixelP_ = gifDecoder.transparentPixelP())) {
      transparentIntPixel = gifDecoder.transparentPixelIndex();
      transparentPixel_ = (byte)transparentIntPixel;
      trans0MI_.setEnabled(true);
    } else {
      trans0MI_.setEnabled(false);
    }
    if (colorTabel1 != null) {
      colorMapSize_ = colorTabel1.byteTable.length / 3;
      int i = 0;
      for (i = 0; i < colorMapSize_; i++) {
        int ix3 = i * 3;
        reds_[i] = colorTabel1.byteTable[ix3];
        greens_[i] = colorTabel1.byteTable[ix3 + 1];
        blues_[i] = colorTabel1.byteTable[ix3 + 2];
      }
      jQCentral_.message("Trap GIF contains " + colorMapSize_ + " colors.  " +
                        (transparentPixelP_ ? ("Transparent index = " + transparentIntPixel) :
                          "There is no transparent pixel."), false);
      for (i = colorMapSize_; i < 256; i++) {
        reds_[i] = 0;
        greens_[i] = 0;
        blues_[i] = 0;
      }
        
      // Move the transparent pixel to the 0th position if necessary.
      // Change the image where it was 0 or the transparent index to
      // point to the same color as before.
      // Although this is not absolutely necessary, it should make life more consistant
      //  for the user.
      if (transparentPixelP_ && transparentIntPixel > 0) {
        jQCentral_.message("Transparent index is being moved to 0.", false);
        swapRGBs(0, transparentIntPixel);
        swapPixelIndices(transparentPixel_, (byte)0);
        transparentPixel_ = 0;
      }
        
      // updateColorModel:
      colorModel256_ = new IndexColorModel(bitsPerByte, indexedColors, reds_, greens_, blues_);
   
      if (copyColorMapToCanvasP) {
        for (i = 0; i < 256; i++) {
          jQCentral_.setIndexAndColor(i, getColorAt(i), (i == 255));
        }
        jQCentral_.updateColorMapEd(false, this);
      }
    }
  }
  
  private boolean writeGIF ()
    // Return true if the write was successful.
  {
    boolean interlace_p = false;
    double version = 2.0;
    JQIO outUtility = new JQIO(jQCentral_, this);
    OutputStream stream1 = null;
    String comment = "";
    
    String pathname = outUtility.pathFromDialog(false); // false => use write dialog

    if (pathname == null || pathname.length() == 0) {
      jQCentral_.message("Write Trap GIF image - No action taken.", false);
      return false;
    }

    // Save TRAP info in jQCentral_.
    jQCentral_.setTrapPathname(pathname);

    try {
      stream1 = outUtility.toOutputStream(pathname, false, false);
      if (stream1 == null)
        throw new IOException("NOT FOUND: " + pathname);
        
      GIFEncoder gifWriter = 
        new GIFEncoder(width_, height_, pixels_, colorModel256_);
      if (transparentPixelP_) {
        gifWriter.setTransparentPixel(transparentPixel_);
      }
      gifWriter.JavaQuatWrite(stream1, comment, false, version);
      jQCentral_.message("Trap Image Saved.", false);

    } catch (Throwable e) {
      jQCentral_.message("Error on stream or encode: " + e,true);
      jQCentral_.message(e.getMessage() == null ? e.toString() : e.getMessage(), true);
      return false;
    } finally {
      try {
        if (stream1 != null)
          stream1.flush();
          stream1.close();
      } catch (Throwable e) {}
    }
    return true;
  }


  /* ******************************** DISPLAY **************************** */
  
  
  class TrapCanvas extends Canvas implements ImageObserver, MouseListener
  {
    private int width_;
    private int height_;
    private byte [] pixels_;
    
    TrapCanvas ()
    {
      addMouseListener(this);
    }
    
    public void showImage (byte [] pixels, int width, int height, ColorModel colorModel256)
      // Similar to drawing fractals but we have allthe pixels_ already.
    {
      width_ = width;
      height_ = height;
      pixels_ = pixels;
      TpShowProd producer1 = new TpShowProd(pixels, width, height, colorModel256);
      refresher_ = createImage(producer1);
      if (canvas_.prepareImage(refresher_, this)) {
        // Either the test succeeds or the system calls imageUpdate.
        repaint();
      }
    }
  
    public boolean imageUpdate (Image  img, int  infoflags,
                                int  x, int  y, int  width, int  height)
    {
//    System.out.println("TrapFrame$TrapCanvas imageUpdate entry.");
      super.imageUpdate(img, infoflags, x, y, width, height);
    
      if ((infoflags & ALLBITS) > 0) {
        repaint();
        return false; // No more information needed.
      } 
      return true; // Still want to be notified when image is complete.
    }
    
    public void paint (Graphics g)
    {
      g.drawImage(refresher_, 0, 0, this);
    }
    
    /* ******************************* Mouse Listener methods ************************ */
    
    public void mouseClicked(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
  
    public void mousePressed(MouseEvent evt)
    {
    }
  
    public void mouseReleased(MouseEvent evt)
    {
      double widthExpansion = width_/(double)showWidth;
      double heightExpansion = height_/(double)showHeight;
      int col = (int)(evt.getX() * widthExpansion);
      int row = (int)(evt.getY() * heightExpansion);
      int pixIndex = row * width_ + col;
      if (pixIndex < 0 || pixIndex >= pixels_.length) {
        jQCentral_.message("Illegal pixel index into trap image: " + pixIndex, true);
        jQCentral_.message("  Legal values are [0, " + (pixels_.length - 1) + "].", false);
        return;
      }
      int color_map_index = pixels_[pixIndex];
      if (color_map_index < 0) {
        color_map_index += 256;
      }
      Color c = getColorAt(color_map_index);
      jQCentral_.setFocusOfColorMapEditor(color_map_index, c, true);
    }  
    
  } // End of Canvas
  
  class TpShowProd implements ImageProducer
  {
    Vector consumers = new Vector();
    Hashtable properties = new Hashtable();
    int width_;
    int height_;
    byte [] localPixelSet;
    byte [] pixels_;
    ColorModel colorModel_;
  
    TpShowProd (byte [] allPixels, int allWidth, int allHeight, ColorModel colorModel)
    {
      width_ = showWidth;
      height_ = showHeight;
      double widthRatio = allWidth/(double)width_;
      double heightRatio = allHeight/(double)height_;
      localPixelSet = new byte[width_ * height_];
      pixels_ = allPixels;
      colorModel_ = colorModel;
      int i = 0;
      for (int h = 0; h < height_; h++) {
        int allRowOffset = ((int)(h * heightRatio)) * allWidth;
        for (int w = 0; w < width_; w++) {
          int allHoriz = (int)(w * widthRatio);
          localPixelSet[i++] = allPixels[allRowOffset + allHoriz];
        }
      }
    }
    /*  ******************* ImageProducer implementation help ******************* */

    protected void produce(ImageConsumer ic) 
    {
      //System.out.println("produce entry. colorModel is of type " + parent.colorModel.getClass());
      ic.setDimensions(width_, height_);
      ic.setProperties(properties);
      ic.setColorModel(colorModel_);
      ic.setHints(ImageConsumer.SINGLEPASS |
                  ImageConsumer.SINGLEFRAME);
//      if (true)
        ic.setPixels(0, 0, width_, height_, colorModel_, localPixelSet, 0, width_);
//      else
//        ic.setPixels(0, 0, width, height, parent.colorModel, intPixels_, 0, width);
      ic.imageComplete(ImageConsumer.STATICIMAGEDONE);
    }
    
  /*  ******************* ImageProducer implementation ************************ */

    public synchronized void startProduction(ImageConsumer ic)
      // This is one of the two interface methods for ImageProducer, other than add,is,remove consumer.
      //  It guarantees that drawTheBox sets up the picture before produce is called
      //  to make settings in the ImageConsumer(s).
    {
      //System.out.println("TrapFrame.startProduction entry");
      if (ic != null) {
        addConsumer(ic);
        int temp;
        while ((temp = consumers.size()) > 0) {
          ic = (ImageConsumer)consumers.elementAt(temp - 1);
          produce(ic);
          consumers.removeElement(ic);
        }
      }
    }
      

    public void requestTopDownLeftRightResend(ImageConsumer ic) 
    {
//      System.out.println("TrapFrame::requestTopDownLeftRightResend entry ic=" + ic);
      if (colorsIs256p_ ? (pixels_ != null) : false) {
        //System.out.println("Q1ImageProducer::requestTopDownLeftRightResend IF entry");
        ic.setDimensions(width_, height_);
        ic.setProperties(properties);
        ic.setColorModel(colorModel_);
        ic.setHints(ImageConsumer.SINGLEPASS |
                    ImageConsumer.TOPDOWNLEFTRIGHT |
                    ImageConsumer.SINGLEFRAME);
        if (colorsIs256p_)
          ic.setPixels(0, 0, width_, height_, colorModel_, pixels_, 0, width_);
        //else
        //  ic.setPixels(0, 0, width, height, parent.colorModel, intPixels_, 0, width);
      }
    }

    public synchronized void addConsumer (ImageConsumer ic) 
    {
      produce(ic);
    }

    public synchronized boolean isConsumer (ImageConsumer ic) 
    {
      return consumers.contains(ic);
    }

    public synchronized void removeConsumer (ImageConsumer ic) 
    {
      consumers.removeElement(ic);
    }
  }
  
  /* ---------------------- FRACTAL HITS --------------------- */
  
  
  public void setCenter (double real, double imaginary)
  {
    realCenter_ = real;
    imagCenter_ = imaginary;
    realCenterPix_ = (int)(width_ / 2.0);
    imagCenterPix_ = (int)(height_ / 2.0);
  }
  
  public boolean setRange (double range)
  {
    if (range != 0 && ready_ && width_ > 0) {
      rangePerPixel_ = range / width_;
      return true;
    }
    if (range == 0) {
      jQCentral_.message("Error setting range for trap GIF.  Given range is 0.", true);
    }
    if (!ready_) {
      jQCentral_.message("Error setting range for trap GIF.  Trap GIF not ready.", true);
    }
    if (width_ <= 0) {
      jQCentral_.message("Error setting range for trap GIF.  Trap GIF width is " + width_, true);
    }
    return false;
  }
  
  public void setRotation (double degrees)
    // This is degrees gif is to be rotated counter clockwise.
  {
    rotation_ = degrees;
    if (degrees != 0) {
      double angle = Math.PI * degrees / 180;
      real2real_ = Math.cos(angle);
      imag2real_ = -Math.sin(angle);
      real2imag_ = Math.sin(angle);
      imag2imag_ = Math.cos(angle);
    }
  }
  
  public boolean hitP (double real, double imaginary)
  {
    int deltaRealPix;
    int deltaImagPix;
    if (rotation_ == 0) {
      deltaRealPix = (int)((real - realCenter_) / rangePerPixel_);
      deltaImagPix = (int)((imagCenter_ - imaginary) / rangePerPixel_);  // Bigger Y is up.
    } else {
      double deltaReal = real - realCenter_;
      double deltaImag = imagCenter_ - imaginary; // Y is upside down on the scene, bigger is up.
      deltaRealPix = (int)((deltaReal * real2real_ + deltaImag * imag2real_) / rangePerPixel_);
      deltaImagPix = (int)((deltaReal * real2imag_ + deltaImag * imag2imag_) / rangePerPixel_);
    }
    int realHitPix = realCenterPix_ + deltaRealPix;
    if (realHitPix < 0 || realHitPix >= width_) {
      return false;
    }
    int imagHitPix = imagCenterPix_ + deltaImagPix;
    if (imagHitPix < 0 || imagHitPix >= height_) {
      return false;
    }
    int pixelIndex = imagHitPix * width_ + realHitPix;
    colorMapHit_ = pixels_[pixelIndex];
    if (transparentPixelP_ && colorMapHit_ == transparentPixel_) {
      return false;
    }
//    if (debugCount_++ < 6) {
//      System.out.println("TrapFrame.hitP returning T");
//    }
    return true;
  }
  
  public int hitColorMapIndex ()
  {
    int answer = colorMapHit_;
    if (answer < 0) {
//      if (debugCount_++ < 6) {
//        System.out.println("TrapFrame.hitColorMapIndex returning " + (answer + 256));
//      }
      return answer + 256;
    }
//      if (debugCount_++ < 6) {
//        System.out.println("TrapFrame.hitColorMapIndex returning " + answer);
//      }
    return answer;
  }

  /* ------------------------ LISTENERS ---------------------- */

  class MyWindowLNR implements WindowListener
  {
    public void windowOpened (WindowEvent e)
    {
      //jQCentral_.message("TrapFrame$MyWindowLNR.windowOpened", false);
      jQCentral_.setTrapFrameIsTargetP();
      jQCentral_.updateColorMapEd(false, self_);
    }

    public void windowClosing (WindowEvent e)
    {
      //jQCentral_.message("TrapFrame$MyWindowLNR.windowClosing", false);
      self_.setVisible(false);
    }

    public void windowClosed (WindowEvent e)
    {
      //jQCentral_.message("TrapFrame$MyWindowLNR.windowClosed", false);
    }

    public void windowIconified (WindowEvent e)
    {
//    jQCentral_.message("TrapFrame$MyWindowLNR.windowIconified", false);
    }

    public void windowDeiconified (WindowEvent e)
    {
//      jQCentral_.message("TrapFrame$MyWindowLNR.windowDeiconified", true);
      jQCentral_.setTrapFrameIsTargetP();
      jQCentral_.updateColorMapEd(false, self_);
    }

    public void windowActivated (WindowEvent e)
    {
//      System.out.println("TrapFrame.windowActivated set lastCommandWasAdjust_ = false;");
      jQCentral_.setTrapFrameIsTargetP();
      jQCentral_.updateColorMapEd(false, self_);
    }

    public void windowDeactivated (WindowEvent e)
    {
      //jQCentral_.message("TrapFrame$MyWindowLNR.windowDeactivated", false);
    }
  }
 
  class trapFileLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.openTrapImage();
    }
  }
  
  
  class cMapLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.updateColorMapEd(true, self_);
    }
  }
  
  class trans0LNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      jQCentral_.message("Transparent index is being moved to 0.", false);
      int transparentIntPixel = transparentPixel_ < 0 ? transparentPixel_ + 256 : transparentPixel_;
      swapRGBs(0, transparentIntPixel);
      swapPixelIndices(transparentPixel_, (byte)0);
      transparentPixel_ = 0;
      updateColorModel();
      jQCentral_.setIndexAndColor(0, getColorAt(0), false);
      jQCentral_.setIndexAndColor(transparentIntPixel, getColorAt(transparentIntPixel), true);
    }
  }
  
  class saveLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      writeGIF();
    }
  }
  
  class closeLNR implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      setVisible(false);
    }
  }
  
  
}
