package JavaQuat20;

import java.awt.*;
import java.awt.image.*;
import java.util.*;
import JavaQuat20.FractalLoops.*;

class TrapImageProducer extends Q1ImageProducer 
{
  TrapFrame trap_ = null;
  private int debugCount_ = 0;
  
  TrapImageProducer(FractalSpec arg, JQCanvas p, TrapFrame trap) 
    // This is for normal creation when we are going to compute a fractal
  {
    super(arg, p);
    //System.out.println("TrapImageProducer constructor 1");
    trap_ = trap;
  }
   
  TrapImageProducer(Image refresher, FractalSpec arg, JQCanvas p, TrapFrame trap)
    // This is for dummy creation, when we have loaded an image file, GIF or PNG.
  {
    super(refresher, arg, p);
    //System.out.println("TrapImageProducer constructor 2");
    trap_ = trap;
  }
  
  public boolean trapImageProducerP ()
  {
    return true;
  }
  
  public double call_loop_function(FastQuaternion point_in, FastQuaternion coeff, 
                                   FastQuaternion trans, int depth)
  {
//    if (debugCount_++ < 5) {
//      System.out.println("TrarImageProducer.call_loop_function called.");
//    }
    if (spec_.pts_per_pixel == 1) {
      return loop_function.trapLoop(point_in, coeff, trans, depth, trap_);
    }
    FastQuaternion point = new FastQuaternion(point_in);
    if (spec_.pts_per_pixel == 2) {  // Currently pts_per_pixel is never 2.
      FastQuaternion halfx = incxFast.times_scalar(0.5);
      FastQuaternion halfy = incyFast.times_scalar(0.5);
      double loops = loop_function.trapLoop(point, coeff, trans, depth, trap_);
      point.plus_equal(halfx).plus_equal(halfy);
      loops += loop_function.trapLoop(point, coeff, trans, depth, trap_);
      if (spec_.colorsIs256_p)
        return Math.round(loops / 2.0);
      return loops / (double)2.0;
    }
    if (spec_.pts_per_pixel <= 4) {
      FastQuaternion halfx = incxFast.times_scalar(0.5);
      FastQuaternion halfy = incyFast.times_scalar(0.5);
      FastQuaternion point2 = point.plus(halfx);
      double loops = 0;
      for (int i = 0; ; i++) {
        loops += loop_function.trapLoop(point, coeff, trans, depth, trap_);
        loops += loop_function.trapLoop(point2, coeff, trans, depth, trap_);
        if (i == 1) {
          if (spec_.colorsIs256_p)           //+++++++++++++++++++
            return Math.round(loops / 4.0);
          return loops / (double)4.0;
        }
        point.plus_equal(halfy);
        point2.plus_equal(halfy);
      }
    }
    if (spec_.pts_per_pixel > 4) {
      FastQuaternion thirdx = incxFast.times_scalar(0.333333333333);
      FastQuaternion thirdy = incyFast.times_scalar(0.333333333333);
      FastQuaternion point2 = point.plus(thirdx);
      FastQuaternion point3 = point2.plus(thirdx);
      double loops = 0;
      for (int i = 0; ; i++) {
        loops += loop_function.trapLoop(point, coeff, trans, depth, trap_);
        loops += loop_function.trapLoop(point2, coeff, trans, depth, trap_);
        loops += loop_function.trapLoop(point3, coeff, trans, depth, trap_);
        if (i == 2) {
          if (spec_.colorsIs256_p)           //+++++++++++++++++++
            return Math.round(loops / 9.0);
          return loops / (double)9.0;
        }
        point.plus_equal(thirdy);
        point2.plus_equal(thirdy);
        point3.plus_equal(thirdy);
      }      
    } 
    System.err.println("Q1ImageProducer::call_loop_function error: " + 
                        spec_.pts_per_pixel);
    return (double)loop_function.trapLoop(point, coeff, trans, depth, trap_);
  }
  
  // The following methods preserve the 'loop count' as the new color map index.
  int reverseLoops256 (int depth, double loops)
  {
    return Math.min(depth, (int)loops);
  }
  
  double reverseLoopsMillions (int depth, double loops)
  {
    return loops;
  }

  double pixout(int x, int y, int loops, int gdidx, boolean recordd)
    //loops is actually max-loops minus actual loops, so 0 means non-divergence and to be black.
  {
    //System.out.println("Q1ImageProducer::pixout ent");
    loops = Math.min(255, loops);
    loops = Math.max(0, loops);
    
    // Record the real value
    if (recordd) {
      guide.aset(gdidx, x, loops);
    }
        
    // Draw the point.
    int i = x + y * width;
    pixels_[i] = (byte)loops;

    // In the following line the returned value indicates whether the graphics instance is 
    // present.
    // So if the graphics instance is gone and we were not planning to stop, plan to stop.
    Color indexColor = parent.getColorAt(loops);
    if (!parent.keepCurrentDrawLine(x, x, y, indexColor) && !stop_p_)
      stop_p_ = true;
    //System.out.println("Q1ImageProducer::pixout ex");
    return loops;
  }

  void frDrawline(int xleft, int xright, int yndx, int loops)
    // 256 color version
  {
    //System.out.println("Q1ImageProducer::frDrawline ent");
    // ldivc is a number recorded by pixout, so we pick up the computations for
    // drawing from there
    
    // Draw the horizontal line.
    int end = xright + yndx * width;
    for (int i = xleft + yndx * width; i <= end; i++) {
      pixels_[i] = (byte)loops;
    }    
    
    Color indexColor = parent.getColorAt(loops);
    parent.keepCurrentDrawLine(xleft, xright, yndx, indexColor);
    //System.out.println("Q1ImageProducer::frDrawline ex");
  }


}


